# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
ContentDataStorage the main handler for content data storage functional.

@package: app
@file ContentDataStorage.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import dc.EventObjects as dc_event
import app.Utils as Utils
from dc_crawler.DBTasksWrapper import DBTasksWrapper


class ContentDataStorage(object):
  # # Constants used in class
  PUT_DICT_OPTION_ID = "id"
  PUT_DICT_OPTION_DATA = "data"
  PUT_DICT_OPTION_CDATE = "cDate"

  contentTypesDict = {
    dc_event.Content.CONTENT_RAW_CONTENT:'RAW_CONTENT',
    dc_event.Content.CONTENT_TIDY_CONTENT:'TIDY_CONTENT',
    dc_event.Content.CONTENT_HEADERS_CONTENT:'HEADERS_CONTENT',
    dc_event.Content.CONTENT_REQUESTS_CONTENT:'REQUESTS_CONTENT',
    dc_event.Content.CONTENT_META_CONTENT:'META_CONTENT',
    dc_event.Content.CONTENT_COOKIES_CONTENT:'COOKIES_CONTENT',
    dc_event.Content.CONTENT_DYNAMIC_CONTENT:'DYNAMIC_CONTENT',
    dc_event.Content.CONTENT_PROCESSOR_CONTENT:'PROCESSOR_CONTENT',
    dc_event.Content.CONTENT_CHAIN_PARTS:'CHAIN_PARTS'
    }

  DEFAULT_VALUE_CONTENT_TYPE_MASK = dc_event.URLContentRequest.CONTENT_TYPE_RAW_LAST + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_RAW + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_TIDY + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_HEADERS + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_REQUESTS + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_META + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_COOKIES + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_DYNAMIC + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_PROCESSED + \
                                    dc_event.URLContentRequest.CONTENT_TYPE_PROCESSED_ALL

  # # Error messages used in class
  ERROR_INPUT_DATA = "Wrong input data '%s' = %s"
  ERROR_TYPE_CONTENT = "Wrong value a type of content: %s"

  # # Initialization
  def __init__(self):
    pass


  ## Get content type name
  #
  # @param contentType - type of content
  # @param log - logger instance
  # @return content type name
  @staticmethod
  def getContentTypeName(contentType, log=None):
    #variable for result
    ret = ''
    try:
      if not isinstance(contentType, int):
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('contentType', str(contentType)))

      if not contentType in ContentDataStorage.contentTypesDict:
        raise Exception(ContentDataStorage.ERROR_TYPE_CONTENT % str(contentType))

      ret = ContentDataStorage.contentTypesDict[contentType]

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  # # Get contents
  #
  # @param dbWrapper - DBTasksWrapper instance
  # @param siteId - site/project ID
  # @param url - URLs string
  # @param contentTypeMask - the content types mask defines types of content that will be collected and returned
  # @param log - logger instance
  # @return dictionary with extracted contents (content type as key and content as value)
  @staticmethod
  def getContents(dbWrapper, siteId, url, contentTypeMask=DEFAULT_VALUE_CONTENT_TYPE_MASK, log=None):
    # variables for result
    ret = {}

    # #Local function for check and print of data
    #
    # @param rawContentData - raw content data
    # @param rawContentName - raw content name
    # @param log - logger instance
    def printContentStatus(rawContentData, rawContentName, log=None):
      if rawContentData is not None:
        if log is not None:
          log.debug("Some %s content size %s on disk", rawContentName, str(len(rawContentData)))
      else:
        if log is not None:
          log.debug("NO %s content on disk", rawContentName)


    try:
      if not isinstance(dbWrapper, DBTasksWrapper):
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('dbWrapper', str(dbWrapper)))

      if not isinstance(siteId, basestring) or siteId == "":
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('siteId', str(siteId)))

      if not isinstance(url, basestring) or url == "":
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('url', str(url)))

      if not isinstance(contentTypeMask, int):
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('contentTypeMask', str(contentTypeMask)))

      # make request
      urlContentObj = dc_event.URLContentRequest(siteId, url, contentTypeMask)
      rawContentData = dbWrapper.urlContent([urlContentObj])
#       if log is not None:
#         log.debug("UrlContentResponse:\n%s", Utils.varDump(rawContentData, stringifyType=0))

      if rawContentData is not None and len(rawContentData) > 0:
        # get processed content
        if isinstance(rawContentData[0].processedContents, list) and len(rawContentData[0].processedContents) > 0 and \
          isinstance(rawContentData[0].processedContents[0], dc_event.Content):
          ret[dc_event.Content.CONTENT_PROCESSOR_CONTENT] = rawContentData[0].processedContents[0].buffer

        # get headers content
        if isinstance(rawContentData[0].headers, list) and len(rawContentData[0].headers) > 0 and isinstance(rawContentData[0].headers[0], dc_event.Content):
          ret[dc_event.Content.CONTENT_HEADERS_CONTENT] = rawContentData[0].headers[0].buffer

        # get requests content
        if isinstance(rawContentData[0].requests, list) and len(rawContentData[0].requests) > 0 and isinstance(rawContentData[0].requests[0], dc_event.Content):
          ret[dc_event.Content.CONTENT_REQUESTS_CONTENT] = rawContentData[0].requests[0].buffer

        # get meta content
        if isinstance(rawContentData[0].meta, list) and len(rawContentData[0].meta) > 0 and isinstance(rawContentData[0].meta[0], dc_event.Content):
          ret[dc_event.Content.CONTENT_META_CONTENT] = rawContentData[0].meta[0].buffer

        # get cookies content
        if isinstance(rawContentData[0].cookies, list) and len(rawContentData[0].cookies) > 0 and isinstance(rawContentData[0].cookies[0], dc_event.Content):
          ret[dc_event.Content.CONTENT_COOKIES_CONTENT] = rawContentData[0].cookies[0].buffer

        if isinstance(rawContentData[0].rawContents, list):
          # get raw content
          if len(rawContentData[0].rawContents) > 0 and isinstance(rawContentData[0].rawContents[0], dc_event.Content):
            ret[dc_event.Content.CONTENT_RAW_CONTENT] = rawContentData[0].rawContents[0].buffer

          # get dynamic content
          if len(rawContentData[0].rawContents) > 1 and isinstance(rawContentData[0].rawContents[1], dc_event.Content):
            ret[dc_event.Content.CONTENT_DYNAMIC_CONTENT] = rawContentData[0].rawContents[1].buffer

        # print log information
        for contentType, contentValue in ret.items():
          printContentStatus(contentValue, ContentDataStorage.getContentTypeName(contentType, log))

    except Exception, err:
      if log is not None:
        log.error(str(err))
        log.info(Utils.getTracebackInfo())

    return ret

#   # # Put the content use certain request type
#   #
#   # @param dbWrapper - DBTasksWrapper instance
#   # @param siteId - site/project ID
#   # @param urlMd5 - urlMd5 string
#   # @param rawContentData - raw content data value as string
#   # @param contentType - type of content
#   # @param log - logger instance
#   # @return - None
#   @staticmethod
#   def putContentOfType(dbWrapper, siteId, urlMd5, rawContentData, contentType, log=None):
#     try:
#       if not isinstance(siteId, basestring) or siteId == "":
#         raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('siteId', str(siteId)))
#
#       if not isinstance(urlMd5, basestring) or urlMd5 == "":
#         raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('urlMd5', str(urlMd5)))
#
#       if not isinstance(contentType, int):
#         raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('contentType', str(contentType)))
#
#       rawContent = ''
#       if rawContentData is not None:
#         rawContent = rawContentData
#
#       # put raw content to DB
#       putDict = {}
#       putDict[ContentDataStorage.PUT_DICT_OPTION_ID] = urlMd5
#       putDict[ContentDataStorage.PUT_DICT_OPTION_DATA] = rawContent
#       putDict[ContentDataStorage.PUT_DICT_OPTION_CDATE] = Utils.SQLExpression("NOW()")
#
#       if isinstance(dbWrapper, DBTasksWrapper):
#         dbWrapper.putURLContent([dc_event.URLPut(siteId, urlMd5, contentType, putDict)])
#
#     except Exception, err:
#       if log is not None:
#         log.error(str(err))
#         log.info(Utils.getTracebackInfo())


  # # Put contents
  #
  # @param dbWrapper - DBTasksWrapper instance
  # @param siteId - site/project ID
  # @param urlMd5 - urlMd5 string
  # @param contentsDict - dictionary with different contents (content type as key and content as value)
  # @param log - logger instance
  # @return - None
  @staticmethod
  def putContents(dbWrapper, siteId, urlMd5, contentsDict, log=None):
    try:
      if not isinstance(dbWrapper, DBTasksWrapper):
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('dbWrapper', str(dbWrapper)))

      if not isinstance(siteId, basestring) or siteId == "":
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('siteId', str(siteId)))

      if not isinstance(urlMd5, basestring) or urlMd5 == "":
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('urlMd5', str(urlMd5)))

      if not isinstance(contentsDict, dict):
        raise Exception(ContentDataStorage.ERROR_INPUT_DATA % ('contentsDict', str(contentsDict)))

      urlPutList = []
      for contentType, contentValue in contentsDict.items():
        putDict = {}
        putDict[ContentDataStorage.PUT_DICT_OPTION_ID] = urlMd5
        putDict[ContentDataStorage.PUT_DICT_OPTION_DATA] = contentValue
        putDict[ContentDataStorage.PUT_DICT_OPTION_CDATE] = Utils.SQLExpression("NOW()")
        urlPutList.append(dc_event.URLPut(siteId, urlMd5, contentType, putDict))

      dbWrapper.putURLContent(urlPutList)

    except Exception, err:
      if log is not None:
        log.error(str(err))
