'''
Created on Aug 27, 2019

@package: app
@file HostRequestStorage.py
@author Evgeniy Arefa <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
'''

import json
import re
import time
import os
import tempfile
import random

import app.Utils as Utils


# # HostRequestStorage Class, implements functional http delay by hosts frequence
#
class HostRequestStorage(object):

  MSG_ERROR_MAKEDIR = ">>> makedir exception "
  MSG_ERROR_DEFAULT_DELAY = ">>> there isn't correctly value of min_delay. Get DEFAULT_DELAY = %s sec"

  MSG_DEBUG_NO_FIND_LIMIT = ">>> there aren't limits for this host and this domain"
  MSG_DEBUG_FIND_LIMIT = "Finded next limits: %s"
  MSG_DEBUG_PATH_OF_TEMP = "Path of temp: %s"
  MSG_DEBUG_WRONG_IN_METHOD = ">>> wrong in HostRequestStorage.availableLimitHost method; "
  MSG_DEBUG_USE_RANDOMIZED = "Use 'randomized' for delay: %s"

  ERROR_INPUT_URL = "Incorrect value of Url"
  ERROR_INPUT_SITE_ID = "Incorrect value of Site Id"

  FILE_SUFF = ".json"
  PATH_STORAGE_KEY = "storage"
  HTTP_FILE_PREFIX = "http_limits_"
  LAST_START = "last_start"
  MIN_DELAY = "min_delay"
  MAX_DELAY = "max_delay"
  DEFAULT_DELAY = 60
  RANDOMIZED = "randomized"

  def __init__(self):
    pass

  # #Safe conversion of limits from request to a dictionary
  #
  # @hostRequestStorage - limits of request from HTTP_FREQ_LIMITS parameter
  # @return Dict of limits from request or None
  @staticmethod
  def getLimitsToDict(hostRequestStorage, log=None):
    ret = None

    try:
      ret = json.loads(hostRequestStorage)

    except Exception, err:
      if log is not None:
        log.error(">>> wrong with limits json.loads: " + str(err))

    return ret

  # #Find limits for selected IP and domain
  #
  # @url - url of request
  # @propertiesDict - all limits of request from HTTP_FREQ_LIMITS parameter
  # @path - path of temporary dir in system
  # @return selected limits of Dict or None and full path for temporary dir
  @staticmethod
  def gethttpLimit(url, propertiesDict, path, log=None):
    ret = None

    try:
      if not path:
        path = tempfile.gettempdir()
      path = os.path.join(path, Utils.UrlParser.getDomain(url))

      for elem in propertiesDict:
        if len(propertiesDict[elem].values()) > 0 and isinstance(propertiesDict[elem].values()[0], dict):
          if re.search(elem, url):
            for item in propertiesDict[elem]:
              if item in (Utils.getCurrentHostName(), Utils.getIP()):
                ret = propertiesDict[elem][item]
                path = os.path.join(path, item)
                break
          else:
            continue

        CurrentHostIP = []
        for item in (Utils.getCurrentHostName(), Utils.getIP()):
          if isinstance(item, basestring) and item:
            CurrentHostIP.append(item)
        if Utils.getCurrentIPInterfaces() and isinstance(Utils.getCurrentIPInterfaces(), dict):
          for key, values in Utils.getCurrentIPInterfaces().items():
            if isinstance(values, basestring) and values:
              CurrentHostIP.append(values)
            elif isinstance(values, list):
              for value in values:
                if isinstance(value, basestring) and value:
                  CurrentHostIP.append(value)

        if elem in CurrentHostIP:
          ret = propertiesDict[elem]
          path = os.path.join(path, elem)
          break

    except Exception, err:
      if log is not None:
        log.error(str(err))

    if log is not None:
      if ret is None:
        log.debug(HostRequestStorage.MSG_DEBUG_NO_FIND_LIMIT)
      else:
        log.debug(HostRequestStorage.MSG_DEBUG_FIND_LIMIT, str(ret))

    return ret, path

  # #Get time of last start from temporary file and make a directory if it is missing
  #
  # @path - path of temporary dir in system
  # @temporaryFile - path and name of temporary file
  # @return dict whith time in int format
  @staticmethod
  def getLastStartFromFile(path, temporaryFile, log=None):
    ret = {HostRequestStorage.LAST_START: 0}
    if os.path.exists(path):
      ret = Utils.jsonLoadsSafe(jsonString=Utils.FileSafe.read(temporaryFile), default=ret, log=log)
    else:
      try:
        os.makedirs(path)
      except Exception as excp:
        if log is not None:
          log.error(HostRequestStorage.MSG_ERROR_MAKEDIR + str(excp))
          log.info(Utils.getTracebackInfo())

    return ret

  # #Check available host for this request
  #
  # @url - url of request
  # @siteId - id of site
  # @properties - properties limits of request from HTTP_FREQ_LIMITS parameter
  # @return Boolean value
  @staticmethod
  def isAvailableHost(url, siteId, properties, log=None):
    ret = True

    try:
      if not isinstance(url, basestring) or not url:
        raise IOError(HostRequestStorage.ERROR_INPUT_URL)
      elif not isinstance(siteId, basestring) or not siteId:
        raise IOError(HostRequestStorage.ERROR_INPUT_SITE_ID)

      propertiesDict = HostRequestStorage.getLimitsToDict(properties, log)
      path = propertiesDict.pop(HostRequestStorage.PATH_STORAGE_KEY, None)
      limitDict, path = HostRequestStorage.gethttpLimit(url, propertiesDict, path, log)
      temporaryFile = os.path.join(path, HostRequestStorage.HTTP_FILE_PREFIX + str(siteId) + HostRequestStorage.FILE_SUFF)
      if log is not None:
        log.debug(HostRequestStorage.MSG_DEBUG_PATH_OF_TEMP, str(path))

      if limitDict is not None:
        lastTimeDict = HostRequestStorage.getLastStartFromFile(path, temporaryFile, log=log)
        lastStart = lastTimeDict.get(HostRequestStorage.LAST_START) if lastTimeDict.get(HostRequestStorage.LAST_START) else 0
        lastStart = lastStart if lastStart <= int(time.time()) else 0

        minDelay, maxDelay, randomized = None, None, None
        for index, limit in enumerate(limitDict):
          if limit.lower() == HostRequestStorage.MIN_DELAY.lower():
            minDelay = Utils.toInt(limitDict.get(limit), log=log)
          elif limit.lower() == HostRequestStorage.MAX_DELAY.lower():
            maxDelay = Utils.toInt(limitDict.get(limit), log=log)
          elif limit.lower() == HostRequestStorage.RANDOMIZED.lower():
            randomized = Utils.toInt(limitDict.get(limit), log=log)

        if not minDelay:
          minDelay = HostRequestStorage.DEFAULT_DELAY
          if log is not None:
            log.error(HostRequestStorage.MSG_ERROR_DEFAULT_DELAY, HostRequestStorage.DEFAULT_DELAY)

        if randomized > 0 and maxDelay > minDelay:
          delayTime = random.randint(minDelay, maxDelay)
          if log is not None:
            log.debug(HostRequestStorage.MSG_DEBUG_USE_RANDOMIZED, delayTime)
        else:
          delayTime = minDelay or maxDelay

        if int(time.time()) < (lastStart + delayTime):
          ret = False
        else:
          Utils.FileSafe.write(temporaryFile, json.dumps({HostRequestStorage.LAST_START: int(time.time())}))

    except Exception, err:
      ret = False
      if log is not None:
        log.error(str(err))
        log.info(Utils.getTracebackInfo())

    return ret
