# coding: utf-8

"""
HCE project, Python bindings, Distributed Tasks Manager application.
Collect URLs by xpath list functional.

@package: dc_crawler
@file CollectURLsXPath.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import re
import time
import datetime
import lxml.html
import lxml.etree

from app.Filters import Filters
import app.Utils as Utils
import app.Consts as APP_CONSTS


class CollectURLsXPath(object):
  # # Constants used in class
  MSG_DEBUG_NOT_COLLECTED = "Warning! No one XPath set matched URL %s, URLs not collected!"
  MSG_DEBUG_USE_XPATH_LIST_FROM_PROPERTY = "Site has '%s' property: %s"
  MSG_DEBUG_USE_XPATH_LIST_FROM_CONFIG = "Site has '%s' property, but use xpath list from configuration file: %s"

  OPTION_SETS = 'sets'
  OPTION_MODE = 'mode'
  OPTION_DATE_FORMAT = 'date_format'

  MODE_VALUE_APPEND = 1

  DEFAULT_VALUE_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
  
  HREF_NAME = 'href'

  # # Constants of error messages used in class
  MSG_ERROR_BAD_PROPERTY_FORMAT = "The property '%s' has  wrong format: %s"


  # # Initialization
  def __init__(self):
    pass


  # # Evaluate date macro
  #
  # @param localPattern - input pattern as string
  # @param dateFromat - format for %@DATE... macro
  # @param log - logger instance
  # @return patterns list with date macro evaluated and replaced with values
  @staticmethod
  def evaluateDateMacro(localPattern, dateFromat, log=None):
    try:
      if log is not None:
        log.info("dateFromat: %s", str(dateFromat))

      regex = re.compile("%@(DATE)\\(([\\+|\\-]\\d{1,2})\\)%")
      matchArray = regex.findall(localPattern)
      if log is not None:
        log.info("matchArray: %s", str(matchArray))

      for i in matchArray:
        t = time.strftime(dateFromat, time.gmtime(time.time() + datetime.timedelta(hours=(+int(i[1]))).seconds))
        localPattern = localPattern.replace("%@" + i[0] + "(" + i[1] + ")%", t)

      localPattern = Filters.getGmtTime(localPattern, log)

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return localPattern


  # # execute collect URLs by xpath
  #
  # @param siteProperties - site properties dictionary
  # @param urlSet - incoming URLs list (links from page)
  # @param dom - page's DOM model
  # @param urlsXpathListFromConfig - URLs xpath list from configuration file
  # @param url - URL string
  # @param log - logger instance
  # @return urlSet - filled URLs list
  @staticmethod
  def execute(siteProperties, urlSet, dom, urlsXpathListFromConfig, url, log=None):
    try:
      if isinstance(siteProperties, dict)  and APP_CONSTS.COLLECT_URLS_XPATH_LIST in siteProperties:
        urlXpathList = Utils.jsonLoadsSafe(jsonString=siteProperties[APP_CONSTS.COLLECT_URLS_XPATH_LIST],
                                           default=siteProperties[APP_CONSTS.COLLECT_URLS_XPATH_LIST],
                                           log=log)
        if len(urlXpathList) > 0:
          if log is not None:
            log.debug(CollectURLsXPath.MSG_DEBUG_USE_XPATH_LIST_FROM_PROPERTY, str(APP_CONSTS.COLLECT_URLS_XPATH_LIST), str(urlXpathList))
        else:
          # Set URLs xpath list from configuration file
          urlXpathList = {CollectURLsXPath.OPTION_SETS: {'': urlsXpathListFromConfig}}
          if log is not None:
            log.debug(CollectURLsXPath.MSG_DEBUG_USE_XPATH_LIST_FROM_CONFIG, str(APP_CONSTS.COLLECT_URLS_XPATH_LIST), str(urlXpathList))

        if not isinstance(urlXpathList, dict) or CollectURLsXPath.OPTION_SETS not in urlXpathList or \
          not isinstance(urlXpathList[CollectURLsXPath.OPTION_SETS], dict):
          raise Exception(CollectURLsXPath.MSG_ERROR_BAD_PROPERTY_FORMAT % (APP_CONSTS.COLLECT_URLS_XPATH_LIST, str(urlXpathList)))

        matchedSets = 0
        if CollectURLsXPath.OPTION_DATE_FORMAT in urlXpathList:
          dformat = str(urlXpathList[CollectURLsXPath.OPTION_DATE_FORMAT])
        else:
          dformat = CollectURLsXPath.DEFAULT_VALUE_DATE_FORMAT

        for rexpr in urlXpathList[CollectURLsXPath.OPTION_SETS]:
          if rexpr == '' or re.search(rexpr, url) is not None:
            if CollectURLsXPath.OPTION_MODE in urlXpathList and int(urlXpathList[CollectURLsXPath.OPTION_MODE]) == CollectURLsXPath.MODE_VALUE_APPEND:
              xpathl = urlsXpathListFromConfig + urlXpathList[CollectURLsXPath.OPTION_SETS][rexpr]
            else:
              xpathl = urlXpathList[CollectURLsXPath.OPTION_SETS][rexpr]

            matchedSets += 1
            for xpath in xpathl:
              xpathPattern = CollectURLsXPath.evaluateDateMacro(xpath, dformat, log)
              if log is not None:
                log.debug("xpathPattern: '%s'", str(xpathPattern))

              elements = dom.xpath(xpathPattern)

              if log is not None:
                log.debug("dom.xpath() return: %s, type = %s", Utils.varDump(elements), str(type(elements)))

#                 if isinstance(elem, list):
#                   for el in elem:
#                     log.debug("type = %s, attrib: %s, tail: %s, text: %s, get(): %s",
#                               str(type(el)), Utils.varDump(el.attrib), Utils.varDump(el.tail), Utils.varDump(el.text), Utils.varDump(el.get('href')))

#               if isinstance(elem, list) and len(elem) > 0 and hasattr(elem[0], "tail"):
#                 urlSet.extend([el.tail for el in elem])
#               elif isinstance(elem, list) and len(elem) > 0 and isinstance(elem[0], lxml.html.HtmlElement):
#                 urlSet.extend([el.text for el in elem])
#               else:
#                 urlSet.extend(elem)

              if isinstance(elements, list):
                for element in elements:
                  if isinstance(element, lxml.etree._Element):  # pylint: disable=E1101,W0212
                    extractedURL = element.get(CollectURLsXPath.HREF_NAME)
                    if log is not None:
                      log.debug("extractedURL = %s", str(extractedURL))

                    if extractedURL is not None:
                      urlSet.append(extractedURL)

                  else:
                    urlSet.append(element)

        if matchedSets == 0 and log is not None:
          log.warning(CollectURLsXPath.MSG_DEBUG_NOT_COLLECTED, str(url))

    except Exception, err:
      if log is not None:
        log.error(str(err))
        log.info(Utils.getTracebackInfo())

    return urlSet
