# coding: utf-8

"""
HCE project, Python bindings, Distributed Tasks Manager application.
Event objects definitions.

@package: dc
@file CrawlerTask.py
@author scorp <developers.hce@gmail.com>, Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""
import os.path
import sys
import hashlib
import datetime

try:
  import cPickle as pickle
except ImportError:
  import pickle

import time
from time import strftime

from collections import namedtuple
from collections import OrderedDict

import signal
import logging.config
import ConfigParser
import re
import types
import copy
import json
import base64
import math

# import urllib
import lxml.etree
import requests
from cement.core import foundation
from dc_crawler.CollectURLs import CollectURLs
from dc_crawler.CollectProperties import CollectProperties
from dc_crawler.DetectModified import DetectModified
from dc_crawler.Fetcher import BaseFetcher
from dc_crawler.FetcherType import FetcherType
from dc_crawler.Fetcher import Response
from dc_crawler.Fetcher import SeleniumFetcher
from dc_crawler.CrawledResource import CrawledResource
from dc_crawler.UrlSchema import UrlSchema
from dc_crawler.Exceptions import SyncronizeException
from dc_crawler.Exceptions import TimeoutException
from dc_crawler.Exceptions import ConnectionErrorException
from dc_crawler.Exceptions import HTTPErrorException
from dc_crawler.Exceptions import URLRequiredException
from dc_crawler.Exceptions import InternalCrawlerException
from dc_crawler.Exceptions import BadContentTypeException
from dc_crawler.Exceptions import NotModifiedException
from dc_crawler.DBTasksWrapper import DBTasksWrapper
from dc_crawler.RobotsParser import RobotsParser
from dc_crawler.RefererHeaderResolver import RefererHeaderResolver
from dc_crawler.HTTPHeadersExtractor import HTTPHeadersExtractor
from dc_crawler.HTTPCookieResolver import HTTPCookieResolver
# from dc_crawler.HTTPProxyResolver import HTTPProxyResolver
# from dc_crawler.HTTPProxy import HTTPProxy
from dc_crawler.RequestsRedirectWrapper import RequestsRedirectWrapper
from dc_crawler.ResourceProcess import ResourceProcess
from dc_crawler.URLProcess import URLProcess
from dc_crawler.HTMLCanonicalRedirect import HTMLCanonicalRedirect
import dc_crawler.Constants as CONSTS
from dc_crawler.UrlChain import UrlChain
from dc_crawler.MimeTypeExtractor import MimeTypeExtractor
from dc_crawler.CrawlerDataStorage import CrawlerDataStorage
from dc_crawler.HTTPHeaders import HTTPHeaders
from dc_crawler.HTTPRedirectExtractor import HTTPRedirectExtractor
from dc_crawler.HTTPRedirectResolver import HTTPRedirectResolver

from app.Utils import varDump
from app.Utils import PathMaker
from app.Filters import Filters
from app.Utils import SQLExpression
from app.LFSDataStorage import LFSDataStorage
from app.HostRequestStorage import HostRequestStorage
from app.PDateFormat import PDateFormat
from app.PDateExtractor import PDateExtractor
from app.PDateSourceMask import PDateSourceMask
from app.Exceptions import SeleniumFetcherException
from app.Exceptions import UrlAvailableException
from app.Exceptions import DatabaseException
from app.Exceptions import ProxyException
from app.Utils import ExceptionLog
from app.Utils import UrlNormalizator
from app.FieldsSQLExpressionEvaluator import FieldsSQLExpressionEvaluator
import app.Profiler
import app.Consts as APP_CONSTS
import app.Utils as Utils
from app.UrlNormalize import UrlNormalize
from app.LogSwitch import LogSwitch
import dc.Constants as DC_CONSTS
import dc.EventObjects as dc_event
from dc.EventObjects import Site
from dc.EventObjects import Batch
import dc_processor.Constants as PCONSTS

DB_SITES = "dc_sites"
DB_URLS = "dc_urls"

MSG_ERROR_LOAD_CONFIG = "Error loading config file. Exciting. "
MSG_ERROR_LOAD_OPTIONS = "Error loading options. Exciting. "
MSG_ERROR_LOAD_LOG_CONFIG_FILE = "Can't load logging config file. Exiting. "
MSG_ERROR_LOAD_SITE_DATA = "Can't load site data: "
MSG_ERROR_UPDATE_SITE_DATA = "Can't update site data: "
MSG_ERROR_LOAD_URL_DATA = "Can't load url data: "
MSG_ERROR_PROCESS_BATCH_ITEM = "Can't process batch item "
MSG_ERROR_WRITE_CRAWLED_DATA = "Can't write crawled data "
MSG_ERROR_COLLECT_URLS = "Can't collect urls "
MSG_ERROR_ADD_URL_TO_BATCH_ITEM = "Can't add url to batch item "
MSG_ERROR_LOAD_SITE_PROPERTIES = "Can't load site properties "
MSG_ERROR_CRAWL_SITE = "Can't crawl site "
MSG_ERROR_CHECK_SITE = "Site don't passed check site "
MSG_ERROR_GET_DIR = "Can't get dir "
MSG_ERROR_READ_SITE_FROM_DB = "Can't read site data from DB for SiteID: '%s'"
MSG_ERROR_EMPTY_RESPONSE_SIZE = "Empty response"
MSG_ERROR_NOT_EXIST_ANY_VALID_PROXY = "Not exist any valid proxy"
MSG_ERROR_EMPTY_CONFIG_FILE_NAME = "Config file name is empty."
MSG_ERROR_WRONG_CONFIG_FILE_NAME = "Config file name is wrong: %s"
MSG_ERROR_LOAD_APP_CONFIG = "Error loading application config file. %s"
MSG_ERROR_EXTRACT_BASE_URL = "Extract base url failed. Error: %s"


MSG_INFO_PROCESS_BATCH = "ProcessBatch "
MSG_INFO_STORE_COOKIES_FILE = "Store cookies file on disk."

MSG_DEBUG_NON_PROCESSING = "ProcessorName is NONE. Exclude batch item from further processing."

SITE_MD5_EMPTY = "d41d8cd98f00b204e9800998ecf8427e"

DEFAULT_MAX_SIZE = 1000000
EMPTY_RESPONSE_SIZE = "0"

APP_NAME = "crawler-task"

HTTP_COOKIE = "HTTP_COOKIE"
DEFAULT_HTTP_COOKIE = ""
# HTTP_HEADERS = "HTTP_HEADERS"
# DEFAULT_HTTP_HEADER = ""

DC_URLS_DB_NAME = "dc_urls"
DC_URLS_TABLE_PREFIX = "urls_"
DC_SITES_DB_NAME = "dc_sites"
DC_SITES_PROPERTIES_TABLE_NAME = "sites_properties"
DC_SITES_TABLE_NAME = "sites"
DC_URLS_TABLE_NAME = "urls"
COOKIES_FILE_POSTFIX = ".cookies.txt"

NON_PROCESSING = "NONE"

HTTP_REDIRECT = "<Response [301]>"
HTML_REDIRECT = ""
MAX_HTTP_REDIRECTS_UNLIMITED = 0
MAX_HTML_REDIRECTS_UNLIMITED = 0
META_XPATH = "//meta[contains(@content, 'url')]/@content"

Results = namedtuple("Results", "exit_code, output, err")

ROBOTS_PATTERN = re.compile(r'(https?://[^/]+).*', re.I)

TEXT_CONTENT_TYPE_PATTERN = re.compile('text', re.I)

ENV_CRAWLER_STORE_PATH = "ENV_CRAWLER_STORE_PATH"

EXIT_SUCCESS = 0
EXIT_FAILURE = 1

# #The CrawlerTask class, is a interface for fetching content from the web
#
# This object is a run at once application
class CrawlerTask(foundation.CementApp):

  # Dict matches of error masks and http codes
  errorMaskHttpCodeDict = { \
                          APP_CONSTS.ERROR_FETCH_INVALID_URL : SeleniumFetcher.ERROR_NAME_NOT_RESOLVED, \
                          APP_CONSTS.ERROR_FETCHER_INTERNAL : SeleniumFetcher.ERROR_FATAL, \
                          APP_CONSTS.ERROR_FETCHER_INTERNAL : SeleniumFetcher.ERROR_GENERAL, \
                          APP_CONSTS.ERROR_FETCH_TOO_MANY_REDIRECTS : SeleniumFetcher.ERROR_TOO_MANY_REDIRECTS, \
                          APP_CONSTS.ERROR_PROXY_CONNECTION_ERROR : SeleniumFetcher.ERROR_PROXY_CONNECTION_FAILED, \
                          APP_CONSTS.ERROR_FETCH_CONNECTION_TIMEOUT : SeleniumFetcher.ERROR_CONNECTION_TIMED_OUT, \
                          APP_CONSTS.ERROR_FETCH_FORBIDDEN : SeleniumFetcher.ERROR_TUNNEL_CONNECTION_FAILED, \
                          APP_CONSTS.ERROR_EMPTY_RESPONSE : SeleniumFetcher.ERROR_EMPTY_RESPONSE, \
                          APP_CONSTS.ERROR_FETCH_FORBIDDEN : SeleniumFetcher.ERROR_SERVICE_UNAVAILABLE, \
                          APP_CONSTS.ERROR_NOT_EXIST_ANY_VALID_PROXY : SeleniumFetcher.ERROR_PROXY_CONNECTION_FAILED, \
                          APP_CONSTS.ERROR_FETCH_HTTP_ERROR : SeleniumFetcher.ERROR_PROXY_CONNECTION_FAILED
                          }

  # Configuration settings options names
  DB_HOST_CFG_NAME = "db_host"
  DB_PORT_CFG_NAME = "db_port"
  DB_USER_CFG_NAME = "db_user"
  DB_PWD_CFG_NAME = "db_pwd"
  DB_SITES_CFG_NAME = "db_dc_sites"
  DB_URLS_CFG_NAME = "db_dc_urls"

  RAW_DATA_DIR = "raw_data_dir"
  SITE_TEMPLATES = "dc_site_template"
  KEY_VALUE_STORAGE_DIR = "key_value_storage_dir"
  DB_DATA_DIR = "db_data_dir"
  URL_SCHEMA_DIR = "url_schema_data_dir"
  URLS_XPATH_LIST_FILE = "urls_xpath_list_file"

  # Constants used in class
  SEARCH_BASE_URL_PATTERN = r'<base[^>]+href="([^">]+)"'

  NODE_NUMBER_NAME_ENV = "BATCH_NODE_NUMBER"
  NODES_NUMBER_NAME_ENV = "BATCH_NODES_NUMBER"
  BATCH_MODE_NAME_ENV = "BATCH_MODE"

  BATCH_MODE_OFF = 0
  BATCH_MODE_ON = 1

  BATCH_MODE_OFF_NAME = "OFF"
  BATCH_MODE_ON_NAME = "ON"
  BATCH_MODE_DEFAULT_VALUE = BATCH_MODE_OFF


  # Mandatory
  class Meta(object):
    label = APP_NAME
    def __init__(self):
      pass


  # #constructor
  # initialize default fields
  def __init__(self):
    # call base class __init__ method
    foundation.CementApp.__init__(self)
    self.raw_data_dir = None
    self.batchItem = None
    self.logger = None
    self.urlTable = None
    self.batch = None
    self.defaultHeaderFile = None
    self.kvDbDir = None
    self.defaultIcrCrawlTime = None
    self.dbWrapper = None
    self.store_items = []
    self.url = None
    self.max_fetch_time = CONSTS.FETCHER_TIME_LIMIT_MAX
    self.collect_additional_prop = False
    self.exit_code = EXIT_SUCCESS
    self.keep_old_resources = False
    self.urlProcess = URLProcess()
    self.resourceProcess = ResourceProcess()
    self.siteHTTPHeaders = None
    self.headerFileDir = None
    self.robotsFileDir = None
    self.fileHeaders = None
    self.hTTPHeadersStorage = LFSDataStorage()
    self.robotsParser = RobotsParser()
    self.feedItems = None
    self.collectURLsItems = []
    self.errorMask = APP_CONSTS.ERROR_OK
    self.httpApplyHeaders = None
    self.chainDict = OrderedDict()
    self.chainIndex = 0
    self.schemaBatchItems = []
    self.useZeroSiteIdSiteNotExists = False
    self.dir = None
    self.curBatchIterations = 1
    self.urlProcess.currentBatchIterations = self.curBatchIterations
    self.crawledResource = None
    self.realUrl = None
    self.res = None
    self.storeHttpRequest = True
    self.store_http_headers = True
    self.postForms = {}
    self.external_url = None
    self.auto_remove_props = {}
    self.allowedContentTypes = []
    self.allowedWriteContentTypes = []
    self.siteProperties = {}
    self.needStoreMime = None
    self.allow_http_redirects = True
    self.processorName = None
    self.storeCookies = True
    self.dom = None
    self.urlTempalteRegular = None
    self.urlTempalteRealtime = None
    self.urlTempalteRegularEncode = None
    self.urlTempalteRealtimeEncode = None
    self.site = None
    self.siteCacheDict = {}
    self.urlSchemaDataDir = None
    self.normMask = UrlNormalizator.NORM_DEFAULT
    self.urlsXpathList = None
    # status update variables from 'URL_SCHEMA'
    self.statusUpdateEmptyProxyList = None
    self.statusUpdateNoAvailableProxy = None
    self.statusUpdateTriesLimit = None
    self.feedUrl = {}
    self.urlChain = None  # # instance of UrlChain
    self.resetVars()
    # for support max execution time
    self.maxExecutionTimeReached = False
    self.maxExecutionTimeValue = 0
    self.removeUnprocessedItems = False
    self.maxTimeCli = None
    # from config for hard exit
    self.maxExecutionTimeExitUseAbort = False
    self.dbTaskIni = None
    self.inputFile = None
    self.outputFile = None
    self.nodeNumber = os.getenv(self.NODE_NUMBER_NAME_ENV, "")
    if self.nodeNumber == "":
      self.nodeNumber = None
    else:
      self.nodeNumber = int(self.nodeNumber)
    self.nodesNumber = os.getenv(self.NODES_NUMBER_NAME_ENV, "")
    if self.nodesNumber == "":
      self.nodesNumber = None
    else:
      self.nodesNumber = int(self.nodesNumber)

    self.batchMode = os.getenv(self.BATCH_MODE_NAME_ENV, "")
    if self.batchMode == "":
      self.batchMode = self.BATCH_MODE_DEFAULT_VALUE
    else:
      if self.batchMode.isdigit():
        self.batchMode = int(self.batchMode)
      else:
        if self.batchMode == self.BATCH_MODE_OFF_NAME:
          self.batchMode = self.BATCH_MODE_OFF
        elif self.batchMode == self.BATCH_MODE_ON_NAME:
          self.batchMode = self.BATCH_MODE_ON
        else:
          self.batchMode = self.BATCH_MODE_DEFAULT_VALUE


  # # Check is ON batch mode
  #
  # @param - None
  # @return True if batch mode is ON or Fakse otherwise
  def isOnBatchMode(self):
    return int(bool(self.batchMode)) == self.BATCH_MODE_ON


  # #setup
  # setup application
  def setup(self):
    # call base class setup method
    foundation.CementApp.setup(self)


  # #run
  # run application
  def run(self):
    # call base class run method
    foundation.CementApp.run(self)

    # config section
    self.loadConfig()

    # # load cli arguments
    self.loadAppParams()

    # load logger config file
    self.loadLogConfigFile()

    # load mandatory options
    self.loadOptions()

    # load key-value db
    self.loadKeyValueDB()

    # make processing batch data
    self.processBatch()

    # Finish logging
    self.logger.info(APP_CONSTS.LOGGER_DELIMITER_LINE)


  # #RLs collect URL from response body and fills new batch items for batching iterations
  #
  # @return is collect successfully
  def collectURLs(self):
    collectURLsResult = False
    if True or self.dom is not None:
      collectURLs = CollectURLs(isAbortedByTTL=self.isAbortedByTTL)
      collectURLs.crawledResource = self.crawledResource
      collectURLs.url = self.url
      collectURLs.dom = self.dom
      collectURLs.realUrl = self.realUrl
      collectURLs.baseUrl = self.batchItem.urlObj.baseUrl
      collectURLs.processorName = self.processorName
      collectURLs.batchItem = self.batchItem
      collectURLs.siteProperties = self.siteProperties
      collectURLs.site = self.site
      collectURLs.autoRemoveProps = self.auto_remove_props
      collectURLs.processContentTypes = self.allowedContentTypes
      collectURLs.postForms = self.postForms
      collectURLs.urlProcess = self.urlProcess
      collectURLs.urlsXpathList = self.urlsXpathList
      collectURLs.normMask = self.normMask

      self.logger.debug("!!! self.batchItem.urlObj.baseUrl = %s", str(self.batchItem.urlObj.baseUrl))
      self.logger.debug("!!! self.batchItem.urlObj = %s", str(self.batchItem.urlObj))
      if self.batchItem.urlObj is not None:
        self.logger.debug("!!! self.batchItem.urlObj.type = %s", str(self.batchItem.urlObj.type))

      if self.batch.crawlerType != dc_event.Batch.TYPE_REAL_TIME_CRAWLER:
        collectURLs.dbWrapper = self.dbWrapper

      if "ROBOTS_COLLECT" not in self.siteProperties or int(self.siteProperties["ROBOTS_COLLECT"]) > 0:
        collectURLs.robotsParser = self.robotsParser

      self.logger.info("!!!!>> self.curBatchIterations = %s", str(self.curBatchIterations))

      collectURLsResult, internalLinks, externalLinks, urlObjects, self.feedItems, chainUrls, isValidDomStructure = \
                     collectURLs.process(self.crawledResource.http_code, not self.batch.dbMode & dc_event.Batch.DB_MODE_W, self.httpApplyHeaders)

      self.logger.debug("!!! internalLinks (%s): %s", str(len(internalLinks)), str(internalLinks))
      self.logger.debug("!!! externalLinks (%s): %s", str(len(externalLinks)), str(externalLinks))
#       self.logger.debug("!!! self.feedItems (%s): %s", str(len(self.feedItems)), str(self.feedItems))

      # set error mask if DOM is invalid
      if not isValidDomStructure:
        self.errorMask = self.errorMask | APP_CONSTS.ERROR_PARSE_ERROR
        self.logger.debug("Set error_mask = %s", str(self.errorMask))
        self.updateSiteParams(self.errorMask)
        self.updateURLForFailed(self.errorMask, self.crawledResource.http_code, dc_event.URL.STATUS_CRAWLING)

      self.logger.info("!!! TEST CONTENT_TYPE !!!")
      self.logger.info("!!! self.processorName = '%s'", str(self.processorName))
      self.logger.info("!!! crawledResource.content_type = '%s'", str(self.crawledResource.content_type))
      self.logger.info("!!! len(urlObjects) = %s", str(len(urlObjects)))
      self.logger.info("!!! len(self.feedItems) = %s", str(len(self.feedItems)))

      # Check is empty RSS
      if self.url is not None and self.url.parentMd5 is not None and self.url.parentMd5 == "" and self.processorName == PCONSTS.PROCESSOR_RSS and \
        len(urlObjects) == 0 and len(internalLinks) == 0 and len(externalLinks) == 0:
        self.logger.info("RSS IS EMPTY")
        self.errorMask = self.errorMask | APP_CONSTS.ERROR_RSS_EMPTY
        self.logger.debug("Set error_mask = %s", str(self.errorMask))
        self.updateSiteParams(self.errorMask)
        self.updateURLForFailed(self.errorMask, httpCode=self.crawledResource.http_code, status=dc_event.URL.STATUS_CRAWLING)

#       # Check processor name, content-type and number of objects
#       if self.url is not None and self.url.parentMd5 is not None and self.url.parentMd5 == "" and self.processorName == PCONSTS.PROCESSOR_RSS and \
#         self.crawledResource.content_type not in CONSTS.RSS_ALLOWED_CONTENT_TYPES and len(urlObjects) == 0:
#         self.logger.info("RETURN WRONG CONTENT_TYPE FOR RSS AND NOTHING COLLECTED ITEMS")
#
#         self.errorMask = self.errorMask | APP_CONSTS.ERROR_PARSE_ERROR | APP_CONSTS.ERROR_RSS_EMPTY
#         self.logger.debug("Set error_mask = %s", str(self.errorMask))
#         self.updateSiteParams(self.errorMask)
#         self.updateURLForFailed(self.errorMask, httpCode=self.crawledResource.http_code, status=dc_event.URL.STATUS_CRAWLING)

      if self.dom is not None and collectURLsResult and self.collect_additional_prop and (len(internalLinks) > 0 or len(externalLinks)):
        collectPropertiesObj = CollectProperties()
        collectPropertiesObj.siteId = Utils.autoFillSiteId(self.batchItem.siteId, self.logger)
        collectPropertiesObj.kvDbDir = self.kvDbDir
        collectPropertiesObj.res = self.res
        collectPropertiesObj.batchItem = self.batchItem
        collectPropertiesObj.realUrl = self.realUrl
        collectPropertiesObj.process(self.dom, internalLinks, externalLinks)

      # Fill new batch items in self.collectURLsItems
      if self.batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER and\
        urlObjects is not None and len(urlObjects) > 0 and (self.batchItem.depth > 0 or self.processorName == PCONSTS.PROCESSOR_RSS):
        self.fillItemsFromIterations(urlObjects)

      if chainUrls is not None and len(chainUrls) > 0:
        self.fillItemsFromIterationsWithChain(chainUrls, self.batchItem)


  # # writeData  write the response body and headers to file
  #
  # @param batchItem - batch item instance
  # @return - None
  def writeData(self, batchItem):
#     if self.dbWrapper is None:
#       self.logger.info("self.dbWrapper is None")
#       return

    if self.crawledResource is None:
      self.logger.info("self.crawledResource is None")
      return

    self.logger.debug("self.needStoreMime = %s", str(self.needStoreMime))
    # check is need store to disk
    if self.needStoreMime is not None and self.needStoreMime != '*':
      if self.crawledResource.content_type.lower() not in self.needStoreMime:
        self.logger.info("Content not set on disk because a content-type `%s` is not in MIME types list:\n%s",
                         str(self.crawledResource.content_type), str(self.needStoreMime))
        return

    # Create instance CrawlerDataStorage for put contents on FS
    crawlerDataStorage = CrawlerDataStorage(self.dbWrapper, self.logger, self.raw_data_dir, self.keep_old_resources)

    raw_unicode_content = ""
    if self.crawledResource.content_type in self.allowedWriteContentTypes:
      # save UTF-8 encoding text for text content types
      if self.crawledResource.dynamic_fetcher_type:
        raw_unicode_content = self.crawledResource.meta_content
      else:
        raw_unicode_content = self.crawledResource.html_content

      if raw_unicode_content is None:
        raw_unicode_content = ""

      self.logger.debug("self.crawledResource.charset = %s", str(self.crawledResource.charset))
      self.logger.debug("self.crawledResource.etag = '%s'", str(self.crawledResource.etag))
#       self.logger.debug("batchItem: %s", varDump(batchItem, stringifyType=0, strTypeMaxLen=500))
#       self.logger.debug("raw_unicode_content: %s", varDump(raw_unicode_content, stringifyType=0, strTypeMaxLen=500))

      raw_unicode_content = Utils.decodeRawContent(text=raw_unicode_content, charset=self.crawledResource.charset, encoding='utf-8', log=self.logger)

      crawlerDataStorage.addRawContent(batchItem.siteId, batchItem.urlObj.urlMd5, raw_unicode_content)

      self.logger.debug("!!! TIDY !!!")
      self.logger.debug("hasattr(self.crawledResource, 'tidyContent'): %s", str(hasattr(self.crawledResource, 'tidyContent')))

      tidyContent = None
      if hasattr(self.crawledResource, 'tidyContent'):
        tidyContent = getattr(self.crawledResource, 'tidyContent')
      else:
        # Trying recover HTML and fill tidyContent
        ResourceProcess.getDom(siteProperties=self.siteProperties, crawledResource=self.crawledResource, log=self.logger)
        if hasattr(self.crawledResource, 'tidyContent'):
          tidyContent = getattr(self.crawledResource, 'tidyContent')

      if isinstance(tidyContent, basestring):
        crawlerDataStorage.addTidyContent(batchItem.siteId, batchItem.urlObj.urlMd5, tidyContent)
        self.logger.debug("Add 'tidy' content len = %s", str(len(tidyContent)))

    else:
#       # save origin binary data for non-text content types
#       binary_content = '' if self.crawledResource.binary_content is None else self.crawledResource.binary_content
#       crawlerDataStorage.addRawContent(batchItem.siteId, batchItem.urlObj.urlMd5, binary_content)

      # save empty raw content for not allowed content-types
      crawlerDataStorage.addRawContent(batchItem.siteId, batchItem.urlObj.urlMd5, "")

    # save rendered file
    if self.crawledResource.dynamic_fetcher_type and self.crawledResource.html_content:
      crawlerDataStorage.addDynamicContent(batchItem.siteId, batchItem.urlObj.urlMd5, self.crawledResource.html_content)

    # html header
    if self.store_http_headers and self.crawledResource.response_header:
      crawlerDataStorage.addHeaderContent(batchItem.siteId, batchItem.urlObj.urlMd5, self.crawledResource.response_header)

    self.logger.debug("!!! Store HTTP request !!!")
    self.logger.debug("!!! self.storeHttpRequest = %s", str(self.storeHttpRequest))
    self.logger.debug("!!! self.crawledResource.html_request:\n%s", varDump(self.crawledResource.html_request, stringifyType=0))
    # html request
    if self.storeHttpRequest and self.crawledResource.html_request:
      self.logger.debug("Add HTTP request content...")
      crawlerDataStorage.addRequestContent(batchItem.siteId, batchItem.urlObj.urlMd5, self.crawledResource.html_request)

    self.logger.debug("!!! self.storeCookies = %s", str(self.storeCookies))
    # store cookies
    if self.storeCookies:
      cookiesStr = ''
      if self.res is not None and RequestsRedirectWrapper.RESPONSE_COOKIE_HEADER_NAME in self.res.headers:
        # set raw cookies string from header
        cookiesStr = self.res.headers[RequestsRedirectWrapper.RESPONSE_COOKIE_HEADER_NAME]
      elif isinstance(self.crawledResource.cookies, dict):
        # set cookies from cookies dictionary
        cookiesStr = ''.join([key + ": " + value + "; " for (key, value) in self.crawledResource.cookies.items()])

      if cookiesStr != "":
        self.logger.debug("Try to save the cookies string: %s", str(cookiesStr))
        crawlerDataStorage.addCookiesContent(batchItem.siteId, batchItem.urlObj.urlMd5, cookiesStr)

    # Put contents on FS
    crawlerDataStorage.putContents()

    # change URL's contentMask
    self.url.contentMask = dc_event.URL.CONTENT_STORED_ON_DISK
    batchItem.urlObj.contentMask = dc_event.URL.CONTENT_STORED_ON_DISK


  # #getDir prepare directory
  #
  # @param
  def getDir(self):
    if len(self.batchItem.siteId):
      self.dir = os.path.join(self.raw_data_dir, self.batchItem.siteId, PathMaker(self.batchItem.urlId).getDir())
    else:
      self.dir = os.path.join(self.raw_data_dir, "0", PathMaker(self.batchItem.urlId).getDir())


  # #makeDir prepare dir
  #
  # @param
  def makeDir(self):
    self.logger.debug('!!! makeDir() enter .... self.dir = ' + str(self.dir))
    if not os.path.exists(self.dir):
      os.makedirs(self.dir)

    if not os.path.isdir(self.dir):
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_WRITE_FILE_ERROR
      self.updateURLForFailed(APP_CONSTS.ERROR_WRITE_FILE_ERROR)
      raise Exception("path %s exists, but is not a directory" % (self.dir,))


  # #updateURLForFailed
  #
  # @param error_bit BitMask of error
  def updateURLForFailed(self, errorBit, httpCode=CONSTS.HTTP_CODE_400, status=dc_event.URL.STATUS_CRAWLED, updateUdate=True):

    contentType = None
    if self.crawledResource is not None:
      contentType = self.crawledResource.content_type
      self.batchItem.urlObj.eTag = self.crawledResource.etag

    self.urlProcess.siteId = self.batchItem.siteId
    self.urlProcess.urlObj = self.batchItem.urlObj
    self.urlProcess.updateURLForFailed(errorBit, self.batchItem, httpCode, status, updateUdate, contentType)

    self.logger.debug("!!! self.crawledResource: %s", varDump(self.crawledResource))
    if self.crawledResource is not None:
      self.crawledResource.http_code = httpCode
      # save raw data on local file system
      self.getDir()
      if hasattr(self.url, 'urlOld') and self.url.parentMd5 != "":
        self.writeData(self.updateResolvedURL(siteProperties=self.siteProperties, batchItem=self.batchItem, urlOld=self.url.urlOld, urlNew=self.url.url))
      else:
        self.writeData(self.batchItem)


  # #httpRequestWrapper method makes http request, using detectModified object if it present
  #
  # @param url - http requests url
  # @param headers - http headers, using in http request
  # @param urlObj - url's Obj
  # @param incomingContent - raw content that is externally provided
  # @param macroCode - list of the strings with the source code to be executed at the dynamic fetcher,
  # for example JavaScript
  # @return http resource (http response)
  def httpRequestWrapper(self, url, headers, urlObj, incomingContent, macroCode=None):
    ret = None
    # (1) Fetcher type detection section
    localFetchType = copy.deepcopy(self.site.fetchType)
    if localFetchType == BaseFetcher.TYP_DYNAMIC:
      if urlObj is not None and urlObj.parentMd5 is not None and \
      "DYNAMIC_FETCH_ONLY_FOR_ROOT_URL" in self.siteProperties and \
      int(self.siteProperties["DYNAMIC_FETCH_ONLY_FOR_ROOT_URL"]) > 0:
        if urlObj.parentMd5 == "":
          localFetchType = BaseFetcher.TYP_DYNAMIC
        else:
          localFetchType = BaseFetcher.TYP_NORMAL
    elif localFetchType == BaseFetcher.TYP_AUTO:
      localFetchType = BaseFetcher.TYP_NORMAL
      if self.curBatchIterations == 2 and self.processorName == PCONSTS.PROCESSOR_RSS:
        localFetchType = BaseFetcher.TYP_DYNAMIC

    if urlObj is not None and urlObj.parentMd5 is not None and urlObj.parentMd5 == "" and \
    self.processorName == PCONSTS.PROCESSOR_RSS and localFetchType == BaseFetcher.TYP_DYNAMIC:
      localFetchType = BaseFetcher.TYP_NORMAL

    tmpDirOptions = None
    if urlObj is not None and APP_CONSTS.FETCHER_TYPE in self.siteProperties and self.siteProperties[APP_CONSTS.FETCHER_TYPE] is not None:
      self.logger.debug("!!! %s !!!", str(APP_CONSTS.FETCHER_TYPE))
      fetcherType = FetcherType(self.siteProperties[APP_CONSTS.FETCHER_TYPE], self.logger)
      options = fetcherType.getFetcherTypeOptions(urlObj.url)
      if options is not None:
        if options.type is not None:
          localFetchType = options.type

        if options.options is not None:
          tmpDirOptions = options.options

    # (1) END
    self.logger.debug("FetchType = %s, tmpDirOptions: %s", str(localFetchType), str(tmpDirOptions))

    if urlObj is not None:
      if incomingContent is None:
        externalUrlFilterPassed = True
        # Check filter for external URL
        if Filters(None, self.dbWrapper, self.batchItem.siteId, 0, None, Filters.OC_RE, Filters.STAGE_COLLECT_URLS,
                   Filters.SELECT_SUBJECT_EXTERNAL_URL).isExistStage(Filters.STAGE_COLLECT_URLS):
          self.logger.debug("Found filter for check external URL on stage 'COLLECT_URLS'")
          # Apply filter for URL content-type

          if CollectURLs.filtersApply(inputFilters=self.site.filters, subject=url, depth=self.url.depth, wrapper=self.dbWrapper, siteId=self.batchItem.siteId,
                                      stage=Filters.STAGE_COLLECT_URLS, selectSubject=Filters.SELECT_SUBJECT_EXTERNAL_URL, defaultValue=True,
                                      currentBatchIterations=self.curBatchIterations,
                                      siteProperties=self.siteProperties, url=url):
            self.logger.debug("Check pattern for external URL - SUCCESS. Url: %s", str(url))
          else:
            self.logger.debug("Check pattern for external URL - FAILED. Url: %s", str(url))
            externalUrlFilterPassed = False

        # make request URL
        requestURL = url
        if externalUrlFilterPassed and self.external_url is not None and isinstance(self.external_url, basestring):
          self.logger.debug("external_url: %s, url: %s", str(self.external_url), str(url))
          if '%URL%' in self.external_url:
            requestURL = self.external_url.replace('%URL%', str(url))  # ##urllib.quote(url))
            self.logger.debug("New external URL: %s", str(requestURL))

        cookiesStage = HTTPCookieResolver.STAGE_REGULAR
        if self.processorName == PCONSTS.PROCESSOR_RSS:
          cookiesStage = HTTPCookieResolver.STAGE_RSS

        self.logger.debug("!!! Before fetcher.open() for URL: %s", str(requestURL))
        self.logger.debug("!!! self.site.id = '%s'", str(self.site.id))
        self.logger.debug("!!! self.batchItem.siteId = '%s'", str(self.batchItem.siteId))
#         self.logger.debug("!!! self.siteProperties:\n%s", varDump(self.siteProperties, stringifyType=0))

        kwarg = {}
        if self.headerFileDir != "":
          kwarg[HTTPHeadersExtractor.ARGUMENT_STORAGE_DIR] = self.headerFileDir
        kwarg[DetectModified.ARGUMENT_URL_OBJECT ] = urlObj

        ret = HTTPRedirectExtractor.fetchUrl(siteProperties=self.siteProperties,
                                             siteId=self.batchItem.siteId,
                                             url=requestURL,
                                             headers=headers,
                                             httpTimeout=self.url.httpTimeout,
                                             allowRedirects=self.allow_http_redirects,
                                             proxies=None,
                                             allowedContentTypes=self.allowedContentTypes,
                                             maxResourceSize=self.site.maxResourceSize,
                                             maxRedirects=CONSTS.MAX_HTTP_REDIRECTS_LIMIT,
                                             httpMethod=self.url.httpMethod,
                                             fetchType=localFetchType,
                                             dbWrapper=self.dbWrapper,
                                             filters=None if urlObj.parentMd5 == "" else self.site.filters,
                                             macro=macroCode,
                                             log=self.logger,
                                             stage=HTTPRedirectResolver.RedirectProperty.VALUE_STAGE_CRAWLER,
                                             cookiesStage=cookiesStage,
                                             useDetectModified=True,
                                             **kwarg)

      else:
        fetcher = BaseFetcher.get_fetcher(BaseFetcher.TYP_CONTENT, None, self.dbWrapper, self.batchItem.siteId, tmpDirOptions)
        ret = fetcher.open(url, inputContent=incomingContent, log=self.logger)

    return ret


  # # replace macro name for macro data
  #
  # @param macroDict - macro dictionary
  # @param macroName - macro name
  # @param macroData - macro data
  # @return macro dictionary
  def __replaceMacroName(self, macroDict, macroName, macroData):

    if isinstance(macroDict, basestring):
      macroDict = re.sub(pattern=macroName, repl=macroData, string=macroDict, flags=re.U + re.I)

    elif isinstance(macroDict, list):
      for index, value in enumerate(macroDict):
        macroDict[index] = self.__replaceMacroName(value, macroName, macroData)

    elif isinstance(macroDict, dict):
      for key, value in macroDict.items():
        macroDict[key] = self.__replaceMacroName(value, macroName, macroData)

        if key == macroName:
          macroDict[macroData] = value
          del macroDict[key]

    return macroDict


  # #crawl site
  #
  # @return should continue to write data and collect URLs
  def crawl(self, incomingContent, startTime):
    self.crawledResource = None
    # delay
    self.logger.debug("Make request delay " + str(self.url.requestDelay / 1000.0) + " sec.")
    time.sleep(self.url.requestDelay / 1000.0)
    # (1) Decode url - [UrlProcess]
    # TODO move to - [UrlProcess]
#     self.logger.debug("!!! self.url.url = '%s'", str(self.url.url))
#     self.logger.debug("!!! self.siteProperties:\n%s", varDump(self.siteProperties, stringifyType=0))

    self.urlProcess.url = self.url.url
    self.urlProcess.site = self.site
    # (1) END
    url = self.urlProcess.getRealUrl()
    self.realUrl = url

    self.logger.debug("Crawling URL: '%s'", str(url))

    macro = None
    if APP_CONSTS.FETCHER_MACRO in self.siteProperties and self.siteProperties[APP_CONSTS.FETCHER_MACRO] is not None\
      and self.siteProperties[APP_CONSTS.FETCHER_MACRO] != '':
      try:
        macro = Utils.jsonLoadsSafe(self.siteProperties[APP_CONSTS.FETCHER_MACRO], self.siteProperties[APP_CONSTS.FETCHER_MACRO], self.logger)
        macro = self.__replaceMacroName(macroDict=macro, macroName='%URL%', macroData='"' + url + '"')

      except Exception, err:
        self.logger.error("Initialization of macro error: %s, source: %s", str(err), str(self.siteProperties[APP_CONSTS.FETCHER_MACRO]))
        self.errorMask = self.errorMask | APP_CONSTS.ERROR_MACRO_DESERIALIZATION
        self.updateSiteParams(APP_CONSTS.ERROR_MACRO_DESERIALIZATION)
        self.updateURLForFailed(self.errorMask)
        return False

    try:
      # (3) Resolve HTTP Method (move convertToHttpDateFmt to Utils) - [UrlProcess]
      # TODO move to - [UrlProcess]
      self.urlProcess.urlObj = self.url
#       postData = self.urlProcess.resolveHTTP(self.postForms, self.httpApplyHeaders)
      # (3) END

      # Workaround
      url = self.urlProcess.urlTemplateApply(url, self.batch.crawlerType, self.urlTempalteRegular,
                                             self.urlTempalteRealtime, self.urlTempalteRegularEncode,
                                             self.urlTempalteRealtimeEncode)

      self.logger.debug("!!! urlTemplateApply() return url = '%s'", str(url))

# TODO
#       # Check robots mode use proxy if allowed
#       if "ROBOTS_MODE" not in self.siteProperties or int(self.siteProperties["ROBOTS_MODE"]) > 0:
#         self.logger.debug("Robots.txt obey mode ON")
#
#         # get proxy name
#         proxyName, isValid = HTTPProxyResolver.getProxyName(siteProperties=self.siteProperties,
#                                                             siteId=self.site.id,
#                                                             url=self.url.url,
#                                                             dbWrapper=self.dbWrapper,
#                                                             log=self.logger)
#         if not isValid and proxyName is not None:
#           self.logger.debug("Try usage invalid proxy '%s'. Skipped...", str(proxyName))
#           self.errorMask = self.errorMask | APP_CONSTS.ERROR_NOT_EXIST_ANY_VALID_PROXY
#           self.updateSiteParams(APP_CONSTS.ERROR_NOT_EXIST_ANY_VALID_PROXY)
#           self.updateURLForFailed(self.errorMask)
#           return False
#
#         if self.robotsParser.loadRobots(url, self.batchItem.siteId, self.httpApplyHeaders, proxyName):
#           self.httpApplyHeaders = self.updateHeadersByCookies(self.httpApplyHeaders,
#                                                               url,
#                                                               HTTPCookieResolver.STAGE_ROBOTS,
#                                                               self.site.id)
#
#           isAllowed, retUserAgent = self.robotsParser.checkUrlByRobots(url, self.batchItem.siteId,
#                                                                        self.httpApplyHeaders)
#           if not isAllowed:
#             self.logger.debug(">>> URL " + url + " is NOT Allowed by user-agent:" + str(retUserAgent))
#             self.errorMask = self.errorMask | APP_CONSTS.ERROR_ROBOTS_NOT_ALLOW
#             self.updateSiteParams(APP_CONSTS.ERROR_ROBOTS_NOT_ALLOW)
#             self.updateURLForFailed(self.errorMask)
#             return False

      # (3.2) HTTP Fetcher with html redirect resolving
      localUrl = url
#       prevUrl = localUrl
      res = self.makeDefaultResponse(Response())
      self.resourceProcess.urlObj = self.url
      self.crawledResource = self.resourceProcess.generateResource(startTime, res, self.httpApplyHeaders, time.time(), self.defaultIcrCrawlTime)

      self.logger.info("start to fetch: %s", str(localUrl))
      res = self.httpRequestWrapper(localUrl, self.httpApplyHeaders, self.url, incomingContent, macro)
      self.logger.info("fetch finished for %s", str(localUrl))

      if res is None:
        self.logger.info("response from fetch is None!")
        return False

      if self.external_url is None and res is not None and res.url is not None:
        self.logger.debug("!!! Check old and new URLs !!!")
        # URL normalization
        self.logger.debug("Before URL normalization url: %s, type = %s", str(res.url), str(type(res.url)))
        normalizedUrl = UrlNormalize.execute(siteProperties=self.siteProperties, base=res.url, url=res.url, supportProtocols=None, log=self.logger)
        self.logger.debug(" After URL normalization url: %s", str(normalizedUrl))
        # set normalized URL value
        if normalizedUrl is not None:
          self.url.url = normalizedUrl
          res.url = normalizedUrl

        if localUrl == res.url:
          self.logger.debug("!!! Old and new URLs are EQUAL !!!")
        else:
          self.logger.debug("!!! Old and new URLs are NOT EQUAL !!!")
          self.logger.debug("!!! localUrl: %s", str(localUrl))
          self.logger.debug("!!! res.url: %s", str(res.url))

        # save old value anyway
        self.url.urlOld = localUrl

      # save res to self, will use it in next steps
      # (4) Create and fill resource structure - [ResourceProcess]
      self.res = res
      if self.res is not None:
        self.logger.info("!!! response status_code: '%s'", str(self.res.status_code))
        self.logger.info("!!! response cookies: '%s'", str(self.res.cookies))
        self.logger.info("!!! response error_mask: '%s'", str(self.res.error_mask))
        self.logger.info("!!! response error_msg: '%s'", str(self.res.error_msg))
#         if self.res.rendered_unicode_content is not None:
#           self.logger.info("!!! response:\n%s", varDump(self.res, stringifyType=1))

      # remove NoScript from raw content for dynamic fetcher
      if self.site.fetchType == BaseFetcher.TYP_DYNAMIC:
        self.res.rendered_unicode_content = Utils.eraseNoScript(self.res.rendered_unicode_content)
      # re-calculate content size
      if self.res.rendered_unicode_content is not None:
        self.res.content_size = len(self.res.rendered_unicode_content)
        # save raw content
# #         if self.dbWrapper is None:
# #           self.batchItem.urlContentResponse = dc_event.URLContentResponse(url=self.batchItem.urlObj.url, \
# #             rawContents=[dc_event.Content(contentBuffer=base64.b64encode(self.res.rendered_unicode_content))])

      # use charset to improve encoding detect
      self.resourceProcess.urlObj = self.url
      resource = self.resourceProcess.generateResource(startTime, self.res, self.httpApplyHeaders, time.time(), self.defaultIcrCrawlTime,
                                                       self.siteProperties["CONTENT_TYPE_MAP"] if "CONTENT_TYPE_MAP" in self.siteProperties else None)

      resource.dynamic_fetcher_type = res.dynamic_fetcher_type
      resource.dynamic_fetcher_result_type = res.dynamic_fetcher_result_type
      self.crawledResource = resource

      # save raw content
      if self.dbWrapper is None and self.res.rendered_unicode_content is not None:
        rendered_unicode_content = Utils.decodeRawContent(text=self.res.rendered_unicode_content, charset=resource.charset, encoding='utf-8', log=self.logger)

        self.batchItem.urlContentResponse = dc_event.URLContentResponse(url=self.batchItem.urlObj.url, \
          rawContents=[dc_event.Content(contentBuffer=base64.b64encode(rendered_unicode_content))])

      if self.dbWrapper is None and isinstance(self.crawledResource.html_request, basestring):
        self.logger.info("!!! self.crawledResource.html_request: '%s'", str(self.crawledResource.html_request))
        self.batchItem.urlContentResponse.requests.append(dc_event.Content(base64.b64encode(self.crawledResource.html_request),
                                                                           typeId=dc_event.Content.CONTENT_REQUESTS_CONTENT))

      # (4) END

      # (5) Checks error mask
      if self.res is not None and self.res.error_mask != APP_CONSTS.ERROR_OK:
        self.logger.debug("Resource has status_code = %s, error_mask = %s, error_msg: %s", str(self.res.status_code), str(self.res.error_mask),
                          str(self.res.error_msg))
        self.errorMask = self.errorMask | self.res.error_mask
        self.updateURLForFailed(self.errorMask, self.res.status_code)
        return False
      # (5) END

      if self.res is not None and self.res.headers is not None and "content-length" in self.res.headers and \
        self.res.headers["content-length"] == EMPTY_RESPONSE_SIZE:
        self.logger.debug('Zero content-length!')
        self.errorMask = self.errorMask | APP_CONSTS.ERROR_EMPTY_RESPONSE
        self.updateURLForFailed(self.errorMask)
        return False

      self.logger.debug("!!! self.crawledResource.http_code = %s", str(self.crawledResource.http_code))
      if self.crawledResource.http_code >= CONSTS.HTTP_CODE_400:
        self.errorMask = self.errorMask | APP_CONSTS.ERROR_HTTP_ERROR
        # Add error mask about forbidden fetch
        if self.crawledResource.http_code == CONSTS.HTTP_CODE_403:
          self.errorMask = APP_CONSTS.ERROR_FETCH_FORBIDDEN

        isOk = self.updateSiteParams(self.errorMask)
        self.logger.debug("!!! updateSiteParams by error mask = '%s' is OK = %s", str(self.errorMask), str(isOk))

        self.updateURLForFailed(self.errorMask, self.crawledResource.http_code)
        return False

      # (8) Check Content size and response code - [ResourceProcess]
      # TODO move to - [ResourceProcess]
      self.resourceProcess.resource = resource
      self.resourceProcess.batchItem = self.batchItem
      if not self.resourceProcess.checkResourcesResponse(res, self.site.maxResourceSize, self.updateSiteParams):
        self.errorMask = self.errorMask | resource.error_mask
        self.updateURLForFailed(self.errorMask, res.status_code)
        return False
      # (8) END

      # # Block handlers 'STAGE_BEFORE_DOM_PRE'
      self.logger.debug("+++++++++++++++++++++++++++++++++++++")
      self.logger.debug("Block handlers 'STAGE_BEFORE_DOM_PRE'")
      collectURLs = CollectURLs()
      self.logger.debug("self.site.filters: " + varDump(self.site.filters))
      self.logger.debug("self.batchItem.urlObj.parentMd5: %s", str(self.batchItem.urlObj.parentMd5))
      if self.batchItem.urlObj.parentMd5 == "":
        self.logger.debug("It's root url. Skipped block handlers 'STAGE_BEFORE_DOM_PRE'")
        return True

      # Create class Filters instance for check 'raw content' use regular expressions
      localFilters = Filters(None, self.dbWrapper, self.batchItem.siteId, 0,
                             None, Filters.OC_RE, Filters.STAGE_BEFORE_DOM_PRE, Filters.SELECT_SUBJECT_RAW_CONTENT)

      self.logger.debug('>>> localFilters.filters: ' + varDump(localFilters.filters))
      self.logger.debug(">>> isExistStage('STAGE_BEFORE_DOM_PRE'): " + \
                        str(localFilters.isExistStage(Filters.STAGE_BEFORE_DOM_PRE)))
      # (9) Check RAW content text regular expression
      if localFilters.isExistStage(Filters.STAGE_BEFORE_DOM_PRE):
        self.logger.debug("Check RAW content text regular expression ...")
        if collectURLs.filtersApply(None, resource.binary_content, 0, self.dbWrapper,
                                    self.batchItem.siteId, None, Filters.OC_RE,
                                    Filters.STAGE_BEFORE_DOM_PRE, Filters.SELECT_SUBJECT_RAW_CONTENT, True,
                                    siteProperties=self.siteProperties, url=self.res.url):
          self.logger.debug("RAW content text regular expression check SUCCESS")
        else:
          self.logger.debug("RAW content text regular expression check FAILED")
          self.errorMask = self.errorMask | APP_CONSTS.ERROR_CRAWLER_FILTERS_BREAK
          self.updateURLForFailed(self.errorMask)
          return False
      # (9) END

      # Create class Filters instance for check 'headers' use regular expressions
      localFilters = Filters(None, self.dbWrapper, self.batchItem.siteId, 0,
                             None, Filters.OC_RE, Filters.STAGE_BEFORE_DOM_PRE, Filters.SELECT_SUBJECT_HEADERS_ALL)

      self.logger.debug('>>> localFilters.filters: ' + varDump(localFilters.filters))
      self.logger.debug(">>> isExistStage('STAGE_BEFORE_DOM_PRE'): " + \
                        str(localFilters.isExistStage(Filters.STAGE_BEFORE_DOM_PRE)))
      # (10) Check HTTP headers by name text regular expression check
      if localFilters.isExistStage(Filters.STAGE_BEFORE_DOM_PRE):
        self.logger.debug("Check HTTP headers by name text regular expression check ...")
        if collectURLs.filtersApply(None, resource.response_header, 0, self.dbWrapper,
                                    self.batchItem.siteId, None, Filters.OC_RE,
                                    Filters.STAGE_BEFORE_DOM_PRE, Filters.SELECT_SUBJECT_HEADERS_ALL, True,
                                    siteProperties=self.siteProperties, url=self.res.url):
          self.logger.debug("HTTP headers by name text regular expression check SUCCESS")
        else:
          self.logger.debug("HTTP headers by name text regular expression check FAILED")
          self.errorMask = self.errorMask | APP_CONSTS.ERROR_CRAWLER_FILTERS_BREAK
          self.updateURLForFailed(self.errorMask)
          return False
      # (10) END

      # (11) Check Last modified datetime value date comparison check
      self.logger.debug("Check Last modified datetime value date comparison check ...")
      self.logger.debug('resource.last_modified = ' + str(resource.last_modified))

      localFilters = Filters(None, self.dbWrapper, self.batchItem.siteId, 0,
                             {'PDATE':str(resource.last_modified)}, Filters.OC_SQLE, Filters.STAGE_BEFORE_DOM_PRE,
                             Filters.SELECT_SUBJECT_LAST_MODIFIED)

      self.logger.debug('>>> localFilters.filters: ' + varDump(localFilters.filters))
      self.logger.debug(">>> isExistStage('STAGE_BEFORE_DOM_PRE'): " + \
                        str(localFilters.isExistStage(Filters.STAGE_BEFORE_DOM_PRE)))

      if localFilters.isExistStage(Filters.STAGE_BEFORE_DOM_PRE):
        if collectURLs.filtersApply(None, '', 0, self.dbWrapper, self.batchItem.siteId,
                                    {'PDATE':str(resource.last_modified)}, Filters.OC_SQLE,
                                    Filters.STAGE_BEFORE_DOM_PRE, Filters.SELECT_SUBJECT_LAST_MODIFIED, True,
                                    siteProperties=self.siteProperties, url=self.res.url):
          self.logger.debug("Last modified datetime value date comparison check SUCCESS")
        else:
          self.logger.debug("Last modified datetime value date comparison check FAILED")
          self.errorMask = self.errorMask | APP_CONSTS.ERROR_CRAWLER_FILTERS_BREAK
          self.updateURLForFailed(self.errorMask)
          return False
      # (11) END

    except TimeoutException, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_REQUEST_TIMEOUT
      self.updateURLForFailed(APP_CONSTS.ERROR_REQUEST_TIMEOUT)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except requests.exceptions.InvalidURL:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_BAD_URL
      self.updateURLForFailed(APP_CONSTS.ERROR_BAD_URL)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except requests.exceptions.ChunkedEncodingError:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_PAGE_CONVERT_ERROR
      self.updateURLForFailed(APP_CONSTS.ERROR_PAGE_CONVERT_ERROR)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except ConnectionErrorException, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_CONNECTION_ERROR
      self.updateURLForFailed(APP_CONSTS.ERROR_CONNECTION_ERROR)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except requests.exceptions.ContentDecodingError:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_PAGE_CONVERT_ERROR
      self.updateURLForFailed(APP_CONSTS.ERROR_PAGE_CONVERT_ERROR)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except lxml.etree.XMLSyntaxError:  # pylint: disable=E1101
      self.logger.debug("XML HTML syntax error")
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_DTD_INVALID
      self.updateURLForFailed(APP_CONSTS.ERROR_DTD_INVALID)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except BadContentTypeException, err:
      self.logger.debug("Bad content-type error : %s", str(err))
      self.errorMask = self.errorMask | err.code
      self.crawledResource = self.resourceProcess.generateResource(startTime, err.response, self.httpApplyHeaders, time.time(), self.defaultIcrCrawlTime)
      self.updateURLForFailed(self.errorMask, httpCode=err.response.status_code)
      self.res = self.crawledResource
      return False
    except ProxyException, err:
      self.logger.debug('self.errorMask = ' + str(self.errorMask) + ' err.code = ' + str(err.code) + ' err.statusUpdate = ' + str(err.statusUpdate))
      status = dc_event.URL.STATUS_CRAWLED
      if err.statusUpdate is not None:
        status = err.statusUpdate
      self.logger.debug('Set status update = ' + str(status))
      self.errorMask = self.errorMask | err.code
      self.updateURLForFailed(self.errorMask, SeleniumFetcher.ERROR_PROXY_CONNECTION_FAILED, status)
      self.res = self.makeDefaultResponse(self.res, SeleniumFetcher.ERROR_PROXY_CONNECTION_FAILED)
      return False
    except SeleniumFetcherException, err:
      self.logger.error("Selenium fetcher error: " + str(err) + ' code = ' + str(err.code))
      httpCode = CONSTS.HTTP_CODE_400
      if err.code in self.errorMaskHttpCodeDict:
        httpCode = self.errorMaskHttpCodeDict[err.code]
      self.errorMask = self.errorMask | err.code
      self.updateURLForFailed(self.errorMask, httpCode)
      self.res = self.makeDefaultResponse(self.res, httpCode)
      return False
    except UrlAvailableException, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_CONNECTION_ERROR
      self.updateURLForFailed(APP_CONSTS.ERROR_CONNECTION_ERROR)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except HTTPErrorException, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_FETCH_HTTP_ERROR
      self.updateURLForFailed(APP_CONSTS.ERROR_FETCH_HTTP_ERROR)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except URLRequiredException, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_FETCH_INVALID_URL
      self.updateURLForFailed(APP_CONSTS.ERROR_FETCH_INVALID_URL)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except requests.exceptions.RequestException, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_FETCH_AMBIGUOUS_REQUEST
      self.updateURLForFailed(APP_CONSTS.ERROR_FETCH_AMBIGUOUS_REQUEST)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except NotModifiedException, err:
      self.logger.debug('=' * 30)
      self.logger.debug("%s siteId = %s, url = %s", str(err), str(self.batchItem.siteId), str(self.realUrl))
      self.logger.debug("urlMd5 = %s, url = %s", str(self.batchItem.urlObj.urlMd5), str(self.batchItem.urlObj.url))
      httpCode, status, updateUDate = DetectModified.extractResult(self.batchItem.siteId, self.batchItem.urlObj.urlMd5, self.dbWrapper, err.detectModified,
                                                                   err.response)

      self.crawledResource = self.resourceProcess.generateResource(startTime, err.response, self.httpApplyHeaders, time.time(), self.defaultIcrCrawlTime)
      self.updateURLForFailed(self.errorMask, httpCode, status, updateUDate)
      self.res = self.crawledResource
      return False
    except DatabaseException, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_DATABASE_ERROR
      self.updateURLForFailed(APP_CONSTS.ERROR_DATABASE_ERROR)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except InternalCrawlerException, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_FETCHER_INTERNAL
      self.updateURLForFailed(APP_CONSTS.ERROR_FETCHER_INTERNAL)
      self.res = self.makeDefaultResponse(self.res)
      return False
    except Exception, err:
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_GENERAL_CRAWLER
      self.updateURLForFailed(APP_CONSTS.ERROR_GENERAL_CRAWLER)
      ExceptionLog.handler(self.logger, err, "Crawler fatal error.", (err), {ExceptionLog.LEVEL_NAME_ERROR:ExceptionLog.LEVEL_VALUE_DEBUG})
      return False

    return True


  # # Make default response
  #
  # @param response - instance of Response
  # @param httpCode - HTTP code
  # @return response - updated instance of Response
  def makeDefaultResponse(self, response, httpCode=CONSTS.HTTP_CODE_400):

    if response is None:
      response = Response()
    # set default values for response
    response.status_code = httpCode
    response.unicode_content = ""
    response.str_content = ""
    response.rendered_unicode_content = ""
    response.content_size = 0
    response.encoding = ""
    response.headers = {"content-length": 0, "content-type": ""}
    response.meta_res = ""

    return response


  # updateAVGSpeed update sites.AVGSpeed property
  def updateAVGSpeed(self):
    if self.res.status_code == CONSTS.HTTP_CODE_304:
      # not modified will return empty response body
      return

    avgSpeed = (self.site.avgSpeed * self.site.avgSpeedCounter + self.crawledResource.bps)\
               / (self.site.avgSpeedCounter + 1)
    self.site.avgSpeed = avgSpeed
    self.site.avgSpeedCounter += 1

    if self.dbWrapper is not None:
      localSiteUpdate = dc_event.SiteUpdate(self.batchItem.siteId)
      for attr in localSiteUpdate.__dict__:
        if hasattr(localSiteUpdate, attr):
          setattr(localSiteUpdate, attr, None)
      localSiteUpdate.id = self.batchItem.siteId
      localSiteUpdate.avgSpeed = avgSpeed
      localSiteUpdate.avgSpeedCounter = SQLExpression("`AVGSpeedCounter` + 1")
      self.dbWrapper.siteNewOrUpdate(localSiteUpdate)


  # #prepare request headers and cookie
  #
  # @param site - site object
  # TODO move to - [SiteProcess]
  def readSiteProperties(self):
    # TODO not sure is it right that fetch headers and cookie
    # by `Name` = 'headers'/'cookie'
    self.storeHttpRequest = True
    self.store_http_headers = True
#     self.headersDict = {}
    self.postForms = {}
#     self.headers = None
    self.external_url = None
    self.auto_remove_props = {}
    self.allowedContentTypes = []
    self.allowedWriteContentTypes = []

    try:
#       self.logger.info('#' * 50)
#       self.logger.info("\n       self.batchItem.siteId = '%s',\n                self.site.id = '%s',\nself.batchItem.urlObj.siteId = '%s'",
#                        str(self.batchItem.siteId), str(self.site.id), str(self.batchItem.urlObj.siteId))
#
#       self.logger.debug("self.site.properties:\n%s", varDump(self.site.properties, stringifyType=0))
#       self.logger.debug("self.batchItem.properties:\n%s", varDump(self.batchItem.properties, stringifyType=0))

      # Update site properties from batch item properties
      keys = [localProperty["name"] for localProperty in self.site.properties]
      # self.logger.debug("keys: %s" % str(keys))
      for key in self.batchItem.properties.keys():
        if key in keys:
          for localProperty in self.site.properties:
            if localProperty["name"] == key:
              self.logger.debug("%s present in site properties. Rewrite localProperty", str(key))
              localProperty["value"] = self.batchItem.properties[key]
        else:
          self.logger.debug("%s not present in site properties. Add localProperty", str(key))
          self.site.properties.append({"name":key, "value":self.batchItem.properties[key], "URLMd5":self.batchItem.urlId, "siteId":self.batchItem.siteId})

#       self.logger.debug("Updated site's properties:\n%s", varDump(self.site.properties, stringifyType=0))

      self.siteProperties = {}
      for item in self.site.properties:
        self.siteProperties[item["name"]] = item["value"]

      if len(self.siteProperties) == 0:
        raise Exception("Empty site properties. Abort process batch item for siteId = %s" % str(self.batchItem.siteId))

      if 'STORE_HTTP_REQUEST' in self.siteProperties:
        self.storeHttpRequest = int(self.siteProperties['STORE_HTTP_REQUEST']) > 0
      if 'STORE_HTTP_HEADERS' in self.siteProperties:
        self.store_http_headers = int(self.siteProperties['STORE_HTTP_HEADERS']) > 0
      if 'MIME_TYPE_STORE_ON_DISK' in self.siteProperties:
        allowMimes = self.siteProperties['MIME_TYPE_STORE_ON_DISK']
        if allowMimes is not None and allowMimes != '' and allowMimes != '*':
          self.needStoreMime = set([mime.lower() for mime in allowMimes.split(',')])
      for key in self.siteProperties.keys():
        if key.startswith('HTTP_POST_FORM_'):
          self.postForms[key[15:]] = self.siteProperties[key]
      if 'EXTERNAL_URL' in self.siteProperties:
        self.external_url = self.siteProperties['EXTERNAL_URL']  # pylint: disable=W0201

      # set all allowed content-types
      self.allowedContentTypes = CONSTS.ALL_ALLOWED_CONTENT_TYPE
      if APP_CONSTS.ALLOWED_CTYPES in self.siteProperties:
        self.allowedContentTypes = self.siteProperties[APP_CONSTS.ALLOWED_CTYPES].lower().split(',')
      self.allowedWriteContentTypes = self.allowedContentTypes

      if 'PROCESSOR_NAME' in self.siteProperties:
        self.processorName = self.siteProperties['PROCESSOR_NAME']
      if DC_CONSTS.SITE_PROP_SAVE_COOKIES in self.siteProperties:
        self.storeCookies = int(self.siteProperties[DC_CONSTS.SITE_PROP_SAVE_COOKIES]) > 0
      if DC_CONSTS.SITE_PROP_AUTO_REMOVE_RESOURCES in self.siteProperties:
        self.auto_remove_props[DC_CONSTS.SITE_PROP_AUTO_REMOVE_RESOURCES] = \
        self.siteProperties[DC_CONSTS.SITE_PROP_AUTO_REMOVE_RESOURCES]
      if DC_CONSTS.SITE_PROP_AUTO_REMOVE_ORDER in self.siteProperties:
        self.auto_remove_props[DC_CONSTS.SITE_PROP_AUTO_REMOVE_ORDER] = \
        self.siteProperties[DC_CONSTS.SITE_PROP_AUTO_REMOVE_ORDER]
      if DC_CONSTS.SITE_PROP_AUTO_REMOVE_WHERE in self.siteProperties:
        self.auto_remove_props[DC_CONSTS.SITE_PROP_AUTO_REMOVE_WHERE] = \
        self.siteProperties[DC_CONSTS.SITE_PROP_AUTO_REMOVE_WHERE]
      if DC_CONSTS.SITE_PROP_AUTO_REMOVE_WHERE_ACTIVE in self.siteProperties:
        self.auto_remove_props[DC_CONSTS.SITE_PROP_AUTO_REMOVE_WHERE_ACTIVE] = \
        self.siteProperties[DC_CONSTS.SITE_PROP_AUTO_REMOVE_WHERE_ACTIVE]
      if 'URL_NORMALIZE_MASK' in self.siteProperties:
        self.normMask = int(self.siteProperties['URL_NORMALIZE_MASK'])
      if APP_CONSTS.URL_NORMALIZE in self.siteProperties:
        self.normMask = UrlNormalize.getNormalizeMask(siteProperties=self.siteProperties, defaultValue=self.normMask)
      if self.urlProcess is not None:
        self.urlProcess.normMask = self.normMask
      if 'URL_TEMPLATE_REGULAR' in self.siteProperties:
        self.urlTempalteRegular = self.siteProperties['URL_TEMPLATE_REGULAR']
      if 'URL_TEMPLATE_REALTIME' in self.siteProperties:
        self.urlTempalteRealtime = self.siteProperties['URL_TEMPLATE_REALTIME']
      if 'URL_TEMPLATE_REGULAR_URLENCODE' in self.siteProperties:
        self.urlTempalteRegularEncode = self.siteProperties['URL_TEMPLATE_REGULAR_URLENCODE']
      if 'URL_TEMPLATE_REALTIME_URLENCODE' in self.siteProperties:
        self.urlTempalteRealtimeEncode = self.siteProperties['URL_TEMPLATE_REALTIME_URLENCODE']
      if 'PROTOCOLS' in self.siteProperties:
        if self.urlProcess is not None:
          self.urlProcess.setProtocols(self.siteProperties['PROTOCOLS'])
      if 'CONTENT_TYPE_MAP' in self.siteProperties:
        self.siteProperties['CONTENT_TYPE_MAP'] = Utils.jsonLoadsSafe(jsonString=self.siteProperties['CONTENT_TYPE_MAP'], default={}, log=self.logger)

      if APP_CONSTS.URL_CHAIN in self.siteProperties:
        self.urlChain = UrlChain(self.siteProperties[APP_CONSTS.URL_CHAIN], self.logger)

      self.httpApplyHeaders = self.getHeadersDict(self.defaultHeaderFile, self.logger)
      self.logger.info("Initialized  self.httpApplyHeaders: %s", str(self.httpApplyHeaders))

      self.urlProcess.siteProperties = self.siteProperties

      self.logger.debug("self.siteProperties:\n%s", varDump(self.siteProperties, stringifyType=0))

    except Exception, err:
      ExceptionLog.handler(self.logger, err, MSG_ERROR_LOAD_SITE_PROPERTIES)
      self.errorMask |= APP_CONSTS.ERROR_CRAWLER_FATAL_INITIALIZATION_PROJECT_ERROR


  # # Get HTTP headers dictionary from default headers file
  #
  # @param fileName - headers file name
  # @param log - logger instance
  # @return headers dictionary
  @staticmethod
  def getHeadersDict(fileName, log=None):
    # variable for result
    resultDict = {}
    try:
      if not isinstance(fileName, basestring) or fileName == "" or not os.path.isfile(fileName):
        raise Exception("Argument '%s' has wrong value: %s" % ('fileName', str(fileName)))

      fileData = CrawlerTask.loadFromFile(fileName)
      if log is not None:
        log.debug("Loaded from %s fileData:\n%s" , str(fileName), str(fileData))

      for headerLine in fileData.splitlines():
        if not headerLine:
          continue
        try:
#           if log is not None:
#             log.info("Try usage '%s', type = %s", str(headerLine), str(type(headerLine)))

          key, value = headerLine[:headerLine.index(':')].strip(), headerLine[headerLine.index(':') + len(':'):].strip()
          if key[0] != '#':
            resultDict[key] = value
        except Exception, err:
          if log is not None:
            log.error("Error: %s, headerLine: %s", str(err), headerLine)

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return resultDict


  # # loadFromFile read file content
  #
  # @param fullFileName - full file name
  # @return file content as string
  @staticmethod
  def loadFromFile(fullFileName):
    with open(fullFileName, 'r') as f:
      return ''.join(f.readlines())


  # #update site params
  # update site params
  # @param mask ErrorMask
  # @param is_suspend should set state to suspend for this site
  def updateSiteParams(self, mask, is_suspend=False):
    ret = False
    if self.dbWrapper is None:
      ret = True
    else:
      try:
        localSiteUpdate = dc_event.SiteUpdate(self.batchItem.siteId)
        for attr in localSiteUpdate.__dict__:
          if hasattr(localSiteUpdate, attr):
            setattr(localSiteUpdate, attr, None)

        # TODO: possible redundant for the URLFetch algorithm and need to be removed
        if mask:
          # localSiteUpdate.errors = SQLExpression("`Errors` + 1")
          localSiteUpdate.errorMask = SQLExpression(("`ErrorMask` | %s" % mask))
          if is_suspend:
            localSiteUpdate.state = Site.STATE_SUSPENDED

        localSiteUpdate.id = self.batchItem.siteId
        localSiteUpdate.updateType = dc_event.SiteUpdate.UPDATE_TYPE_UPDATE
        updated_count = self.dbWrapper.siteNewOrUpdate(siteObject=localSiteUpdate, stype=dc_event.SiteUpdate)
        if updated_count > 0:
          ret = True
      except DatabaseException, err:
        ExceptionLog.handler(self.logger, err, MSG_ERROR_UPDATE_SITE_DATA, (err))
        raise err
      except Exception, err:
        ExceptionLog.handler(self.logger, err, MSG_ERROR_UPDATE_SITE_DATA, (err))
        raise err

    return ret


  # # update site use sql expression evaluator
  #
  # @param - None
  # @return ret - True if success or False otherwise
  def updateSite(self):
    ret = False

    if self.dbWrapper is None:
      ret = True
    else:
      if self.site is not None and APP_CONSTS.SQL_EXPRESSION_FIELDS_UPDATE_CRAWLER in self.siteProperties:
        try:
          localSiteUpdate = dc_event.SiteUpdate(self.batchItem.siteId)
          for attr in localSiteUpdate.__dict__:
            if hasattr(localSiteUpdate, attr):
              setattr(localSiteUpdate, attr, None)

          # Evaluate 'Site' class values if neccessary
          changedFieldsDict = FieldsSQLExpressionEvaluator.execute(
            self.siteProperties, self.dbWrapper, self.site, None, self.logger,
            APP_CONSTS.SQL_EXPRESSION_FIELDS_UPDATE_CRAWLER, self.crawledResource.response_header
          )
          # Update 'Site' class values
          for name, value in changedFieldsDict.items():
            if hasattr(localSiteUpdate, name) and value is not None and name not in ['CDate', 'UDate', 'tcDate']:
              setattr(localSiteUpdate, name, value)

          localSiteUpdate.errorMask = SQLExpression(("`ErrorMask` | %s" % self.site.errorMask))
          localSiteUpdate.id = self.batchItem.siteId
          localSiteUpdate.updateType = dc_event.SiteUpdate.UPDATE_TYPE_UPDATE

          updatedCount = self.dbWrapper.siteNewOrUpdate(siteObject=localSiteUpdate, stype=dc_event.SiteUpdate)
          self.logger.debug('Updated ' + str(updatedCount) + ' rows.')
          if updatedCount > 0:
            ret = True
        except DatabaseException, err:
          self.logger.error("Update 'Site' failed, error: %s", str(err))
          raise err
        except Exception, err:
          self.logger.error("Update 'Site' failed, error: %s", str(err))
          raise err

    return ret


  # #Check sevarl common conditions
  #
  # @return Boolean that means - should continue execute this BatchItem or not
  def commonChecks(self, urlObj):
    if self.site is None or urlObj is None:
      self.logger.error('Error: self.site or urlObj is None!')
      return False
    # self.logger.debug("urlObj:" + str(urlObj))

    if((self.batch.crawlerType != dc_event.Batch.TYPE_REAL_TIME_CRAWLER) and (self.site.state != Site.STATE_ACTIVE)) \
      or ((self.batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER) and (self.site.state == Site.STATE_DISABLED)):
      self.logger.debug("Warning: Batch CrawlerType: %s, site state is %s but is not STATE_ACTIVE!", str(self.batch.crawlerType), str(self.site.state))
      self.errorMask = self.errorMask | APP_CONSTS.ERROR_MASK_SITE_STATE
      return False

    if not self.isRootURL(urlObj):
      if self.site.maxErrors > 0 and self.site.errors > self.site.maxErrors:
        self.logger.debug("Site max errors: %s limit: %s is reached", str(self.site.errors), str(self.site.maxErrors))
        # TODO: possible improve suspend logic to use the property to define, suspend the site if limit reached or not
        # self.updateSiteParams(CONSTS.ERROR_SITE_MAX_ERRORS, True)
        self.errorMask = self.errorMask | APP_CONSTS.ERROR_SITE_MAX_ERRORS
        self.updateSiteParams(APP_CONSTS.ERROR_SITE_MAX_ERRORS)
        return False

    return True


  # #Is URL object a root URL
  #
  # @param urlObj
  # @param urlString
  # @return True if URL object a root URL or False
  def isRootURL(self, urlObj, urlString=None):
    ret = False

    if urlString is None:
      if urlObj.parentMd5 == '':
        ret = True
    else:
      if urlString == '':
        ret = True

    return ret


  # # load site structure
  # the site object to crawl
  # @param batch - batch object instance
  def loadSite(self, batch):
    try:
      # # FIXED alexv 2015-11-11
      # if not len(self.batchItem.siteId):
      #  self.batchItem.siteId = "0"

      if batch.crawlerType != dc_event.Batch.TYPE_REAL_TIME_CRAWLER:  # # FIXED alexv 2017-07-24
        self.site = self.loadSiteFromDB(self.batchItem.siteId)

        # Check if site not exist then read site for siteId=0
        if self.site.id == SITE_MD5_EMPTY and bool(self.useZeroSiteIdSiteNotExists):
          self.logger.debug("Site not found. Assume a site id as: `0`")
          self.batchItem.siteId = '0'
          self.batchItem.urlObj.siteId = '0'
          self.site = self.loadSiteFromDB(self.batchItem.siteId)
      else:
        # Create empty Site object
        self.site = dc_event.Site("")
        self.site.id = '0'

      # self.logger.debug(">>> site before = " + varDump(self.site))
      if self.site is not None and self.batchItem.siteObj is not None:
        self.site.rewriteFields(self.batchItem.siteObj)
      # self.logger.debug(">>> site after = " + varDump(self.site))
    except Exception as err:
      ExceptionLog.handler(self.logger, err, MSG_ERROR_LOAD_SITE_DATA, (err))
      self.errorMask |= APP_CONSTS.ERROR_CRAWLER_FATAL_INITIALIZATION_PROJECT_ERROR
      raise err


  # # load Site object from DB
  #
  # @param siteId - site ID
  # @return Site object from DB
  def loadSiteFromDB(self, siteId):
    #variable for result
    site = None

    # check cached Site objects
    if siteId in self.siteCacheDict and self.siteCacheDict[siteId] is not None:
      site = self.siteCacheDict[siteId]
    else:
      # get Site object from DB
      site = self.dbWrapper.siteStatus(dc_event.SiteStatus(siteId))
      if site is None or site.state == dc_event.Site.STATE_NOT_FOUND or (len(site.urls) > 0 and site.urls[0] == ''):
        self.logger.info("Error read Site '%s' from DB:\n%s", str(siteId), varDump(site, stringifyType=0))
        raise Exception(MSG_ERROR_READ_SITE_FROM_DB % str(siteId))

      # add to cache
      self.siteCacheDict[siteId] = site

    return site


  # #load url
  # the site object to crawl
  # @param site - object to crawl
  def loadURL(self):
    # do siteId

    if len(self.batchItem.siteId):
      self.urlTable = DC_URLS_TABLE_PREFIX + self.batchItem.siteId
    else:
      self.urlTable = DC_URLS_TABLE_PREFIX + "0"
    self.logger.debug("db backend Table for siteId %s is: %s", self.batchItem.siteId, self.urlTable)
    self.urlProcess.siteId = self.batchItem.siteId
    self.urlProcess.urlTable = self.urlTable
    self.url = self.batchItem.urlObj
    self.urlProcess.urlDBSync(self.batchItem, self.batch.crawlerType, self.site.recrawlPeriod,
                              self.auto_remove_props)
    if self.urlProcess.isUpdateCollection:
      self.updateCollectedURLs()


  # #resetVars
  def resetVars(self):
    self.needStoreMime = None  # pylint: disable=F0401,W0201
    self.site = None  # pylint: disable=F0401,W0201
#     self.headersDict = None  # pylint: disable=F0401,W0201
    self.store_http_headers = True  # pylint: disable=F0401,W0201
    self.realUrl = None  # pylint: disable=F0401,W0201
    self.url = None
    self.crawledResource = None
#     self.headers = None  # pylint: disable=F0401,W0201
    self.storeHttpRequest = True  # pylint: disable=F0401,W0201
    self.dir = None  # pylint: disable=F0401,W0201
    self.kvConnector = None  # pylint: disable=F0401,W0201
    self.kvCursor = None  # pylint: disable=F0401,W0201
    self.processorName = None  # pylint: disable=F0401,W0201
    self.auto_remove_props = None  # pylint: disable=F0401,W0201
    self.storeCookies = True  # pylint: disable=F0401,W0201
    self.allow_http_redirects = True
    self.allow_html_redirects = True  # pylint: disable=F0401,W0201
    self.httpRedirects = 0  # pylint: disable=F0401,W0201
    self.htmlRedirects = 0  # pylint: disable=F0401,W0201
    self.dom = None
    self.res = None  # pylint: disable=F0401,W0201
    self.postForms = None  # pylint: disable=F0401,W0201
    self.siteProperties = {}  # pylint: disable=F0401,W0201
    self.errorMask = APP_CONSTS.ERROR_OK
    self.feed = None  # pylint: disable=F0401,W0201
    self.feedItems = None
    self.urlTempalteRegular = None  # pylint: disable=F0401,W0201
    self.urlTempalteRealtime = None  # pylint: disable=F0401,W0201
    self.urlTempalteRegularEncode = None  # pylint: disable=F0401,W0201
    self.urlTempalteRealtimeEncode = None  # pylint: disable=F0401,W0201
    self.collectURLsItems = []
    self.schemaBatchItems = []
    self.urlChain = None


  # # update batch item after crawling
  #
  # @param startTime - start crawling time
  # @param status - the URL status value for update
  # @return - None
  def updateBatchItemAfterCarwling(self, startTime, status=dc_event.URL.STATUS_CRAWLED):
    self.urlProcess.urlObj = self.url
    self.urlProcess.siteId = self.batchItem.siteId
    self.logger.debug("set siteId = '" + str(self.urlProcess.siteId) + "' from 'updateBatchItemAfterCarwling'")

    # update crawling time by the actual value
    self.crawledResource.crawling_time = int((time.time() - startTime) * 1000)

    # Update status value accord to dict HTTP codes
    if "HTTP_CODE_STATUS_UPDATE" in self.siteProperties and self.siteProperties["HTTP_CODE_STATUS_UPDATE"] != "":
      self.logger.debug('!!!!!! HTTP_CODE_STATUS_UPDATE !!!!! ')
#       # Default value
#       status = dc_event.URL.STATUS_CRAWLED
      try:
        statusDict = json.loads(self.siteProperties["HTTP_CODE_STATUS_UPDATE"])
        self.logger.debug('!!!!!! statusDict: ' + str(statusDict))
        if str(self.crawledResource.http_code) in statusDict:
          self.logger.debug("Change status from (%s) to (%s), because http_code = %s", str(status), \
                            str(statusDict[str(self.crawledResource.http_code)]), str(self.crawledResource.http_code))
          status = int(statusDict[str(self.crawledResource.http_code)])
      except Exception, err:
        self.logger.error("Load property 'HTTP_CODE_STATUS_UPDATE' has error: " + str(err))

    if status is not None:
      self.batchItem.urlObj.status = dc_event.URL.STATUS_CRAWLED
    self.batchItem.urlObj.crawlingTime = self.crawledResource.crawling_time
    self.batchItem.urlObj.errorMask |= self.crawledResource.error_mask
    if self.batchItem.urlObj.charset == "":
      self.batchItem.urlObj.charset = self.crawledResource.charset
    if self.batchItem.urlObj.httpCode == 0:
      self.batchItem.urlObj.httpCode = self.crawledResource.http_code
    if self.batchItem.urlObj.contentType == "":
      self.batchItem.urlObj.contentType = self.crawledResource.content_type
    self.batchItem.urlObj.crawled = 1
    if self.batchItem.urlObj.size == 0:
      self.batchItem.urlObj.size = self.res.content_size
    self.batchItem.urlObj.CDate = self.batchItem.urlObj.UDate = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    self.batchItem.urlObj.tagsCount = 0
    self.batchItem.urlObj.tagsMask = 0
    if self.batchItem.urlObj.eTag == "":
      self.batchItem.urlObj.eTag = self.crawledResource.etag

    # update crawled URL
    self.urlProcess.updateCrawledURL(self.crawledResource, self.batchItem, self.res.content_size, status)

    self.logger.info("=====INSERT & UPDATE=====")
    self.logger.info("type(self.url) = %s", str(type(self.url)))
    self.logger.info("self.url: %s", varDump(self.url))

    self.logger.info("hasattr(self.url, 'urlOld') = %s", str(hasattr(self.url, 'urlOld')))

    if self.url.parentMd5 != "":
      urlOld = self.url.urlOld if hasattr(self.url, 'urlOld') else self.url.url

      self.writeData(self.updateResolvedURL(siteProperties=self.siteProperties, batchItem=self.batchItem, urlOld=urlOld, urlNew=self.url.url))

      # update processed content
      self.urlProcess.updateProcessedContent(siteProperties=self.siteProperties, dbWrapper=self.dbWrapper, batchItem=self.batchItem, urlOld=urlOld,
                                             urlNew=self.url.url, normalizeMask=self.normMask, log=self.logger)

#     if hasattr(self.url, 'urlOld') and self.url.parentMd5 != "":
#       self.writeData(self.updateResolvedURL(siteProperties=self.siteProperties, batchItem=self.batchItem, urlOld=self.url.urlOld, urlNew=self.url.url))
#
#       # update processed content
#       self.urlProcess.updateProcessedContent(siteProperties=self.siteProperties, dbWrapper=self.dbWrapper, batchItem=self.batchItem, urlOld=self.url.urlOld,
#                                              urlNew=self.url.url, normalizeMask=self.normMask, log=self.logger)
#


  # # Update crawled URL by resolved
  #
  # @param siteProperties - site properties dictionary
  # @param batchItem - batch item instance
  # @param urlOld - old URL value
  # @param urlNew - new URL value
  # @return updated batch item
  def updateResolvedURL(self, siteProperties, batchItem, urlOld, urlNew):
    # variable for result
    localBatchItem = copy.deepcopy(batchItem)

    if self.dbWrapper is None:
      self.logger.info("self.dbWrapper is None")
      return localBatchItem

    self.logger.info("Update resolved URL...\nurlOld: %s\nurlNew: %s", str(urlOld), str(urlNew))

    # get flag insert as new URL
    insertAsNew = HTTPRedirectExtractor.isInsertAsNewUrl(siteProperties=siteProperties, url=urlNew, log=self.logger)
    self.logger.debug("insertAsNew = %s", str(insertAsNew))

    # create temporary object for use native normalization and urlMd5 calculation algorithm
    tmpUrlObj = dc_event.URL(batchItem.siteId, urlOld, normalizeMask=self.normMask)
    urlUpdateObj = dc_event.URLUpdate(tmpUrlObj.siteId, tmpUrlObj.urlMd5, dc_event.URLStatus.URL_TYPE_MD5, normalizeMask=self.normMask)

    if hasattr(self.res, 'firstRedirectResponse'):
      firstRedirectResponse = getattr(self.res, 'firstRedirectResponse')
#       self.logger.info("firstRedirectResponse: %s", varDump(firstRedirectResponse, stringifyType=0))

      urlUpdateObj.httpCode = firstRedirectResponse.status_code
      urlUpdateObj.contentType = RequestsRedirectWrapper.getContentType(firstRedirectResponse)

    # update old URL
    HTTPRedirectExtractor.updateUrl(siteProperties=siteProperties, urlObj=urlUpdateObj, url=urlOld, log=self.logger)

    self.logger.info("Before update urlUpdateObj: %s", varDump(urlUpdateObj))
    affectDB = self.dbWrapper.affect_db
    self.dbWrapper.affect_db = True
    updatedRowsCount = self.dbWrapper.urlUpdate([urlUpdateObj], "`Status` = 3")
    self.dbWrapper.affect_db = affectDB
    self.logger.info("Updated rows count = %s", str(updatedRowsCount))

    if insertAsNew:  # and urlOld != urlNew:
      # create temporary object for use native normalization and urlMd5 calculation algorithm
      tmpUrlObj = dc_event.URL(batchItem.siteId, urlNew, normalizeMask=self.normMask)

      # initialization object for insert
      localBatchItem.urlObj.url = tmpUrlObj.url
      localBatchItem.urlObj.urlMd5 = tmpUrlObj.urlMd5
      localBatchItem.urlId = tmpUrlObj.urlMd5

      affectDB = self.dbWrapper.affect_db
      self.dbWrapper.affect_db = True
      if urlOld == urlNew:
        # necessary update old by crawled
        urlUpdateObj.crawled = 1
        updatedRowsCount = self.dbWrapper.urlUpdate([urlUpdateObj], "`Status` = 3")
        self.logger.debug("Update URL return updatedRowsCount = %s", str(updatedRowsCount))
      else:
        # necessary insert as new
        result = self.dbWrapper.urlNew([localBatchItem.urlObj])
        self.logger.debug("Insert URL as new return result: " + str(result))

      self.dbWrapper.affect_db = affectDB

    return localBatchItem


  # #process batch item
  # the batch item is the one from list of batch items within batch object
  # @param item - the batch item is the one from list of batch items within batch object
  def processBatchItem(self):
    self.resetVars()
    self.logger.info("Batch item START, batchId: %s siteId: %s, urlId: %s, self.batch.crawlerType: %s",
                     str(self.batch.id), str(self.batchItem.siteId), str(self.batchItem.urlId), str(self.batch.crawlerType))

    startTime = time.time()
    isResourceCrawled = False
    nextStep = True
    detectedMime = None
    mpLogger = Utils.MPLogger()
    self.urlProcess.siteId = self.batchItem.siteId

    try:
      # self.logger.debug("BatchItem: %s" % varDump(self.batchItem))
      self.loadSite(self.batch)

      # prepare headers and cookie
      self.readSiteProperties()

      self.logger.debug("!!! hasattr(self.batchItem, 'processorName') = %s", str(hasattr(self.batchItem, 'processorName')))
      if hasattr(self.batchItem, 'processorName'):
        self.logger.debug("!!! self.batchItem.processorName = '%s'", str(self.batchItem.processorName))

      self.logger.debug("!!! hasattr(self.batchItem, 'parentContentType') = %s", str(hasattr(self.batchItem, 'parentContentType')))
      if hasattr(self.batchItem, 'parentContentType'):
        self.logger.debug("!!! self.batchItem.parentContentType = '%s'", str(self.batchItem.parentContentType))

      if (hasattr(self.batchItem, 'processorName') and self.batchItem.processorName is not None) or \
        (hasattr(self.batchItem, 'parentContentType') and self.batchItem.parentContentType == dc_event.URL.CONTENT_TYPE_TEXT_HTML):
        self.logger.debug("!!! CHANGE PROCESSOR NAME self.processorName = %s, self.batchItem.processorName = %s",
                          str(self.processorName), str(self.batchItem.processorName))
        self.processorName = self.batchItem.processorName
      else:
        self.logger.debug("!!! NOT CHANGED PROCESSOR NAME self.processorName = %s, url: %s", str(self.processorName), str(self.batchItem.urlObj.url))

#       if 'LOGGER' in self.siteProperties and self.siteProperties['LOGGER'] is not None and\
#         self.siteProperties['LOGGER'] != '':
#         self.logger.info("Switch logger to dedicated project %s log", str(self.batchItem.siteId))
#         self.setLogConfigFileProject(mpLogger, self.batchItem.siteId, self.siteProperties['LOGGER'])

      if APP_CONSTS.LOGGER in self.siteProperties and self.siteProperties[APP_CONSTS.LOGGER] is not None and self.siteProperties[APP_CONSTS.LOGGER] != '':
        self.logger = LogSwitch.setLogConfigFileProject(mpLogger, self.batchItem.siteId, self.siteProperties[APP_CONSTS.LOGGER], self.logger)

      self.setChainId(self.batchItem)
      if self.robotsParser is not None:
        self.robotsParser.\
        initFiends(None, (bool(int(self.siteProperties['ROBOTS_CACHE'])) if \
                          'ROBOTS_CACHE' in self.siteProperties else False),
                   self.robotsFileDir)
      # Check if Real-Time crawling
      if self.batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER:
        # TODO: need for comments
        if self.batch.dbMode & dc_event.Batch.DB_MODE_W == 0:
          self.url = self.batchItem.urlObj  # pylint: disable=W0201
      else:
        self.loadURL()

      if not self.commonChecks(self.url):
        self.logger.debug("Common checks failed!")
        nextStep = False

      if nextStep and "HTTP_FREQ_LIMITS" in self.siteProperties:
        self.logger.debug("Enter to block HTTP_FREQ_LIMITS")
        if not HostRequestStorage.isAvailableHost(self.url.url, self.batchItem.siteId, self.siteProperties.get("HTTP_FREQ_LIMITS"), log=self.logger):
          self.logger.debug(">>> Skip url [%s] by http requests freq", self.url.url)
          self.url.status = dc_event.URL.STATUS_NEW
          self.url.errorMask = APP_CONSTS.ERROR_NO_TIME_WINDOW
          self.url.httpCode = CONSTS.HTTP_CODE_400
          self.urlProcess.updateURLStatus(self.batchItem.urlId, self.url.status)
          nextStep = False
        else:
          self.logger.debug("Skipped block HTTP_FREQ_LIMITS")

      # Check if Real-Time crawling resource already crawled
      if nextStep and self.batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER:
        self.logger.debug("Enter to block 'self.batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER'")
        # TODO: strange condition: self.url.tagsCount > 0
        if self.url.crawled > 0 and self.url.errorMask == APP_CONSTS.ERROR_OK and self.url.tagsCount > 0 and \
          self.errorMask == APP_CONSTS.ERROR_OK:
          self.logger.debug("RealTime Crawling: Cashed resource. Resource crawled and error mask is empty")
          if PCONSTS.RECRAWL_KEY not in self.batchItem.properties or \
          int(self.batchItem.properties[PCONSTS.RECRAWL_KEY]) == PCONSTS.RECRAWL_VALUE_NO:
            self.logger.debug("Item not need to be recrawled.")
            # set contentMask stored on disk
            self.batchItem.urlObj.contentMask = dc_event.URL.CONTENT_STORED_ON_DISK
            # return self.batchItem
            nextStep = False
          else:
            self.logger.debug("Property `recrawl` = %s. Item recrawling.", str(self.batchItem.properties[PCONSTS.RECRAWL_KEY]))

      else:
        self.logger.debug("Skipped block 'self.batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER'")

      if nextStep:
        # reset error mask for current crawled url
        self.urlProcess.resetErrorMask(self.batchItem)
        # update state to crawling
        self.urlProcess.urlObj = self.url
        self.urlProcess.updateURL(self.batchItem, self.batch.id)
        # prepare incoming content
        incomingContent = None
        if self.batchItem.urlObj.urlPut is not None and self.batchItem.urlObj.urlPut.putDict is not None and "data" in self.batchItem.urlObj.urlPut.putDict:
          incomingContent = self.batchItem.urlObj.urlPut.putDict["data"]

        self.logger.debug('Start to crawl item')
        isResourceCrawled = self.crawl(incomingContent, startTime)
        self.logger.debug("After crawl() isResourceCrawled: %s", str(isResourceCrawled))

        if self.crawledResource is not None and isResourceCrawled:
          # TODO move to - [SiteProcess]
          self.updateAVGSpeed()

          self.logger.debug("!!!!!! self.batchItem.urlObj.type = %s", str(self.batchItem.urlObj.type))
          # Build DOM object
          if self.batchItem.urlObj.type == dc_event.URL.TYPE_SINGLE:
            self.logger.debug('URL type single, do not pars and build DOM, set self.dom = None')
            self.dom = None
          else:
            self.logger.debug("Build DOM, call getDom()\nself.crawledResource.charset = %s\nurl: %s", str(self.crawledResource.charset), str(self.url.url))
            self.dom = ResourceProcess.getDom(siteProperties=self.siteProperties, crawledResource=self.crawledResource, log=self.logger)
            self.logger.debug("!!! self.dom: %s", str(self.dom))
            self.logger.debug("hasattr(self.crawledResource, 'tidyContent'): %s", str(hasattr(self.crawledResource, 'tidyContent')))

            # # check dom value
            if self.dom is None:
              self.errorMask |= APP_CONSTS.ERROR_PARSE_ERROR
              self.updateURLForFailed(self.errorMask)

          # Set pDate accord to site property 'PDATE_SOURCE_MASK'
          self.logger.debug('>>>>> self.crawledResource.last_modified = ' + str(self.crawledResource.last_modified))
          self.logger.debug("!!!>>>> self.feedItems: %s", str(self.feedItems))

          self.logger.debug('!!! Before getPubdateUseSourceMask() self.batchItem.urlObj.pDate = ' + \
                            str(self.batchItem.urlObj.pDate))
          self.batchItem.urlObj.pDate = self.getPubdateUseSourceMask(self.siteProperties, \
                                                             self.crawledResource, \
                                                             self.batchItem.urlObj)
          self.logger.debug('!!! After getPubdateUseSourceMask() self.batchItem.urlObj.pDate = ' + \
                            str(self.batchItem.urlObj.pDate))

          # Add 'feed_url' to headers
          if self.processorName == PCONSTS.PROCESSOR_FEED_PARSER or self.processorName == PCONSTS.PROCESSOR_RSS:
            # get parent URL
            parentUrl = RefererHeaderResolver.getUrlByMD5(self.batchItem.siteId, self.url.parentMd5, self.dbWrapper, self.logger)
            self.logger.debug("!!! parentUrl: %s", str(parentUrl))
            self.logger.debug("!!! self.url.parentMd5: %s", str(self.url.parentMd5))
            self.logger.debug("!!! self.url.url: %s", str(self.url.url))
            self.logger.debug("!!! self.feedUrl: %s", str(self.feedUrl))

            if self.url.parentMd5 == "":
              self.feedUrl[hashlib.md5(self.url.url).hexdigest()] = self.url.url
            elif self.url.parentMd5 != "" and parentUrl is None:
              if self.url.parentMd5 in self.feedUrl:
                parentUrl = self.feedUrl[self.url.parentMd5]

            if parentUrl is not None and parentUrl != "":
              HTTPHeaders.addFeedUrlToHeader(self.crawledResource, parentUrl)

          self.logger.debug("!!! getHeaderContent() !!!")
          headerContent = CrawlerDataStorage.getHeaderContent(self.raw_data_dir, self.batchItem.siteId, self.batchItem.urlObj.urlMd5, self.logger)
          self.logger.debug("!!! headerContent:\n%s", str(headerContent))

          defaultBaseUrl = HTTPHeaders.getValue(headerContent, HTTPHeaders.BASE_URL_NAME)
          self.logger.debug("!!! defaultBaseUrl from %s: %s", str(HTTPHeaders.BASE_URL_NAME), str(defaultBaseUrl))

          if defaultBaseUrl is None:
            defaultBaseUrl = self.batchItem.urlObj.url
          self.logger.debug("!!! self.batchItem.urlObj.url: %s", str(self.batchItem.urlObj.url))
          self.logger.debug("!!! defaultBaseUrl: %s", str(defaultBaseUrl))
          # Add 'base_url' to batch item and header
          self.batchItem.urlObj.baseUrl = self.extractBaseUrl(self.crawledResource.html_content, defaultBaseUrl)
          headerContent = HTTPHeaders.setValue(headerContent, HTTPHeaders.BASE_URL_NAME, self.batchItem.urlObj.baseUrl)
          self.logger.debug("Add 'base_url': %s", str(self.batchItem.urlObj.baseUrl))

          # Add header from FS to header
          HTTPHeaders.addHeaderToHeader(self.crawledResource, headerContent)

          normalizedUrl = Utils.urlNormalization(base=self.batchItem.urlObj.url, url=self.batchItem.urlObj.baseUrl, log=self.logger)
          if normalizedUrl is not None:
            self.batchItem.urlObj.baseUrl = normalizedUrl

          # Add 'canonical_url to' to batch item and header
          if APP_CONSTS.HTML_CANONICAL_REDIRECT in self.siteProperties and self.siteProperties[APP_CONSTS.HTML_CANONICAL_REDIRECT] != "":
            self.logger.debug("!!! %s !!!", str(APP_CONSTS.HTML_CANONICAL_REDIRECT))

            canonicalRedirectUrl = HTMLCanonicalRedirect.extract(rawContent=self.crawledResource.html_content,
                                                                 properties=self.siteProperties[APP_CONSTS.HTML_CANONICAL_REDIRECT],
                                                                 url=self.batchItem.urlObj.url,
                                                                 log=self.logger)

            self.logger.debug("Extracted canonical url: %s", str(canonicalRedirectUrl))
            if canonicalRedirectUrl is not None:
              HTTPHeaders.addCanonicalUrlToHeader(self.crawledResource, canonicalRedirectUrl)
              self.url.url = copy.copy(canonicalRedirectUrl)

          self.logger.debug("!!! Before writeData() nextStep = %s", str(nextStep))
          try:
            self.getDir()
            if self.processorName != PCONSTS.PROCESSOR_FEED_PARSER:
              self.writeData(self.batchItem)

            if APP_CONSTS.MIME_TYPE_AUTO_DETECT in self.siteProperties:
              self.logger.debug("!!! %s !!!", str(APP_CONSTS.MIME_TYPE_AUTO_DETECT))

            # Detect mime type if necessary
            detectedMime = MimeTypeExtractor.extract(crawledResource=self.crawledResource,
                                                     siteProperties=self.siteProperties,
                                                     urlObj=self.batchItem.urlObj,
                                                     log=self.logger,
                                                     default=detectedMime)
            self.logger.debug("detectedMime = %s", str(detectedMime))
            # apply content type
            if detectedMime is not None:
              self.crawledResource.content_type = detectedMime

          except Exception, err:
            self.errorMask |= APP_CONSTS.ERROR_WRITE_FILE_ERROR
            self.updateURLForFailed(self.errorMask)
            self.logger.error(MSG_ERROR_WRITE_CRAWLED_DATA + ': ' + str(err))

          self.logger.debug("!!! After writeData() nextStep = %s", str(nextStep))

        else:
          nextStep = False
      else:
        self.logger.debug("Skipped block 'Start to crawl item'")

      if nextStep:
        self.logger.debug("Enter in collectURLs()")
        self.collectURLs()

        if self.batchItem.urlObj.chainId is not None and self.batchItem.urlObj.chainId in self.chainDict and \
        self.chainDict[self.batchItem.urlObj.chainId]["batchItem"].urlId == self.batchItem.urlId:
          self.chainDict[self.batchItem.urlObj.chainId]["batchItem"].depth += 1

          self.updateBatchItemAfterCarwling(startTime, None)
          self.logger.debug(">>> ChainId, update URL without status")
        else:
          self.updateBatchItemAfterCarwling(startTime)

        # self.logger.info("!!!! self.chainDict: %s", varDump(self.chainDict, stringifyType=0, maxDepth=15))
        # self.logger.debug('self.feed: %s, self.feedItems: %s', str(self.feed), str(self.feedItems))
        # TODO move to - [UrlProcess]
        self.urlProcess.siteId = self.batchItem.siteId
        if self.res is not None:
          self.urlProcess.updateCollectTimeAndMime(detectedMime, self.batchItem, startTime, self.res.headers, self.res.str_content)
        else:
          self.urlProcess.updateCollectTimeAndMime(detectedMime, self.batchItem, startTime)
      # (1) END
      else:
        self.logger.debug("Skipped block 'Enter in collectURLs()'")

      # # Update site properties if necessary
      if APP_CONSTS.SQL_EXPRESSION_FIELDS_UPDATE_CRAWLER in self.siteProperties:
        self.updateSite()

    except DatabaseException, err:
      self.errorMask |= APP_CONSTS.ERROR_DATABASE_ERROR
      ExceptionLog.handler(self.logger, err, MSG_ERROR_PROCESS_BATCH_ITEM, (err))
    except SyncronizeException, err:
      self.errorMask |= APP_CONSTS.ERROR_SYNCHRONIZE_URL_WITH_DB
      ExceptionLog.handler(self.logger, err, MSG_ERROR_PROCESS_BATCH_ITEM, (err), \
                           {ExceptionLog.LEVEL_NAME_ERROR:ExceptionLog.LEVEL_VALUE_DEBUG})
    except Exception as err:
      ExceptionLog.handler(self.logger, err, MSG_ERROR_PROCESS_BATCH_ITEM, (err))

#     if 'LOGGER' in self.siteProperties and self.siteProperties['LOGGER'] is not None and\
#       self.siteProperties['LOGGER'] != '':
#       self.logger.info("Switch logger back to default from dedicated for project %s", str(self.batchItem.siteId))
#       self.setLogConfigFileDefault(mpLogger)
#       self.logger.info("Switched logger back to default from dedicated for project %s", str(self.batchItem.siteId))

    # set default logger
    if APP_CONSTS.LOGGER in self.siteProperties and self.siteProperties[APP_CONSTS.LOGGER] is not None and self.siteProperties[APP_CONSTS.LOGGER] != '':
      self.logger = LogSwitch.setLogConfigFileDefault(mpLogger, self.batchItem.siteId, self.logger)

    if self.dbWrapper is not None:
      # Update db counters
      self.logger.debug('>>>>> Before self.dbWrapper.fieldsRecalculating([self.batchItem.siteId])')
      self.dbWrapper.fieldsRecalculating([self.batchItem.siteId])
      self.logger.info("Batch item FINISH, siteId: %s, urlId: %s", str(self.batchItem.siteId), str(self.batchItem.urlId))


  # # Get input data
  #
  # @param inputFile - input file name
  # @return extracted data
  @staticmethod
  def getInputData(inputFile):
    # variable for result
    inputData = None

    if inputFile is None:
      inputData = sys.stdin.read()
    else:
      with open(inputFile, 'r') as f:
        inputData = f.read()

    return inputData


  # # Send output data
  #
  # @param outputFile - output file name
  # @param outputData - output data string
  # @return - None
  @staticmethod
  def sendOutputData(outputFile, outputData):
    if outputFile is None:
      sys.stdout.write(outputData)
      sys.stdout.flush()
    else:
      f = open(outputFile, 'w')
      f.write(outputData)
      f.close()


  # #process batch
  # the main processing of the batch object
  def processBatch(self):
    try:
      input_pickled_object = self.getInputData(self.inputFile)
      # self.logger.debug("len(input_pickle)=%i", len(input_pickled_object))
      input_batch = pickle.loads(input_pickled_object)

      if input_batch.crawlerType != dc_event.Batch.TYPE_REAL_TIME_CRAWLER:
        # Create DBWrapper instance
        self.dbWrapper = self.__createDBTasksWrapper(self.dbTaskIni)
        self.urlProcess.dbWrapper = self.dbWrapper
        self.resourceProcess.dbWrapper = self.dbWrapper

      app.Profiler.messagesList.append("Batch.id: " + str(input_batch.id))
      self.logger.info("Input batch id: %s, items: %s", str(input_batch.id), str(len(input_batch.items)))
      # self.logger.debug("input_batch:\n" + varDump(input_batch))
      self.logger.debug("len before (batch_items)=%i", len(input_batch.items))
      self.generateBatchitemsByURLSchema(input_batch)
      self.logger.debug("len after (batch_items)=%i", len(input_batch.items))

      if int(input_batch.maxExecutionTime) > 0 or (self.maxTimeCli is not None and self.maxTimeCli > 0):
        self.maxExecutionTimeValue = input_batch.maxExecutionTime
        # # set value from cli
        if self.maxTimeCli is not None and self.maxTimeCli > 0:
          self.maxExecutionTimeValue = int(self.maxTimeCli)

        signal.signal(signal.SIGALRM, self.signalHandlerTimer)
        signal.alarm(self.maxExecutionTimeValue)
        self.removeUnprocessedItems = bool(input_batch.removeUnprocessedItems)
        self.logger.debug("Set maxExecutionTime = %s, removeUnprocessedItems = %s",
                          str(self.maxExecutionTimeValue), str(self.removeUnprocessedItems))

      self.batch = input_batch
      Utils.storePickleOnDisk(input_pickled_object, ENV_CRAWLER_STORE_PATH, "crawler.in." + str(self.batch.id))
      batch_items = []
      self.curBatchIterations = 1
      maxBatchIterations = input_batch.maxIterations
      # input_batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER:
        # TODO: temporary set as 2 to test w/o support on client side
        # maxBatchIterations = 2

      self.logger.debug("maxBatchIterations=%s", str(maxBatchIterations))
      if self.batch.dbMode & dc_event.Batch.DB_MODE_W == 0:
        if self.dbWrapper is not None:
          self.dbWrapper.affect_db = False
      while True:
        self.curBatchIterations += 1
        self.urlProcess.currentBatchIterations = self.curBatchIterations
        beforeItems = 0
        self.logger.debug("self.curBatchIterations=%s, beforeItems=%s", str(self.curBatchIterations), str(beforeItems))
        inputBatchItemsNumber = len(self.batch.items)
        rangeByNodeNumber = xrange(inputBatchItemsNumber)

        self.logger.debug("="*50)
        self.logger.debug("self.nodeNumber = %s, self.nodesNumber = %s, self.batchMode = %s, isOnBatchMode() = %s",
                          str(self.nodeNumber), str(self.nodesNumber), str(self.batchMode), str(self.isOnBatchMode()))

        if self.nodeNumber is not None and self.nodesNumber is not None and self.isOnBatchMode():
          self.logger.debug("Usage ranges algorithm.")
#           self.logger.debug("self.nodeNumber = %s, self.nodesNumber = %s", str(self.nodeNumber), str(self.nodesNumber))

          rangeItems = math.ceil(inputBatchItemsNumber / float(self.nodesNumber))
          self.logger.debug("rangeItems = %s", str(rangeItems))

          rangeByNodeNumber = xrange(int(rangeItems * self.nodeNumber), int(rangeItems * (self.nodeNumber + 1)))  # - 1))
          self.logger.debug("rangeByNodeNumber = %s", str(rangeByNodeNumber))

        for index, batchItem in enumerate(self.batch.items):
          if index not in rangeByNodeNumber:
            self.logger.debug("Item: %s is not in a range: %s, node number: %s, nodes total: %s, skipped!",
                              str(index), str(rangeByNodeNumber), str(self.nodeNumber), str(self.nodesNumber))
            continue
          # check max execution time
          if self.maxExecutionTimeReached:
            self.logger.debug("Maximum execution time %ss reached, process batch items loop interrupted!",
                              str(self.maxExecutionTimeValue))
            self.errorMask = APP_CONSTS.ERROR_MAX_EXECUTION_TIME
            if self.removeUnprocessedItems:
              break
            else:
              batchItem.urlObj.status = dc_event.URL.STATUS_NEW
              batch_items.append(batchItem)
              continue

          if batchItem.urlObj.status == dc_event.URL.STATUS_NEW or \
          batchItem.urlObj.status == dc_event.URL.STATUS_SELECTED_CRAWLING or \
          batchItem.urlObj.status == dc_event.URL.STATUS_SELECTED_CRAWLING_INCREMENTAL:
            self.errorMask = batchItem.urlObj.errorMask
            batchItem.urlObj.status = dc_event.URL.STATUS_CRAWLING
            self.batchItem = batchItem
            self.logger.debug('========== log flush for batchId = ' + str(self.batch.id) +
                              ' batchItem index = ' + str(index))
            Utils.loggerFlush(self.logger)  # flush logger
            self.processBatchItem()
            self.logger.debug('========== log flush for batchId = ' + str(self.batch.id))
            Utils.loggerFlush(self.logger)  # flush logger
            self.fillChainUrlMD5List(self.batchItem)
            # Currently we update URL.ErrorMask for each url in batch
            self.updateUrlObjInBatchItem(batchItem.urlObj)

            if self.processorName != PCONSTS.PROCESSOR_RSS and self.batchItem is None:
              self.logger.debug('!!! Before self.updateBatchItem(self.batchItem)')
              # self.updateBatchItem(batchItem)
              self.updateBatchItem(self.batchItem)
              input_batch.errorMask |= self.errorMask
            elif self.processorName != PCONSTS.PROCESSOR_RSS:
              self.logger.debug('!!! Before self.updateBatchItem(batchItem)')
              # self.updateBatchItem(self.batchItem)
              # batch_items.append(self.batchItem)
              self.updateBatchItem(batchItem)
              batch_items.append(batchItem)

#             if self.urlChain is not None and maxBatchIterations < self.urlChain.maxBatchIterations and len(self.store_items) > 0:
              if APP_CONSTS.PROCESSOR_NAME in self.siteProperties and \
                self.siteProperties[APP_CONSTS.PROCESSOR_NAME] in [PCONSTS.PROCESSOR_RSS, PCONSTS.PROCESSOR_FEED_PARSER] and \
                len(self.store_items) > 0 and self.urlChain is not None:
                if self.batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER:
                  maxBatchIterations = 3
                else:
                  maxBatchIterations = 2

            # Extend self.store_items with new batch items if iterations >1 and not last iteration
            if maxBatchIterations > 1 and self.curBatchIterations <= maxBatchIterations:
              self.logger.debug("!!! (1) self.batchItemsExtendUnique()")
              self.logger.debug("!!! self.batchItem.urlObj.type = %s", str(self.batchItem.urlObj.type))
              self.logger.debug("!!! self.batchItem.urlObj.contentType = %s", str(self.batchItem.urlObj.contentType))
              self.batchItemsExtendUnique(self.store_items, self.collectURLsItems, processorName="")
            beforeItems += len(self.collectURLsItems)

        # self.logger.debug("!!! self.collectURLsItems: " + varDump(self.collectURLsItems))
        # Exit the batch cycling if no one new item added or max iterations reached
        if self.curBatchIterations > maxBatchIterations or beforeItems == 0:
          self.logger.debug("Exit from batching iteration:" + \
                            "self.curBatchIterations=%s, maxBatchIterations=%s, beforeItems=%s, self.store_items=%s",
                            str(self.curBatchIterations), str(maxBatchIterations), str(beforeItems),
                            str(len(self.store_items)))
          break
        else:
          self.batch.items = self.store_items
          # self.logger.debug("Next batching iteration %s\n%s", str(self.curBatchIterations), varDump(self.store_items))
          self.logger.debug("Next batching iteration %s, len(self.store_items): %s", str(self.curBatchIterations),
                            str(len(self.store_items)))

      if input_batch.crawlerType == dc_event.Batch.TYPE_REAL_TIME_CRAWLER:
        process_task_batch = input_batch
        process_task_batch.items = self.store_items
      else:
        process_task_batch = Batch(input_batch.id, batch_items)
        process_task_batch.errorMask |= self.errorMask

      self.saveChainStorageData()

      if self.processorName == PCONSTS.PROCESSOR_RSS and len(process_task_batch.items) == 0 and \
        self.batchItem is not None:
        self.logger.debug("RSS empty!")
        if self.batchItem.urlObj is not None:
          self.batchItem.urlObj.errorMask |= APP_CONSTS.ERROR_RSS_EMPTY
        process_task_batch.items.append(self.batchItem)

      # self.logger.debug("output_batch:\n" + varDump(process_task_batch))
      self.logger.info("Out batch id: %s, node number: %s, items: %s", str(process_task_batch.id), str(self.nodeNumber), str(len(process_task_batch.items)))
      output_pickled_object = pickle.dumps(process_task_batch)
      Utils.storePickleOnDisk(output_pickled_object, ENV_CRAWLER_STORE_PATH, "crawler.out." + str(self.batch.id))
#       sys.stdout.write(output_pickled_object)
#       sys.stdout.flush()
      self.sendOutputData(self.outputFile, output_pickled_object)

    except Exception, err:
      ExceptionLog.handler(self.logger, err, 'Batch processing failed!', (err))


  # #generateBatchitemsByURLSchema methon added new items into the batch.items list
  #
  # @param batch - incoming batch
  def generateBatchitemsByURLSchema(self, batch):

    additionBatchItems = []
    for batchItem in batch.items:
      self.batchItem = batchItem
      self.resetVars()
      self.loadSite(batch)
      self.readSiteProperties()
      self.batchItem.siteProperties = copy.deepcopy(self.siteProperties)
      batchItem = self.changeBatchItemByUrlSchema(self.batchItem, batch.id)

      self.logger.debug("batchItem.urlObj.url: " + str(batchItem.urlObj.url))

      self.logger.debug('>>> len(additionBatchItems) = ' + str(len(additionBatchItems)) + \
                        ' len(self.schemaBatchItems) = ' + str(len(self.schemaBatchItems)))

      for elem in self.schemaBatchItems:
        self.logger.debug("url: " + str(elem.urlObj.url))

      additionBatchItems += self.schemaBatchItems
    tmpBatchItems = []
    for elem in additionBatchItems:
      if elem.urlId not in [e.urlId for e in tmpBatchItems]:
        tmpBatchItems.append(elem)

    batch.items += tmpBatchItems

    self.logger.debug("len(batch.items) = " + str(len(batch.items)) + \
                      " len(tmpBatchItems) = " + str(len(tmpBatchItems)))

    # TODO maxItems, maxItems <= len(batch.items)
    if batch.maxItems is not None and int(batch.maxItems) < len(batch.items):
      batch.items = batch.items[0: self.batch.maxItems]
      batch.items[-1].urlObj.errorMask |= APP_CONSTS.ERROR_MAX_ITEMS


  # #changeBatchItemByUrlSchema generates new url bu urlSchema and replace batchItem.urlObject if url was change
  #
  # @param batchItem - incoming batchItem
  # @return batchItem with changed urlObj or untoched batchItem
  def changeBatchItemByUrlSchema(self, batchItem, batchId):
    # self.schemaBatchItems
    if "URLS_SCHEMA" in self.siteProperties:
      urlSchema = UrlSchema(self.siteProperties["URLS_SCHEMA"], batchItem.urlObj.siteId, self.urlSchemaDataDir)
      newUrls = urlSchema.generateUrlSchema(batchItem.urlObj.url)
      self.logger.debug("Generated new urls count = %s", str(len(newUrls)))

      # check limits
      if self.site.maxURLs > 0 and len(newUrls) >= self.site.maxURLs:
        newUrls = set(list(newUrls)[:self.site.maxURLs])
        self.logger.debug("Site maxURLs = %s limit reached.", str(self.site.maxURLs))

      if self.site.maxResources > 0 and len(newUrls) >= self.site.maxResources:
        newUrls = set(list(newUrls)[:self.site.maxResources])
        self.logger.debug("Site maxResources = %s limit reached.", str(self.site.maxResources))

      # url update
      if len(newUrls) > 0:
        self.logger.debug("Url was changed. From %s to %s", batchItem.urlObj.url, newUrls[0])

        if self.dbWrapper is not None:
          urlUpdateObj = dc_event.URLUpdate(siteId=batchItem.urlObj.siteId, urlString=batchItem.urlObj.url,
                                            normalizeMask=UrlNormalizator.NORM_NONE)
          urlUpdateObj.urlMd5 = batchItem.urlObj.urlMd5
          urlUpdateObj.batchId = batchId
          urlUpdateObj.crawled = SQLExpression("`Crawled`+1")
          urlUpdateObj.processed = 0
          urlUpdateObj.status = dc_event.URL.STATUS_CRAWLED
          urlUpdateObj.size = 0
          urlUpdateObj.contentType = ""
          result = self.dbWrapper.urlUpdate(urlUpdateObj)
          self.logger.debug("urlUpdate() return result: " + str(result))

        batchItem.urlObj.url = newUrls[0]
        batchItem.urlObj.parentMd5 = batchItem.urlObj.urlMd5
        if urlSchema.externalError != APP_CONSTS.ERROR_OK:
          batchItem.urlObj.errorMask |= urlSchema.externalError
        batchItem.urlId = batchItem.urlObj.urlMd5

        if self.dbWrapper is not None:
          result = self.dbWrapper.urlNew([batchItem.urlObj])
          self.logger.debug("urlNew() return result: " + str(result))

      if len(newUrls) > 1:
        for newUrl in newUrls[1:]:
          localBatchItem = copy.deepcopy(batchItem)
          localBatchItem.urlObj.batchId = 0
          localBatchItem.urlObj.status = dc_event.URL.STATUS_NEW
          localBatchItem.urlObj.url = newUrl
          localBatchItem.urlObj.urlMd5 = hashlib.md5(newUrl).hexdigest()
          localBatchItem.urlObj.parentMd5 = batchItem.urlObj.urlMd5
          localBatchItem.urlId = localBatchItem.urlObj.urlMd5
          localBatchItem.urlObj.CDate = str(datetime.datetime.now())
          localBatchItem.urlObj.errorMask = 0
          localBatchItem.urlObj.tagsCount = 0
          localBatchItem.urlObj.tagsMask = 0
          localBatchItem.urlObj.crawled = 0
          localBatchItem.urlObj.processed = 0
          localBatchItem.urlObj.size = 0
          localBatchItem.urlObj.contentType = ""
          localBatchItem.urlObj.rawContentMd5 = ""
          localBatchItem.urlObj.state = dc_event.URL.STATE_ENABLED

          if urlSchema.externalError != APP_CONSTS.ERROR_OK:
            localBatchItem.urlObj.errorMask |= urlSchema.externalError
          self.schemaBatchItems.append(localBatchItem)

        if self.dbWrapper is not None:
          # result = self.dbWrapper.urlNew([elem.urlObj for elem in self.schemaBatchItems])
          for elem in self.schemaBatchItems:
            result = self.dbWrapper.urlNew([elem.urlObj])
            self.logger.debug("urlNew() for urls list return result: " + str(result))
            if int(result) == 0:  # necessary update url
              urlUpdateObj = dc_event.URLUpdate(siteId=elem.urlObj.siteId, urlString=elem.urlObj.url,
                                                normalizeMask=UrlNormalizator.NORM_NONE)
              urlUpdateObj.urlMd5 = elem.urlObj.urlMd5
              urlUpdateObj.parentMd5 = batchItem.urlObj.urlMd5
              urlUpdateObj.batchId = 0  # #batchId
              urlUpdateObj.status = dc_event.URL.STATUS_NEW
              urlUpdateObj.UDate = SQLExpression("NOW()")
              urlUpdateObj.errorMask = 0
              urlUpdateObj.tagsCount = 0
              urlUpdateObj.tagsMask = 0
              urlUpdateObj.crawled = SQLExpression("`Crawled`+1")
              urlUpdateObj.processed = 0
              urlUpdateObj.size = 0
              urlUpdateObj.contentType = ""
              urlUpdateObj.rawContentMd5 = ""
              urlUpdateObj.state = dc_event.URL.STATE_ENABLED

              result = self.dbWrapper.urlUpdate(urlUpdateObj)
              self.logger.debug("urlUpdate() for urls list return result: " + str(result))

      # # Apply 'batch_insert' property
      if urlSchema.batchInsert == UrlSchema.BATCH_INSERT_ALL_NEW_ITEMS:
        self.logger.debug("UrlSchema use 'batch_insert' as 'BATCH_INSERT_ALL_NEW_ITEMS'")
      elif urlSchema.batchInsert == UrlSchema.BATCH_INSERT_ONLY_FIRST_ITEM:
        self.logger.debug("UrlSchema use 'batch_insert' as 'BATCH_INSERT_ONLY_FIRST_ITEM'")
        if len(self.schemaBatchItems) > 0:
          self.schemaBatchItems = self.schemaBatchItems[0:1]
      elif urlSchema.batchInsert == UrlSchema.BATCH_INSERT_NO_ONE_ITEMS:
        self.logger.debug("UrlSchema use 'batch_insert' as 'BATCH_INSERT_NO_ONE_ITEMS'")
        self.schemaBatchItems = []
      else:
        self.logger.error("UrlSchema use 'batch_insert' an unsupported value: " + str(urlSchema.batchInsert))

    return batchItem


  # #updateUrlObjInBatchItem update urlObj's fields in batchitem
  #
  # @param urlObj - reference to batchItem.urlObj
  # @param batchId - batch's id
  def updateUrlObjInBatchItem(self, urlObj):
    if urlObj is not None:
      urlObj.errorMask |= self.errorMask
      self.logger.debug("Set error_mask: %s", str(urlObj.errorMask))
      if self.crawledResource is not None:
        urlObj.httpCode = self.crawledResource.http_code
        self.logger.debug("Set HTTP Code: %s, contentType: %s", str(self.crawledResource.http_code), str(self.crawledResource.content_type))
        if self.crawledResource.content_type is not None and \
        self.crawledResource.content_type != dc_event.URL.CONTENT_TYPE_UNDEFINED:
          urlObj.contentType = self.crawledResource.content_type


  # #updateBatchItem
  #
  def updateBatchItem(self, batchItem):
    # update batchItem properties
    # RSS feed items -> batch items
    if batchItem is not None:
      self.logger.debug("self.processorName: %s", varDump(self.processorName))

      if self.processorName == PCONSTS.PROCESSOR_FEED_PARSER:
        if self.feedItems is not None and len(self.feedItems) > 0:
          self.logger.debug("len(self.feedItems): %s", str(len(self.feedItems)))
          batchItem = self.createBatchItemsFromFeedItems(batchItem)
      elif self.url is not None:
        self.logger.debug("Before: batchItem urlObj errorMask: %s, url ErrorMask: %s", str(batchItem.urlObj.errorMask), str(self.url.errorMask))
        batchItem.urlObj.errorMask |= self.url.errorMask
        batchItem.urlObj.errorMask |= self.errorMask
        self.logger.debug("After: batchItem urlObj errorMask: %s, url ErrorMask: %s" , str(batchItem.urlObj.errorMask), str(self.url.errorMask))
      if isinstance(batchItem, types.ListType):
        self.logger.debug("!!! (2) self.batchItemsExtendUnique()")
        self.batchItemsExtendUnique(self.store_items, batchItem)
      else:
        self.logger.debug("!!! (3) self.batchItemsExtendUnique()")
        self.batchItemsExtendUnique(self.store_items, [batchItem], False)
      if self.feedItems is not None:
        self.logger.debug("self.feedItems: %s, self.store_items: %s", str(len(self.feedItems)),
                          str(len(self.store_items)))
    else:
      self.logger.debug(">>> wrong !!! updateBatchItem, batchItem is None")


  # #createBatchItems
  #
  def createBatchItemsFromFeedItems(self, parentBatchItem):
    self.logger.debug("!!!  createBatchItemsFromFeedItems() enter ... self.crawledResource: " + \
                      varDump(self.crawledResource))
    items = []
    for elem in self.feedItems:
      if self.batch.maxItems > len(items):  # update only in case allowed high value of limits
        urlMd5 = elem["urlMd5"]
        self.logger.debug("URLMD5: %s", str(urlMd5))
        self.logger.debug("value: %s", str(elem))
        siteId = self.batchItem.siteId
        urlObj = elem["urlObj"]

        # serialize content
        elem.pop("urlObj", None)

        # list of objects must be converted to string before json dumps
        dates = ["published_parsed", "updated_parsed"]
        for date in dates:
          if date in elem["entry"] and elem["entry"][date] is not None:
            elem["entry"][date] = strftime("%a, %d %b %Y %H:%M:%S +0000", elem["entry"][date])
        elem["urlMd5"] = urlMd5
        elem["entry"] = dict(elem["entry"])
        # store raw content on disk
        saveBatchItemUrlId = self.batchItem.urlId
        try:
          self.crawledResource = CrawledResource()  # pylint: disable=W0201
          self.crawledResource.binary_content = json.dumps(elem, ensure_ascii=False)
          self.batchItem.urlId = urlMd5
          self.getDir()
          self.writeData(self.batchItem)
          self.batchItem.urlId = saveBatchItemUrlId
        except Exception, err:
          ExceptionLog.handler(self.logger, err, "Can't save object on disk. Reason:", (), \
                               {ExceptionLog.LEVEL_NAME_ERROR:ExceptionLog.LEVEL_VALUE_DEBUG})
          self.batchItem.urlId = saveBatchItemUrlId
          continue

        urlObj.contentMask = dc_event.URL.CONTENT_STORED_ON_DISK
        if parentBatchItem is not None:
          batchItem = copy.deepcopy(parentBatchItem)
          batchItem.siteId = siteId
          batchItem.urlId = urlMd5
          batchItem.urlObj = urlObj
        else:
          batchItem = dc_event.BatchItem(siteId, urlMd5, urlObj)
        batchItem.urlObj.urlPut = None
        items.append(batchItem)

    return items


  # # load cli arguments
  #
  # @param - None
  # @return - None
  def loadAppParams(self):
    # # max time execution
    if self.pargs.maxtime:
      self.maxTimeCli = int(self.pargs.maxtime)

    # input data file
    if self.pargs.inputFile:
      self.inputFile = str(self.pargs.inputFile)

    # output data file
    if self.pargs.outputFile:
      self.outputFile = str(self.pargs.outputFile)


  # #load config from file
  # load from cli argument or default config file
  def loadConfig(self):
    try:
      if self.pargs.config:
        self.config = ConfigParser.ConfigParser()
        self.config.optionxform = str
        self.config.read(self.pargs.config)
      else:
        raise Exception(MSG_ERROR_EMPTY_CONFIG_FILE_NAME)
    except Exception, err:
      raise Exception(MSG_ERROR_LOAD_CONFIG + ' ' + str(err))


  # #loadKeyValueDB load key-value db
  #
  #
  def loadKeyValueDB(self):
    try:
      className = self.__class__.__name__
      self.kvDbDir = self.config.get(className, self.DB_DATA_DIR)
    except Exception, err:
      ExceptionLog.handler(self.logger, err, "Error load KVDB config option: %s", self.DB_DATA_DIR)
      raise


  # #load logging
  # load logging configuration (log file, log level, filters)
  #
  def loadLogConfigFile(self):
    try:
      log_conf_file = self.config.get("Application", "log")
      logging.config.fileConfig(log_conf_file)
      self.logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)  # Utils.MPLogger().getLogger()
    except:
      print MSG_ERROR_LOAD_LOG_CONFIG_FILE
      raise

#   # #load project/site-specific logging
#   # load logging configuration (log file, log level, filters) for project-specific file name
#   #
#   def setLogConfigFileProject(self, mpLogger, projectId, propertyStr):  # pylint: disable=W0622
#     if propertyStr is not None and propertyStr != '':
#       try:
#         propertyObj = json.loads(propertyStr)
#         if 'suffix' in propertyObj:
#           suffix = propertyObj['suffix'].replace('%PROJECT_ID%', projectId)
#           self.logger = mpLogger.getLogger(fileNameSuffix=suffix)
#         else:
#           self.logger.debug("Suffix field not found for project %s in property: %s", str(projectId), str(propertyObj))
#       except Exception, err:
#         self.logger.error("Error set project-specific logger: %s", str(err))
#     else:
#       self.logger.debug("Wrong or empty file name suffix, project %s logger not set: %s", str(projectId), \
#                         str(propertyObj))
#
#
#   # #set default logging
#   #
#   def setLogConfigFileDefault(self, mpLogger):
#     try:
#       self.logger = mpLogger.getLogger(restore=True)
#     except Exception, err:
#       self.logger.error("Error set default logger: %s", str(err))


  # #load mandatory options
  # load mandatory options
  #
  def loadOptions(self):
    try:
      self.raw_data_dir = self.config.get(self.__class__.__name__, "raw_data_dir")
      self.defaultHeaderFile = self.config.get(self.__class__.__name__, "headers_file")
      self.defaultIcrCrawlTime = self.config.getfloat(self.__class__.__name__, "default_icr_crawl_time")
      self.headerFileDir = self.config.get(self.__class__.__name__, "header_file_dir")
      self.robotsFileDir = self.config.get(self.__class__.__name__, "robots_file_dir")
      self.dbTaskIni = self.config.get(self.__class__.__name__, "db-task_ini")
      self.urlSchemaDataDir = self.config.get(self.__class__.__name__, self.URL_SCHEMA_DIR)
      # Read urls xpath list from file as json and make loads
      self.urlsXpathList = \
      json.loads(open(self.config.get(self.__class__.__name__, self.URLS_XPATH_LIST_FILE), 'r').read())

      if self.config.has_option(self.__class__.__name__, "useZeroSiteIdSiteNotExists"):
        try:
          self.useZeroSiteIdSiteNotExists = \
          bool(int(self.config.get(self.__class__.__name__, "useZeroSiteIdSiteNotExists")))
        except Exception:
          self.useZeroSiteIdSiteNotExists = False

      # does call collectAddtionalProp and collectProperties methods
      self.max_fetch_time = self.config.getint(self.__class__.__name__, "max_fetch_time")
      # keep old resources on disk
      if self.config.has_option(self.__class__.__name__, "keep_old_resources"):
        self.keep_old_resources = self.config.getboolean(self.__class__.__name__, "keep_old_resources")
      self.collect_additional_prop = self.config.getboolean(self.__class__.__name__, "collect_additional_prop")

      if self.config.has_option(self.__class__.__name__, "maxExecutionTimeExitUseAbort"):
        self.maxExecutionTimeExitUseAbort = self.config.getboolean(self.__class__.__name__, "maxExecutionTimeExitUseAbort")

    except Exception, err:
      ExceptionLog.handler(self.logger, err, "Error load config options:")
      raise Exception('CRAWLER FATAL INITIALIZATION INI ERROR: ' + str(err))


  # #fillItemsFromIterations
  #
  # Optionally resets and fills the list of the BatchItems with the new instances initialized from the URLs objects list
  # @param urlObjects the list of the URL objects
  # @param siteObject the site object
  # @param reset the Boolean defines reset the items list before fill, True means reset
  def fillItemsFromIterations(self, urlObjects=None, siteObject=None, reset=True):
    if reset:
      self.collectURLsItems = []

    if urlObjects is not None:
      if siteObject is None:
        siteObjectLocal = self.batchItem.siteObj
        siteIdLocal = self.batchItem.siteId
      else:
        siteObjectLocal = siteObject
        siteIdLocal = siteObject.id

      for urlObject in urlObjects:
        self.logger.debug("Create new batch item, URLMd5: %s, siteId: %s, url: %s", urlObject.urlMd5, siteIdLocal, urlObject.url)
        batchItem = dc_event.BatchItem(siteId=siteIdLocal, urlId=urlObject.urlMd5, urlObj=urlObject, siteObj=siteObjectLocal)
        batchItem.properties = self.batchItem.properties
        batchItem.parentContentType = self.crawledResource.content_type
        self.logger.debug("Set parentContentType = %s", str(batchItem.parentContentType))
        self.collectURLsItems.append(batchItem)


  # #batchItemsExtendUnique
  #
  # Fills the batch items list with unique items from another list
  # @param destinationBatchItems the destination BatchItem objects list
  # @param destinationBatchItems the source BatchItem objects list
  def batchItemsExtendUnique(self, destinationBatchItems, sourceBatchItems, lookIncomingBatch=True, processorName=None):
    self.logger.debug("!!! Enter processorName: %s", str(processorName))
    for srcBatchItem in sourceBatchItems:
      inList = False
      if lookIncomingBatch:
        for dstBatchItem in self.batch.items:
          if srcBatchItem.siteId == dstBatchItem.siteId and srcBatchItem.urlId == dstBatchItem.urlId:
            self.logger.debug("batchItemsExtendUnique baseItems duplicate " + srcBatchItem.urlId + " " +
                              dstBatchItem.urlId)
            inList = True
            break
      if not inList:
        for dstBatchItem in destinationBatchItems:
          if srcBatchItem.siteId == dstBatchItem.siteId and srcBatchItem.urlId == dstBatchItem.urlId:
            self.logger.debug("batchItemsExtendUnique duplicate " + srcBatchItem.urlId + " " + dstBatchItem.urlId)
            inList = True
            break
      if not inList:
        self.logger.debug("batchItemsExtendUnique added, url: %s", str(srcBatchItem.urlObj.url))
        self.logger.debug("batchItemsExtendUnique added, urlId: %s", srcBatchItem.urlId)
        self.logger.debug("batchItemsExtendUnique srcBatchItem.urlObj.type1 = %s", str(srcBatchItem.urlObj.type))

        #TODO: add property field to allow set type 0 for item from RSS feed
#        chained = APP_CONSTS.URL_CHAIN in self.siteProperties and self.siteProperties[APP_CONSTS.URL_CHAIN] is not None and\
#                  len(self.siteProperties[APP_CONSTS.URL_CHAIN]) > 0
#        if self.processorName in [PCONSTS.PROCESSOR_FEED_PARSER, PCONSTS.PROCESSOR_RSS] and not chained:
#          srcBatchItem.urlObj.type = dc_event.URL.TYPE_SINGLE
#        else:
#          srcBatchItem.urlObj.type = dc_event.URL.TYPE_REGULAR

        srcBatchItem.urlObj.type = URLProcess.getTypeByChainAndProcessorName(self.siteProperties, processorName)
        self.logger.debug("batchItemsExtendUnique srcBatchItem.urlObj.type2 = %s", str(srcBatchItem.urlObj.type))

        # check max allowed limits of resources
        if self.batch.maxItems is not None and int(self.batch.maxItems) <= len(destinationBatchItems):
          destinationBatchItems = destinationBatchItems[0: self.batch.maxItems]
          destinationBatchItems[-1].urlObj.errorMask |= APP_CONSTS.ERROR_MAX_ITEMS
          self.logger.debug("Set ErrorMask^ %s", str(destinationBatchItems[-1].urlObj.errorMask))
          break
        else:
          if processorName is not None:
            srcBatchItem.processorName = processorName
          destinationBatchItems.append(srcBatchItem)


  # #updateCollectedURLs
  #
  def updateCollectedURLs(self):
    self.dbWrapper.collectedURLsRecalculating(Utils.autoFillSiteId(self.batchItem.siteId, self.logger))


  # #setChainId
  #
  # @param batchItem incoming batch item
  def setChainId(self, batchItem):
    self.logger.debug("!!! ENTER !!!")
    if batchItem.urlObj.chainId is None and self.urlChain is not None:
      batchItem.urlObj.chainId = self.chainIndex
      self.chainIndex += 1
      self.logger.debug("!!! batchItem.urlObj.chainId = %s", str(batchItem.urlObj.chainId))
      self.logger.debug("!!! batchItem.urlObj.url: %s", str(batchItem.urlObj.url))
      self.logger.debug("!!! batchItem.urlObj.type = %s", str(batchItem.urlObj.type))

    self.logger.debug("!!! LEAVE !!!")


  # #fillItemsFromIterationsWithChain
  #
  # @param urlObjects the list of the URL objects
  def fillItemsFromIterationsWithChain(self, urlObjects, batchItem):

    self.logger.debug("!!! ENTER !!! urlObjects: %s", str(urlObjects))
    if urlObjects is not None and len(urlObjects) > 0:
      if batchItem.urlObj.chainId not in self.chainDict:
        self.chainDict[batchItem.urlObj.chainId] = {}
        self.chainDict[batchItem.urlObj.chainId]["batchItem"] = batchItem
        self.chainDict[batchItem.urlObj.chainId]["chainUrlMD5List"] = []
      for urlObj in urlObjects:
        urlObj.chainId = batchItem.urlObj.chainId
        self.logger.debug("!!! batchItem.urlObj.type = %s, urlObj.type = %s, url: %s", str(batchItem.urlObj.type), str(urlObj.type), str(urlObj.url))

        self.logger.debug("!!! hasattr(batchItem, 'processorName') = %s", str(hasattr(batchItem, 'processorName')))
        if hasattr(batchItem, 'processorName'):
          self.logger.debug("!!! batchItem.processorName = '%s'", str(batchItem.processorName))

        self.logger.debug("!!! hasattr(batchItem, 'parentContentType') = %s", str(hasattr(batchItem, 'parentContentType')))
        if hasattr(batchItem, 'parentContentType'):
          self.logger.debug("!!! batchItem.parentContentType = '%s'", str(batchItem.parentContentType))

      self.fillItemsFromIterations(urlObjects, None, False)
    self.logger.debug("!!! LEAVE !!!")


  # #fillChainUrlMD5List
  #
  # @param batchItem incoming batch item
  def fillChainUrlMD5List(self, batchItem):
    self.logger.debug("!!! ENTER !!! chainId = %s, urlMd5 = %s, url = %s", str(batchItem.urlObj.chainId), str(batchItem.urlObj.urlMd5),
                      str(batchItem.urlObj.url))

    if batchItem.urlObj.chainId is not None and batchItem.urlObj.chainId in self.chainDict:
      localChainElem = self.chainDict[batchItem.urlObj.chainId]
      if batchItem.urlObj.urlMd5 != localChainElem["batchItem"].urlObj.urlMd5 and \
        batchItem.urlObj.urlMd5 not in localChainElem["chainUrlMD5List"]:
        localChainElem["chainUrlMD5List"].append(batchItem.urlObj.urlMd5)
        self.logger.debug("!!! Added urlMd5 = %s, url = %s", str(batchItem.urlObj.urlMd5), str(batchItem.urlObj.url))
      else:
        self.logger.debug("!!! Skipped urlMd5 = %s, url = %s", str(batchItem.urlObj.urlMd5), str(batchItem.urlObj.url))
    else:
      self.logger.debug("!!! Skipped chainId = %s, url = %s", str(batchItem.urlObj.chainId), str(batchItem.urlObj.url))

    self.logger.debug("!!! LEAVE !!! self.chainDict: %s", varDump(self.chainDict, stringifyType=0))


  # #saveChainStorageData
  #
  def saveChainStorageData(self):
#     urlPutList = []
    for localChainKay in self.chainDict:
      localChainElem = self.chainDict[localChainKay]
      saveBuf = '\n'.join(localChainElem["chainUrlMD5List"])
      saveBuf = saveBuf.strip()
      putDict = {"data": base64.b64encode(saveBuf)}
      urlPutObj = dc_event.URLPut(localChainElem["batchItem"].siteId, localChainElem["batchItem"].urlObj.urlMd5, \
                                  dc_event.Content.CONTENT_CHAIN_PARTS, putDict)
#       urlPutList.append(urlPutObj)
#       if len(urlPutList) > 0:
      self.dbWrapper.putURLContent([urlPutObj])

      self.urlProcess.siteId = localChainElem["batchItem"].siteId
      self.urlProcess.updateURLStatus(localChainElem["batchItem"].urlId)
      self.logger.debug("!!! SaveChainStorageData for urlId: %s, saved: %s", str(localChainElem["batchItem"].urlId), str(localChainElem["chainUrlMD5List"]))


  # #Get pubdate accord to 'PDATE_SOURCE_MASK'
  #
  # @param siteProperties - site properites instance
  # @param crawledResource - CrawledResource instance
  # @param urlObj - class URL instance
  # @return datetime as string in iso format
  def getPubdateUseSourceMask(self, siteProperties, crawledResource, urlObj):
    # variable to result
    ret = None
    try:
      if siteProperties is not None and crawledResource is not None:
        # default values
        pdateSourceMask = APP_CONSTS.PDATE_SOURCES_MASK_BIT_DEFAULT
        pdateSourceMaskOverwrite = APP_CONSTS.PDATE_SOURCES_MASK_OVERWRITE_DEFAULT
        pdateSourceMaskFormats = None

        # get value 'PDATE_SOURCES_MASK' from site properties
        if APP_CONSTS.PDATE_SOURCES_MASK_PROP_NAME in siteProperties:
          self.logger.debug("Property '%s' has input value: %s",
                            str(APP_CONSTS.PDATE_SOURCES_MASK_PROP_NAME),
                            str(siteProperties[APP_CONSTS.PDATE_SOURCES_MASK_PROP_NAME]))

          pDateSourceMaskObject = PDateSourceMask(properties=siteProperties[APP_CONSTS.PDATE_SOURCES_MASK_PROP_NAME], url=urlObj.url, log=self.logger)
          self.logger.debug("Extracted pdateSourceMask = %s, pdateSourceMaskOverwrite = %s, formats = %s",
                            str(pDateSourceMaskObject.pdateSourceMask),
                            str(pDateSourceMaskObject.pdateSourceMaskOverwrite),
                            str(pDateSourceMaskObject.formats))

          if isinstance(pDateSourceMaskObject.pdateSourceMask, int):
            pdateSourceMask = pDateSourceMaskObject.pdateSourceMask

          if isinstance(pDateSourceMaskObject.pdateSourceMaskOverwrite, int):
            pdateSourceMaskOverwrite = pDateSourceMaskObject.pdateSourceMaskOverwrite

          if isinstance(pDateSourceMaskObject.formats, list):
            self.urlProcess.stagesFormatList = pdateSourceMaskFormats = pDateSourceMaskObject.formats

        # get value 'PDATE_SOURCES_MASK_OVERWRITE' from site properties
        if APP_CONSTS.PDATE_SOURCES_MASK_OVERWRITE_PROP_NAME in siteProperties:
          self.logger.debug("Property '%s' has input value: %s",
                            str(APP_CONSTS.PDATE_SOURCES_MASK_OVERWRITE_PROP_NAME),
                            str(siteProperties[APP_CONSTS.PDATE_SOURCES_MASK_OVERWRITE_PROP_NAME]))

          pDateSourceMaskObject = PDateSourceMask(properties=siteProperties[APP_CONSTS.PDATE_SOURCES_MASK_OVERWRITE_PROP_NAME],
                                                  url=urlObj.url,
                                                  log=self.logger)

          self.logger.debug("Extracted mask = %s", str(pDateSourceMaskObject.pdateSourceMask))

          if isinstance(pDateSourceMaskObject.pdateSourceMask, int):
            pdateSourceMaskOverwrite = pDateSourceMaskObject.pdateSourceMask

        self.logger.debug('pdateSourceMask = %s, pdateSourceMaskOverwrite = %s',
                          str(pdateSourceMask), str(pdateSourceMaskOverwrite))
        # self.logger.debug('crawledResource.response_header = ' + str(crawledResource.response_header))
        self.logger.debug('crawledResource.last_modified = ' + str(crawledResource.last_modified))

        # Extracted from URL name (not implemented)
        if pdateSourceMask & APP_CONSTS.PDATE_SOURCES_MASK_URL_NAME:  # reserved
          if pdateSourceMaskOverwrite & APP_CONSTS.PDATE_SOURCES_MASK_URL_NAME:  # ON
            pass
          else:  # OFF
            pass

        # URL object the "pdate" field (supposed was got from the RSS feed)
        if pdateSourceMask & APP_CONSTS.PDATE_SOURCES_MASK_RSS_FEED:
          if (pdateSourceMaskOverwrite & APP_CONSTS.PDATE_SOURCES_MASK_RSS_FEED and ret is None) or \
          not pdateSourceMaskOverwrite & APP_CONSTS.PDATE_SOURCES_MASK_RSS_FEED:
            try:
              ret = urlObj.pDate

            except TypeError:
              self.logger.debug("Unsupported date format: <%s>", Utils.varDump(urlObj.pDate))

            self.logger.debug('pubdate from rss feed: ' + str(ret))


        # URL object "Date" field (supposed was got from the web server's HTTP response header)
        if pdateSourceMask & APP_CONSTS.PDATE_SOURCES_MASK_HTTP_DATE and 'date' in crawledResource.response_header:
          if (pdateSourceMaskOverwrite & APP_CONSTS.PDATE_SOURCES_MASK_HTTP_DATE and ret is None) or \
          not pdateSourceMaskOverwrite & APP_CONSTS.PDATE_SOURCES_MASK_HTTP_DATE:
            value = HTTPHeaders.getValue(crawledResource.response_header, 'date', default='')
            dt = PDateExtractor.extract(dataString=value,
                                        siteProperties=siteProperties,
                                        url=urlObj.url,
                                        stage=PDateFormat.STAGE_HTTP_DATE,
                                        log=self.logger,
                                        stagesFormatList=pdateSourceMaskFormats)
            if dt is not None:
              ret = dt.strftime('%Y-%m-%d %H:%M:%S')
              self.logger.debug('pubdate from http header: ' + str(ret))

        # URL object "lastModified" field (supposed was got from the web server's HTTP response header)
        if pdateSourceMask & APP_CONSTS.PDATE_SOURCES_MASK_HTTP_LAST_MODIFIED:
          if (pdateSourceMaskOverwrite & APP_CONSTS.PDATE_SOURCES_MASK_HTTP_LAST_MODIFIED and ret is None) or \
          not pdateSourceMaskOverwrite & APP_CONSTS.PDATE_SOURCES_MASK_HTTP_LAST_MODIFIED:
            dt = None
            if 'last-modified' in crawledResource.response_header:
              value = HTTPHeaders.getValue(crawledResource.response_header, 'last-modified', default='')
              dt = PDateExtractor.extract(dataString=value,
                                          siteProperties=siteProperties,
                                          url=urlObj.url,
                                          stage=PDateFormat.STAGE_HTTP_LAST_MODIFIED,
                                          log=self.logger,
                                          stagesFormatList=pdateSourceMaskFormats)
            else:
              dt = PDateExtractor.extract(dataString=crawledResource.last_modified,
                                          siteProperties=siteProperties,
                                          url=urlObj.url,
                                          stage=PDateFormat.STAGE_HTTP_LAST_MODIFIED,
                                          log=self.logger,
                                          stagesFormatList=pdateSourceMaskFormats)

            if dt is not None:
              ret = dt.strftime('%Y-%m-%d %H:%M:%S')
              self.logger.debug('pubdate from last modified: ' + str(ret))

    except Exception, err:
      self.logger.error('Error: ' + str(err))

    return ret


  # # Extract base url
  #
  # @param htmlContent - batct item
  # @param default - default value of url
  # @return base url
  def extractBaseUrl(self, htmlContent, default):
    # variable for result
    ret = default

    try:
      if isinstance(htmlContent, basestring):
        urlsList = re.findall(pattern=self.SEARCH_BASE_URL_PATTERN, string=htmlContent, flags=re.M + re.I + re.U)
        if len(urlsList) > 0:
          ret = urlsList[0]

    except Exception, err:
      self.logger.error(MSG_ERROR_EXTRACT_BASE_URL, str(err))

    return ret


  # #Timer alarm signal handler
  #
  # @param signum
  # @param frame
  def signalHandlerTimer(self, signum, frame):
    del frame
    self.maxExecutionTimeReached = True
    self.logger.error("Signal %s - timer trapped! (%s)", str(signum), str(self.maxExecutionTimeValue))
    self.exit_code = EXIT_FAILURE

    self.logger.debug("maxExecutionTimeExitUseAbort = %s", str(self.maxExecutionTimeExitUseAbort))
    if self.maxExecutionTimeExitUseAbort:
      self.logger.info("Hard exit!")
      os._exit(self.exit_code)  # pyLint disable=protected-access


  # # Check is set flag max execution time reached
  #
  # @param - None
  # @return True if set or False otherwise
  def isAbortedByTTL(self):
    return self.maxExecutionTimeReached


  # # create dbtask wrapper instance
  #
  # @param configName - dbtask ini file
  # @return instance of DBTasksWrapper class
  def __createDBTasksWrapper(self, configName):
    # variable for result
    dbTasksWrapper = None
    try:
      if configName == "":
        raise Exception(MSG_ERROR_EMPTY_CONFIG_FILE_NAME)

      config = ConfigParser.ConfigParser()
      config.optionxform = str

      readOk = config.read(configName)

      if len(readOk) == 0:
        raise Exception(MSG_ERROR_WRONG_CONFIG_FILE_NAME % configName)

      dbTasksWrapper = DBTasksWrapper(config)

    except Exception, err:
      raise Exception(MSG_ERROR_LOAD_APP_CONFIG % str(err))

    return dbTasksWrapper
