# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
FetcherPost Class content main functional of post process after fetching procedure.

@package: dc_crawler
@file FetcherPost.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2018 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import os
import re
import json
import copy
import datetime
from dc_crawler.Fetcher import Response
from scrapy.selector import Selector
import app.Utils as Utils


class FetcherPost(object):
  # # Constants error messages used in class
  MSG_ERROR_FETCHER_POST_PROCESSING_SKIPPED = "FetcherPost processing skipped. Reason: %s"
  MSG_ERROR_PROCESSING_ID_COLLECT_SKIPPED = "Processing ID collect skipped. Reason: %s"

  MSG_ERROR_WRONG_PROPERTY = "Wrong type of the input property: %s"
  MSG_ERROR_WRONG_PARAMETERS = "Wrong type of the input parameter: %s"
  MSG_ERROR_NOT_FOUND_FILE_NAME = "Not found file name option for the result pages file."
  MSG_ERROR_GET_DOMAIN = "Get domain failed. Possible you used wrong protocol."

  # # Constants used in class
  PROPERTY_OPTION_ID_COLLECT_NAME = 'fb_pages_id_collect'
  PROPERTY_OPTION_DATA_SCRAPING_NAME = 'fb_pages_data_scraping'
  PROPERTY_OPTION_PAGES_FILE = 'pages_file'

  DATATIME_FORMAT = "%Y-%m-%d %H:%M:%S"
  PATTERNS_FOR_ID_COLLECT = ['fb://page/(\d+)', 'fb://page/\?id=(\d+)']
  ERROR_MASK_NO_ERRORS = 0

  # # Initialization
  #
  # @param properties - properties dictionary
  # @param log - logger instance
  # @param data - macro collect protocoll data
  def __init__(self, properties, log, data=None):
    self.properties = properties
    self.logger = log
    self.data = data


  # # Check is active option
  #
  # @param properties - properties dictionary
  # @param optionName - option name
  # @return True if is active or False otherwise
  @staticmethod
  def isActive(properties, optionName):
    # variable for result
    ret = False

    if isinstance(properties, dict) and optionName in properties and int(properties[optionName]) > 0:
      ret = True
    
    return ret


  # # Check is active option 'ID collect'
  #
  # @param properties - properties dictionary
  # @return True if is active or False otherwise
  @staticmethod
  def isActiveIDCollect(properties):
    return FetcherPost.isActive(properties, FetcherPost.PROPERTY_OPTION_ID_COLLECT_NAME)


  # # Get option 'ID collect' name
  #
  # @param - None
  # @return option name
  @staticmethod
  def getOptionIDCollectName():
    return FetcherPost.PROPERTY_OPTION_ID_COLLECT_NAME


  # # Check is active option 'Data scraping'
  #
  # @param properties - properties dictionary
  # @return True if is active or False otherwise
  @staticmethod
  def isActiveDataScraping(properties):
    return FetcherPost.isActive(properties, FetcherPost.PROPERTY_OPTION_DATA_SCRAPING_NAME)


  # # Get option 'Data scraping' name
  #
  # @param - None
  # @return option name
  @staticmethod
  def getOptionDataScrapingName():
    return FetcherPost.PROPERTY_OPTION_DATA_SCRAPING_NAME


  # # Process execution method
  #
  # @param response - Response class instance
  # @return - None
  def process(self, response):
    try:
      if not isinstance(self.properties, dict):
        raise Exception(self.MSG_ERROR_WRONG_PROPERTY % str(type(self.properties)))

      if not isinstance(response, Response):
        raise Exception(self.MSG_ERROR_WRONG_PARAMETERS % str(type(self.properties)))

      # execution some post procedures if necessary
      if self.isActiveIDCollect(self.properties):
        self.logger.debug("Found active option '%s'.", self.getOptionIDCollectName())
        self.processIdCollect(response)

      if self.isActiveDataScraping(self.properties):
        self.logger.debug("Found active option '%s'.", self.getOptionDataScrapingName())
        self.processDataScraping(response)

    except Exception, err:
      self.logger.error(self.MSG_ERROR_FETCHER_POST_PROCESSING_SKIPPED, str(err))


  # # Process ID collect
  #
  # @param response - Response class instance
  # @return - None
  def processIdCollect(self, response):
    try:
      if self.PROPERTY_OPTION_PAGES_FILE not in self.properties or self.properties[self.PROPERTY_OPTION_PAGES_FILE] == "":
        raise Exception(self.MSG_ERROR_NOT_FOUND_FILE_NAME)

      if isinstance(response.page_source, basestring):
        for pattern in self.PATTERNS_FOR_ID_COLLECT:
          matches = re.findall(pattern=pattern, string=response.page_source, flags=re.U + re.I + re.M)
          if len(matches) > 0:
            self.logger.debug("Found Id '%s' by pattern '%s'", str(matches[0]), str(pattern))
            self.saveToFile(matches[0], response.url, self.properties[self.PROPERTY_OPTION_PAGES_FILE])
            break
          
    except Exception, err:
      self.logger.error(self.MSG_ERROR_PROCESSING_ID_COLLECT_SKIPPED, str(err))


  def getDomain(self):
    # variable for result
    ret = None
    self.logger.debug("self.data: %s", Utils.varDump(self.data, stringifyType=0))

    if isinstance(self.data, dict) and 'options' in self.data and isinstance(self.data['options'], dict) and 'currentDataMd5' in self.data['options']:
      currentDataMd5 = self.data['options']['currentDataMd5']
      self.logger.debug("currentDataMd5: %s", str(currentDataMd5))
      if 'data' in self.data and currentDataMd5 in self.data['data'] and isinstance(self.data['data'][currentDataMd5], dict) and \
        'domain' in self.data['data'][currentDataMd5]:
        ret = self.data['data'][currentDataMd5]['domain']
        self.logger.debug("domain: %s", str(ret))

    return ret


  ## save ID to file
  #
  # @param id - ID for saving
  # @param url - url of media page
  # @param fileName - file name
  def saveToFile(self, id, url, fileName):
    domain = self.getDomain()
    if domain is None:
      raise Exception(self.MSG_ERROR_GET_DOMAIN)
    else:
      msg = "%s, %s, %s\n" % (str(domain), str(url), str(id))
      mode = 'a' if os.path.isfile(fileName) else 'w'
      with open(fileName, mode) as f:
        f.write(msg)


  # # Load json from data
  #
  # @param data - data string
  # @param log - logger instance
  # @return json
  @staticmethod
  def loadJsonFromData(data, log=None):
    # variable for result
    ret = {}

    if isinstance(data, basestring):
      pos = data.find('{')
      if pos != -1:
        if log is not None:
          log.debug("data[%s:] = %s", str(pos), str(data[pos:]))

        ret = Utils.jsonLoadsSafe(jsonString=data[pos:], default=ret, log=log)

    return ret


  def getSocialData(self, obj, pageId):
    # variable for result
    ret = None

    if isinstance(obj, dict):
      if 'feedbacktarget' in obj and 'actorid' in obj['feedbacktarget'] and obj['feedbacktarget']['actorid'] == str(pageId):
        ret = obj['feedbacktarget']
      else:
        for key, value in obj.items():
          ret = self.getSocialData(value, pageId)
          if ret is not None:
            break

    elif isinstance(obj, list):
      for elem in obj:
        ret = self.getSocialData(elem, pageId)
        if ret is not None:
          break

    return ret
  

  # # Extract social data
  #
  # @param dataDict - data dictionary
  # @return dictionary of the social data
  def extractSocialData(self, dataDict):
    # variable for result
    socialDataDict = {}

    if 'payload' in dataDict:
      author = Selector(text=dataDict['payload']).xpath("//*[contains(@class, '_6lz') and contains(@class, '_6mb') and contains(@class, 'ellipsis')]/text()").extract()
      self.logger.debug("author: %s", str(author))
      socialDataDict['author'] = author[0] if isinstance(author, list) and len(author) > 0 else None

      title = Selector(text=dataDict['payload']).xpath("//*[contains(@class, 'userContent')]//p/text()").extract()
      self.logger.debug("title: %s", str(title))
      socialDataDict['title'] = title[0] if isinstance(title, list) and len(title) > 0 else ''

      utime = Selector(text=dataDict['payload']).xpath("//abbr[@data-utime]/@data-utime").extract()
      self.logger.debug("utime: %s", str(utime))
      socialDataDict['date'] = utime[0] if isinstance(utime, list) and len(utime) > 0 else None

      udate = datetime.datetime.fromtimestamp(int(utime[0])).strftime(self.DATATIME_FORMAT) if isinstance(utime, list) and len(utime) > 0 else None
      self.logger.debug("udate: %s", str(udate))

      dataBt = Selector(text=dataDict['payload']).xpath("//div[@data-bt]/@data-bt").extract()
      self.logger.debug("dataBt: %s, type: %s", str(dataBt), str(type(dataBt)))

      if len(dataBt) > 0:
        dataBtDict = Utils.jsonLoadsSafe(jsonString=dataBt[0], default={}, log=self.logger)
        self.logger.debug("dataBtDict: %s, type: %s", str(dataBtDict), str(type(dataBtDict)))

        queryData = Utils.jsonLoadsSafe(jsonString=dataBtDict['query_data']['data'], default={}, log=self.logger)

        self.logger.debug("queryData: %s", str(queryData))
        self.logger.debug("page_id: %s", str(queryData['page_id']))

        if 'jsmods' in dataDict and 'require' in dataDict['jsmods']:
          self.logger.debug("require: %s", str(dataDict['jsmods']['require']))

          resDict = self.getSocialData(dataDict['jsmods']['require'], queryData['page_id'])
          self.logger.debug("resDict: %s", str(resDict))

          if 'likecount' in resDict:
            self.logger.debug("like: %s", str(resDict['likecount']))
            socialDataDict['like'] = resDict['likecount']

          if 'sharecount' in resDict:
            self.logger.debug("repost: %s", str(resDict['sharecount']))
            socialDataDict['repost'] = resDict['sharecount']

      else:
        socialDataDict['like'] = None
        socialDataDict['repost'] = None

    return socialDataDict


  # # Process data scraping
  #
  # @param response - Response class instance
  # @return - None
  def processDataScraping(self, response):

#     try:
#       import jsonpickle
# #       self.logger.info("response: %s, type = %s", jsonpickle.dumps(response), str(type(response)))
#
#     except Exception, err:
#       self.logger.error(str(err))

    dataDict = self.loadJsonFromData(response.unicode_content, self.logger)
    self.logger.debug("dataDict: %s, type: %s", Utils.varDump(dataDict), str(type(dataDict)))

    socialDataDict = self.extractSocialData(dataDict)
    self.logger.debug("socialDataDict: %s", Utils.varDump(socialDataDict, stringifyType=0))

    self.logger.debug("self.data: %s, type = %s", str(self.data), str(type(self.data)))

    if isinstance(self.data, dict) and 'data' in self.data and isinstance(self.data['data'], dict):
      # data=[]
      for urlMd5 in self.data['data'].keys():
        # data.append({urlMd5:socialDataDict})l
        if 'data' in self.data['data'][urlMd5]:
          data = copy.deepcopy(self.data)
          # for apply filter stage
          if socialDataDict['title'] is not None and socialDataDict['title'] != "":
            socialDataDict['title'] += (' ' + data['data'][urlMd5]['source_url'])
          # set social data
          data['data'][urlMd5]['data'] = [socialDataDict]
          data['data'][urlMd5]['errorMask'] = self.ERROR_MASK_NO_ERRORS

      response.unicode_content = json.dumps([data], ensure_ascii=False, encoding='utf-8')

