"""
HCE project,  Python bindings, Distributed Tasks Manager application.
HTTPCookieResolver Class content main functional for collect and resolve cookies

@package: dc_crawler
@file HTTPCookieResolver.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2018 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import os
import re
import datetime
import urlparse

import app.Utils as Utils
from app.DateTimeType import DateTimeType
from app.Utils import varDump
from app.LFSDataStorage import LFSDataStorage

logger = Utils.MPLogger().getLogger()

class HTTPCookieResolver(object):
  # Constants names of usage stages
  STAGE_REGULAR = 1
  STAGE_REDIRECT = 2
  STAGE_ROBOTS = 4
  STAGE_RSS = 8
  # Default value of stage
  STAGE_DEFAULT = STAGE_REGULAR | STAGE_REDIRECT | STAGE_ROBOTS | STAGE_RSS
  DOMAIN_DEFAULT = '.*'

  DOMAIN_OPTION_STAGE_NAME = 'stage'
  DOMAIN_OPTION_COOKIE_NAME = 'cookie'

  # Default value of key for host names
  DEFAULT_HOST_NAME_KEY = ''

  ZERO_SITE_ID = '0'
  COOKIES_VALUE_DELIMITER = '; '
  REQUEST_COOKIE_HEADER_NAME = 'Cookie'

  # Constants error message
  ERROR_BAD_TYPE_INPUT_PROPERTY = "Bad type (%s) of input property: %s"
  ERROR_BAD_TYPE_FOUND_PROPERTY = "Bad type (%s) of found property: %s"
  ERROR_INPUT_PROPERTY = "Input wrong properties: %s"
  ERROR_INITIALIZATION = "Initialization class '%s' was failed. Error: '%s'"


  # # Internal class for cookies dictionary
  class CookiesDictionary(object):
#     # Constants used in class
#     OPTION_EXPIRES = 'expires'
#     OPTION_DOMAIN = 'domain'
#     OPTION_PATH = 'path'
#     OPTION_NAME = 'name'
#     OPTION_VALUE = 'value'


    # # Initialization
    #
    # @param dictObj - dictionary object
    def __init__(self, name=None, value=None, domain=None, path=None, expires=None):
      self.name = name
      self.value = value
      self.domain = domain
      self.path = path
      self.expires = expires


    ## Initialization from dictionary
    #
    # @param dictObj - dictionary object
    # @return - None
    def loadFromDict(self, dictObj):
      if isinstance(dictObj, dict):
        for attrName in self.__dict__.keys():
          if attrName in dictObj.keys():
            setattr(self, attrName, dictObj[attrName])


    # # Convert to dictionary
    #
    # @param - None
    # @return converted to dictionary object values
    def toDict(self):
      return self.__dict__


    # # get data
    #
    # @param - None
    # @return cookie data as string
    def getData(self):
      # variable for result
      ret = ''
      if isinstance(self.name, basestring) and isinstance(self.value, basestring) and self.name != "" and self.value != "":
        ret = str(self.name) + '=' + str(self.value)

      return ret


    # # Check is allowed path
    #
    # @param url - URL string
    # @return True if allowed or otherwise False
    def isAllowedPath(self, url):
      # variable for result
      ret = True
      try:
        if isinstance(url, basestring) and url != "" and isinstance(self.path, basestring) and self.path != "" and \
          re.search('.*' + self.path + '.*', url, re.UNICODE) is None:
          ret = False

      except Exception, err:
        logger.error(str(err))

      return ret


    # # Check is allowed domain
    #
    # @param url - URL string
    # @return True if allowed or otherwise False
    def isAllowedDomain(self, url):
      # variable for result
      ret = True
      try:
        if isinstance(url, basestring) and url != "" and isinstance(self.domain, basestring) and self.domain != "" and \
          re.search('.*' + self.domain + '.*', url, re.UNICODE) is None:
          ret = False

      except Exception, err:
        logger.error(str(err))

      return ret


    # # Check is allowed expires
    #
    # @param - None
    # @return True if allowed or otherwise False
    def isAllowedExpires(self):
      # variable for result
      ret = True
      try:
        if self.expires is not None:
          expiresDatetime = DateTimeType.parse(self.expires, True, logger, False)
          if expiresDatetime is not None:
            expiresDatetime = DateTimeType.toUTC(expiresDatetime)
            currentDatetime = datetime.datetime.utcnow()
            currentDatetime = currentDatetime.replace(tzinfo=None)
            if currentDatetime < expiresDatetime:
              ret = False

      except Exception, err:
        logger.error(str(err))

      return ret


  # # Internal class for domain properties
  class DomainProperty(object):
    # # Constants used in class
    OPTION_STAGE_NAME = 'stage'
    OPTION_COOKIE_NAME = 'cookie'
    OPTION_DEBUG_NAME = 'debug'
    OPTION_ACCUMULATE_NAME = 'accumulate'
    OPTION_STORAGE_NAME = 'storage'
    OPTION_STATIC_FILE_NAME = 'static_file'

    DEBUG_VALUE_OFF = 0
    DEBUG_VALUE_ON = 1

    ACCUMULATE_VALUE_OFF = 0
    ACCUMULATE_VALUE_ON = 1

    DEFAULT_VALUE_STAGE = None
    DEFAULT_VALUE_COOKIE = ""
    DEFAULT_VALUE_DEBUG = DEBUG_VALUE_OFF
    DEFAULT_VALUE_ACCUMULATE = ACCUMULATE_VALUE_OFF
    DEFAULT_VALUE_STORAGE = ""
    DEFAULT_VALUE_STATIC_FILE = ""


    def __init__(self):
      setattr(self, self.OPTION_STAGE_NAME, self.DEFAULT_VALUE_STAGE)
      setattr(self, self.OPTION_COOKIE_NAME, self.DEFAULT_VALUE_COOKIE)
      setattr(self, self.OPTION_DEBUG_NAME, self.DEFAULT_VALUE_DEBUG)
      setattr(self, self.OPTION_ACCUMULATE_NAME, self.DEFAULT_VALUE_ACCUMULATE)
      setattr(self, self.OPTION_STORAGE_NAME, self.DEFAULT_VALUE_STORAGE)
      setattr(self, self.OPTION_STATIC_FILE_NAME, self.DEFAULT_VALUE_STATIC_FILE)


  # Initialization
  # @param properties - properties as JSON string or dictionary
  # @param siteId - Site/Project ID
  def __init__(self, properties=None, siteId=None):
    self.cookiesDict = {}
    self.propertyDict = {self.DOMAIN_DEFAULT:{self.DOMAIN_OPTION_STAGE_NAME:self.STAGE_DEFAULT}}
    self.siteId = siteId
    self.dataStorage = LFSDataStorage()

    if properties is not None:
      loadedProperties = HTTPCookieResolver.loadProperty(properties)
      if loadedProperties is not None:
        self.propertyDict = loadedProperties

    logger.debug("%s was initialized by %s", self.__class__.__name__, varDump(self.propertyDict))


  # Destroy storage data for zero Site
  def __del__(self):
    # remove storage data
    self.removeStorage(self.ZERO_SITE_ID)


  ## Remove storage data
  #
  # @param siteId - site ID string
  # @return - None
  def removeStorage(self, siteId):
    try:
      if isinstance(siteId, basestring) and siteId != "" and isinstance(self.cookiesDict, dict) and \
        siteId in self.cookiesDict and isinstance(self.cookiesDict[siteId], dict):
        for domain in self.cookiesDict[siteId].keys():
          logger.debug("Found domain: %s", str(domain))
          # extract storage directory path
          storage = self.getStorage(domain)
          logger.debug("Extracted storage: %s", str(storage))
          if isinstance(storage, basestring) and storage != "":
            self.dataStorage.removeElement(storage, domain, siteId)

    except Exception, err:
      logger.error(str(err))


  # # load property from input JSON
  #
  # @param propertyString - contains string with json format
  # @return object properties
  @staticmethod
  def loadProperty(propertyString):
    # variable for result
    ret = None
    try:
      logger.debug("!!! propertyString: %s", varDump(propertyString))
      propertyObj = Utils.jsonLoadsSafe(jsonString=propertyString, default=propertyString, log=logger)
      if isinstance(propertyObj, dict) and len(propertyObj) > 0:
        ret = propertyObj

    except Exception, err:
      logger.error(HTTPCookieResolver.ERROR_INITIALIZATION, 'HTTPCookieResolver', str(err))

    return ret


  # # Set to cookie dictionary value
  #
  # @param siteId - site ID string
  # @param domain - domain name
  # @param cookie - cookie dictionary
  # @param cookiesDict - cookies dictionary for result
  # @return - None
  @staticmethod
  def setCookieValue(siteId, domain, cookie, cookiesDict):
    try:
      if siteId is None:
        siteId = HTTPCookieResolver.ZERO_SITE_ID

      if siteId not in cookiesDict:
        cookiesDict[siteId] = {}

      logger.info("!!! if '%s' in cookiesDict['%s'] = %s", str(domain), str(siteId), str(domain in cookiesDict[siteId]))
      if domain not in cookiesDict[siteId]:
        cookiesDict[siteId][domain] = []

      if isinstance(cookie, basestring) and cookie != "":
        cookiesDict[siteId][domain].extend(HTTPCookieResolver.getCookiesDictionariesList(cookie))

      elif isinstance(cookie, list) and len(cookie) > 0:
        cookiesDict[siteId][domain].extend(cookie)

      elif isinstance(cookie, dict) and len(cookie) > 0:
        cookiesDict[siteId][domain].append(cookie)

    except Exception, err:
      logger.error(str(err))


  # # Remove duplicate from cookies string
  #
  # @param cookiesString - cookies as string
  # @return cookies string with removed duplicates
  @staticmethod
  def removeDuplicateCookies(cookiesString):
    # make unique
    cookiesPairDict = {}
    for cookiesPair in cookiesString.split(HTTPCookieResolver.COOKIES_VALUE_DELIMITER):
      pairs = cookiesPair.split('=')
      if len(pairs) > 1:
        cookiesPairDict[pairs[0]] = pairs[1]

    return  HTTPCookieResolver.COOKIES_VALUE_DELIMITER.join(['%s=%s' % (k, v) for k, v in cookiesPairDict.iteritems()])


  # # Load storage
  #
  # @param siteId - site ID string
  # @param url - URL string or part of the URL
  # @return - None
  def loadStorage(self, siteId, url):
    # extract domain
    domain = self.getDomainFromUrl(url)
    # extract storage directory path
    storage = self.getStorage(domain)
    logger.debug("Extracted storage: %s", str(storage))
    if isinstance(storage, basestring) and storage != "":
      if siteId not in self.cookiesDict:
        self.cookiesDict[siteId] = {}

      extractedData = self.dataStorage.loadElement(storage, domain, siteId, self.cookiesDict[siteId])
      logger.debug("Extracted data: %s", str(extractedData))
      if isinstance(extractedData, dict) and domain in extractedData:
        self.setCookieValue(siteId, domain, extractedData[domain], self.cookiesDict)


  # # Save storage
  #
  # @param siteId - site ID string
  # @param url - URL string or part of the URL
  # @return - None
  def saveStorage(self, siteId, url):
    # extract domain
    domain = self.getDomainFromUrl(url)
    # check the domain is allowed to accumulate
    isAccumulate = self.isAccumulate(domain)
    logger.debug("For domain '%s' is accumulate = %s", str(domain), str(isAccumulate))
    if isAccumulate:
      # extract storage directory path
      storage = self.getStorage(domain)
      logger.debug("Extracted storage: %s", str(storage))

      if isinstance(storage, basestring) and storage != "" and siteId in self.cookiesDict:
        # remove duplicate and expired cookies
        self.cookiesDict[siteId] = self.removeExpiredAndDuplicateCookies(self.cookiesDict[siteId])
        # save cookies to local storage
        self.dataStorage.saveElement(storage, domain, siteId, self.cookiesDict[siteId])

    if self.isDebug(domain):
      logger.debug("Dumped cookies: %s", varDump(self.cookiesDict, stringifyType=0))


  # # Remove expired cookies and duplicate
  #
  # @param cookiesDict - cookies dictionary
  # @return cookies dictionary after remove expired cookies and duplicate
  @staticmethod
  def removeExpiredAndDuplicateCookies(cookiesDict):
#     logger.info("!!! Input cookiesDict: %s", varDump(cookiesDict, stringifyType=0))
    if isinstance(cookiesDict, dict):
      for domain, cookiesList in cookiesDict.items():
        # remove duplicate
        cookiesDictObjects = {}
        for cookiesData in cookiesList:
          cookiesDictObj = HTTPCookieResolver.CookiesDictionary()
          cookiesDictObj.loadFromDict(cookiesData)
          # check expired
          if cookiesDictObj.isAllowedExpires():
            cookiesDictObjects[cookiesDictObj.name] = cookiesDictObj

        # make result list
        cookiesResultList = []
        for cookiesDictObj in cookiesDictObjects.values():
          cookiesResultList.append(cookiesDictObj.toDict())

        cookiesDict[domain] = cookiesResultList

#     logger.info("!!! Output cookiesDict: %s", varDump(cookiesDict, stringifyType=0))
    return cookiesDict


  # # Get site cookies dict
  #
  # @param siteId - site ID string
  # @return cookies dictionary
  def getSiteCookiesDict(self, siteId):
    return self.cookiesDict[siteId] if siteId in self.cookiesDict else {}


  # # get domain from URL
  #
  # @param url - URL string or part of the URL
  # @return domain from URL
  @staticmethod
  def getDomainFromUrl(url):
    # varaible fo result
    ret = None
    try:
      parts = urlparse.urlsplit(url)
#       logger.debug("!!! parts: %s", str(parts))

      if parts.scheme != "":
        ret = parts.hostname
      elif parts.netloc != "":
        ret = parts.netloc
      elif parts.path != "":
        ret = parts.path
      else:
        ret = url

    except Exception, err:
      logger.error(str(err))

    return ret


  # # Add cookie to dictionary as raw string
  #
  # @param siteId - site ID string
  # @param url - URL string or the part of the URL
  # @param cookie - cookie string value or dict
  # @return - None
  def addCookie(self, siteId, url, cookie):
#     logger.debug("!!! addCookie ENTER !!! cookiesDict: %s", varDump(self.cookiesDict, stringifyType=0))
#     logger.debug("!!! url: '%s', cookie: %s", str(url), str(cookie))

    domain = HTTPCookieResolver.getDomainFromUrl(url)
    logger.debug("Extracted domain: %s", str(domain))

    if isinstance(cookie, basestring):
      cookieData = HTTPCookieResolver.extractCookiesData(cookie)
      logger.debug("cookieData: %s, type = %s", str(cookieData), str(type(cookieData)))
      self.setCookieValue(siteId, domain, str(cookieData), self.cookiesDict)

    elif isinstance(cookie, dict):
      self.setCookieValue(siteId, domain, cookie, self.cookiesDict)

#     logger.debug("!!! addCookie LEAVE !!! cookiesDict: %s", varDump(self.cookiesDict, stringifyType=0))


  # # Add cookie to dictionary as raw string
  #
  # @param siteId - site ID string
  # @param responseCookies - response cookies object
  # @return - None
  def appendCookies(self, siteId, responseCookies):
    if responseCookies is not None and len(responseCookies) > 0:
      for cookie in responseCookies:
        self.addCookie(siteId=siteId, url=cookie.domain, cookie=self.CookiesDictionary(name=cookie.name, value=cookie.value, domain=cookie.domain, \
                                                                                       path=cookie.path, expires=cookie.expires).toDict())


  # # Extract pair of name and values
  #
  # @param element - element string
  # @return name and value extracted from element string
  @staticmethod
  def extractPair(element):
    # variables for result
    name = value = ''
    pairElem = element.split('=')
    if len(pairElem) > 0:
      name = pairElem[0]

    if len(pairElem) > 1:
      value = pairElem[1]

    return name, value


  # # Split cookies string
  #
  # @param cookie - cookie string value
  # @return cookies - list of Cookie class instances
  @staticmethod
  def getCookiesDictionariesList(cookieString):
    # variable for result
    cookies = []
    # extract elements
    elementsList = cookieString.split(';')
    elements = []
    for element in elementsList:
      if element.count('=') > 1:
        endPos = element.rfind('=')
        begPos = element.rfind(' ', 0, endPos)

        first = element[:begPos]
        second = element[begPos:]
        first = first.strip(',')

        elements.append(first.strip())
        elements.append(second.strip())
      else:
        elements.append(element.strip())

    logger.debug("!!! elements: %s", varDump(elements))
    cookieObj = HTTPCookieResolver.CookiesDictionary()
    for element in elements:
      name, value = HTTPCookieResolver.extractPair(element)
      # logger.debug("!!! name = %s, value = %s", name, value)

      if name.lower() not in HTTPCookieResolver.CookiesDictionary.__dict__.keys() and \
        cookieObj.name is not None and cookieObj.value is not None:
        cookies.append(cookieObj)
        cookieObj = HTTPCookieResolver.CookiesDictionary()

      if name.lower() in HTTPCookieResolver.CookiesDictionary.__dict__.keys():
        if hasattr(cookieObj, name.lower()):
          setattr(cookieObj, name.lower(), value)
      elif cookieObj.name is None and cookieObj.value is None:
        cookieObj.name = name
        cookieObj.value = value

    if cookieObj.name is not None and cookieObj.value is not None:
      cookies.append(cookieObj.toDict())

    return cookies


  # # Exctract domain properties
  #
  # @param domain - domain name for search cookie
  # @param properties - properties dictionary
  # @return propertyStage and propertyCookie - extracted stage and cookies from domain properties
  @staticmethod
  def extractDomainProperty(domain, properties):
    # variables for result
    propertyStage = HTTPCookieResolver.STAGE_DEFAULT
    propertyCookie = None

    domainPropertyObj = HTTPCookieResolver.getDomainProperty(properties, domain)
    logger.debug("domainPropertyObj: %s", varDump(domainPropertyObj))

    if domainPropertyObj is not None and hasattr(domainPropertyObj, HTTPCookieResolver.DomainProperty.OPTION_STAGE_NAME):
      stageValue = getattr(domainPropertyObj, HTTPCookieResolver.DomainProperty.OPTION_STAGE_NAME)
      if stageValue is not None:
        propertyStage = stageValue

      logger.debug("propertyStage = %s", str(propertyStage))
      res = HTTPCookieResolver.extractCookiesData(getattr(domainPropertyObj, HTTPCookieResolver.DomainProperty.OPTION_COOKIE_NAME))
      if res is not None and res != "":
        propertyCookie = res

    return propertyStage, propertyCookie


  # # Exctract cookies data
  #
  # @param dataString - input data string
  # @return property cookie
  @staticmethod
  def extractCookiesData(dataString):
    # variable for result
    ret = None

    cookiesData = dataString
    if isinstance(dataString, basestring) and dataString != "":
      if cookiesData[0] == '{' and cookiesData[-1] == '}':
        cookiesData = Utils.jsonLoadsSafe(jsonString=cookiesData, default=cookiesData, log=logger)

    if isinstance(cookiesData, dict):
      if HTTPCookieResolver.DEFAULT_HOST_NAME_KEY in cookiesData:
        ret = cookiesData[HTTPCookieResolver.DEFAULT_HOST_NAME_KEY]
      else:
        hostName = Utils.getCurrentHostName()
        logger.debug("hostName: %s", str(hostName))
        if hostName in cookiesData:
          ret = cookiesData[hostName]
        else:
          ip = Utils.getIP()
          logger.debug("IP: %s", str(ip))
          if ip in cookiesData:
            ret = cookiesData[ip]
    else:
      ret = cookiesData

    return ret


  # # Check is allowed stage
  #
  # @param stage - allowed stage of usage (support of different stages use bitmask)
  # @param propertyStage - stage from properties
  # @return True if allowed or otherwise False
  @staticmethod
  def isAllowedStage(stage, propertyStage):
    # variable for result
    ret = False
    if int(stage) & int(propertyStage):
      ret = True

    return ret


  # # Get domain property
  #
  # @param properties - properties structure
  # @param domain - domain name for extract from property
  # @return DomainProperty instance is success or otherwise None
  @staticmethod
  def getDomainProperty(properties, domain):
    # variable for result
    ret = None
    try:
      if not isinstance(properties, dict):
        raise Exception(HTTPCookieResolver.ERROR_BAD_TYPE_INPUT_PROPERTY % (str(type(properties)), str(properties)))

      if isinstance(domain, basestring) and domain != "":
        foundProperty = None
        for propertyKey in properties.keys():
          if re.search(pattern=propertyKey, string=domain, flags=re.U + re.I) is not None:
            foundProperty = properties[propertyKey]
            break

        if foundProperty is None:
          for propertyKey in properties.keys():
            if re.search(pattern=domain, string=propertyKey, flags=re.U + re.I) is not None:
              foundProperty = properties[propertyKey]
              break

        if foundProperty is None and HTTPCookieResolver.DOMAIN_DEFAULT in properties:
          foundProperty = properties[HTTPCookieResolver.DOMAIN_DEFAULT]

        if foundProperty is not None:
          if not isinstance(foundProperty, dict):
            raise Exception(HTTPCookieResolver.ERROR_BAD_TYPE_FOUND_PROPERTY % (str(type(foundProperty)), str(foundProperty)))

          domainPropertyObj = HTTPCookieResolver.DomainProperty()
          for name, value in foundProperty.items():
            if hasattr(domainPropertyObj, name):
              setattr(domainPropertyObj, name, value)
          ret = domainPropertyObj

    except Exception, err:
      logger.error(str(err))

    return ret


  ## Add additional cookies
  #
  # @param domain - domain string
  # @param cookiesString - cookies string
  # @return cookies string with additional cookies
  def addAdditionalCookies(self, domain, cookiesString):
    #variable for result
    ret = cookiesString
    try:
      # get cookie static file name
      staticFile = self.getStaticFile(domain)
      logger.debug("static file: %s", str(staticFile))

      if staticFile is not None:
        staticFileName = os.path.expanduser(staticFile)
        logger.debug("static file full name: %s, is exists = %s", str(staticFileName), str(os.path.isfile(staticFileName)))

        if os.path.isfile(staticFileName):
          with open(staticFileName) as f:
            staticFileData = f.read()
            logger.debug("staticFileData: %s", str(staticFileData))
            if staticFileData.find(self.REQUEST_COOKIE_HEADER_NAME + ':') > -1:
              dataList = staticFileData.split(self.REQUEST_COOKIE_HEADER_NAME + ':')
              if len(dataList) > 1:
                staticFileData = dataList[1]

            logger.debug("additionalCookies: %s", str(staticFileData))

            ret += self.COOKIES_VALUE_DELIMITER + staticFileData.strip()
            ret = self.removeDuplicateCookies(ret)

    except Exception, err:
      logger.error(str(err))

    return ret


  # # Get valid cookie
  #
  # @param siteId - site ID string
  # @param url - url string for search cookie
  # @param stage - allowed stage of usage (support of different stages use bitmask)
  # @return ret - cookie string if found cookie or otherwise None
  def getCookie(self, siteId, url, stage=STAGE_DEFAULT):
    # variable for result
    ret = None

    logger.debug("!!! getCookie ENTER !!! url: %s", str(url))
    logger.debug("!!! self.cookiesDict: %s, type: %s", str(self.cookiesDict), str(type(self.cookiesDict)))

    domain = HTTPCookieResolver.getDomainFromUrl(url)
    logger.debug("Extracted domain: %s", str(domain))

    # extract domain property
    propertyStage, propertyCookie = HTTPCookieResolver.extractDomainProperty(domain, self.propertyDict)
    logger.debug("!!! propertyStage = %s, propertyCookie: '%s'", str(propertyStage), varDump(propertyCookie))
    logger.info('=' * 20)

    isAllowedStage = self.isAllowedStage(stage, propertyStage)
    logger.debug("is allowed stage = %s", str(isAllowedStage))
    if isAllowedStage:
      if propertyCookie is None:
        cookiesValuesDict = {}
        # extract cookies from storage
        for cookiesDomain, cookiesData in self.getSiteCookiesDict(siteId).items():
          if re.search('.*' + domain + '.*', cookiesDomain, re.UNICODE) is not None or \
            re.search('.*' + cookiesDomain + '.*', domain, re.UNICODE) is not None:
            logger.debug("For domain '%s' was found cookies list: '%s'", str(domain), str(cookiesData))
            cookiesValuesDict.update(self.getCookiesValuesDict(cookiesData, url))

        ret = HTTPCookieResolver.COOKIES_VALUE_DELIMITER.join(['%s=%s' % (k, v) for k, v in cookiesValuesDict.iteritems()])
        logger.debug("Covert cookies to string: '%s'", str(ret))
      else:
        # use default cookies from property
        ret = propertyCookie
        logger.debug("For domain '%s' use default cookie: %s", str(domain), str(ret))

      # add additional cookies if necessary
      ret = self.addAdditionalCookies(domain, ret)
      ret = self.removeDuplicateCookies(ret)

    logger.debug("return cookie: '%s'", str(ret))

    return ret


  # # Get cookies string from list
  #
  # @param cookiesList - cookies list
  # @param url - url string for search cookie
  # @return cookies string
  @staticmethod
  def getCookiesValuesDict(cookiesList, url):
    # variable for result
    ret = {}
    try:
      if isinstance(cookiesList, list):
        for cookieObj in cookiesList:
          if isinstance(cookieObj, dict):
            cookiesDictObj = HTTPCookieResolver.CookiesDictionary()
            cookiesDictObj.loadFromDict(cookieObj)
            # check cookie is allowed
            if cookiesDictObj.isAllowedDomain(url) and cookiesDictObj.isAllowedPath(url) and cookiesDictObj.isAllowedExpires():
              logger.debug("Allowed cookie '%s=%s'", str(cookiesDictObj.name), str(cookiesDictObj.value))
              if isinstance(cookiesDictObj.name, basestring) and cookiesDictObj.value is not None:
                ret[cookiesDictObj.name] = str(cookiesDictObj.value)

    except Exception, err:
      logger.error(str(err))

    return ret


  ## Get property value
  #
  # @param propertyDict - property dictionary
  # @param domain - domain string
  # @param attrName - attribute name
  # @return attribute value from property dictionary
  @staticmethod
  def getPropertyValue(propertyDict, domain, attrName):
    # variable for result
    ret = None
    propertyObj = HTTPCookieResolver.getDomainProperty(propertyDict, domain)
    if isinstance(propertyObj, HTTPCookieResolver.DomainProperty) and hasattr(propertyObj, attrName):
      ret = getattr(propertyObj, attrName)

    return ret


  # # Check debug mode
  #
  # @param domain - domain string
  # @return True if debug mode or otherwise False
  def isDebug(self, domain):
    # variable for result
    ret = False
    value = HTTPCookieResolver.getPropertyValue(self.propertyDict, domain, HTTPCookieResolver.DomainProperty.OPTION_DEBUG_NAME)
    if value is not None and int(value) > 0:
      ret = True

    return ret


  # # Check flag is accumulate
  #
  # @param domain - domain string
  # @return True if accumulate mode or otherwise False
  def isAccumulate(self, domain):
    # variable for result
    ret = False
    value = HTTPCookieResolver.getPropertyValue(self.propertyDict, domain, HTTPCookieResolver.DomainProperty.OPTION_ACCUMULATE_NAME)
    if value is not None and int(value) > 0:
      ret = True

    return ret


  # # Get storage value
  #
  # @param domain - domain string
  # @return storage full path if set or otherwise None
  def getStorage(self, domain):
    # variable for result
    ret = None
    value = HTTPCookieResolver.getPropertyValue(self.propertyDict, domain, HTTPCookieResolver.DomainProperty.OPTION_STORAGE_NAME)
    if isinstance(value, basestring) and value != "":
      ret = value

    return ret


  # # Get static file value
  #
  # @param domain - domain string
  # @return static file full path if set or otherwise None
  def getStaticFile(self, domain):
    # variable for result
    ret = None
    value = HTTPCookieResolver.getPropertyValue(self.propertyDict, domain, HTTPCookieResolver.DomainProperty.OPTION_STATIC_FILE_NAME)
    if isinstance(value, basestring) and value != "":
      ret = value

    return ret
