# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
HTTPHeaders Class content main functional for headers manipulations.

@package: dc_crawler
@file HTTPHeaders.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import datetime

from dc_crawler.CrawledResource import CrawledResource


class HTTPHeaders(object):
  # # Constants used in class
  HEADERS_LINE_DELIMITER = '\r\n'
  HEADERS_OPTIONS_DELIMITER = ': '

  DEFAULT_TIMEZONE_STRING_VALUE = '+0000'

  PUBDATE_RSS_FEED_NAME = "X-pubdateRssFeed"
  PUBDATE_RSS_RAW_NAME = "X-pubdateRssRaw"
  RAW_URL_NAME = "X-raw_url"
  BASE_URL_NAME = "X-base_url"
  RSS_FEED_URL_NAME = "X-feed_url"
  CANONICAL_URL_NAME = "X-canonical_url"
  REQUEST_DATE_NAME = "X-request_date"

  DEFAULT_DATE_ISO_FORMAT = "%Y-%m-%d %H:%M:%S"


  # # Initialization
  def __init__(self):
    pass


  # # Header to dictionary
  #
  # @param responseHeader - input response header as string
  # @return headers as dictionary
  @staticmethod
  def headerToDict(responseHeader):
    # variable for result
    ret = {}
    if isinstance(responseHeader, basestring):
      responseHeaderList = responseHeader.split(HTTPHeaders.HEADERS_LINE_DELIMITER)

      for elem in responseHeaderList:
        pair = elem.split(HTTPHeaders.HEADERS_OPTIONS_DELIMITER)
        if len(pair) > 1 :
          ret[pair[0]] = pair[1]

    return ret


  # # Dictionary to header
  #
  # @param headersDict - input dictionary of a headers
  # @return header as string
  @staticmethod
  def dictToHeader(headersDict):
    # variable for result
    ret = ''
    if isinstance(headersDict, dict):

      values = []
      for key, value in headersDict.items():
        values.append(str(key) + HTTPHeaders.HEADERS_OPTIONS_DELIMITER + str(value))

      ret = HTTPHeaders.HEADERS_LINE_DELIMITER.join(values)

    return ret


  # # Set headers value
  #
  # @param responseHeader - input response header as string
  # @param name - option name string
  # @param value - option value string
  # @return - None
  @staticmethod
  def setValue(responseHeader, name, value):
    if isinstance(responseHeader, basestring):
      headersDict = HTTPHeaders.headerToDict(responseHeader)
      headersDict.update({name: value})
      responseHeader = HTTPHeaders.dictToHeader(headersDict)

    return responseHeader


  # # Get headers value
  #
  # @param responseHeader - input response header as string
  # @param name - option name string
  # @param default - default value
  # @return value as string
  @staticmethod
  def getValue(responseHeader, name, default=None):
    # variable for result
    ret = default
    if isinstance(responseHeader, basestring):
      headersDict = HTTPHeaders.headerToDict(responseHeader)
      if name in headersDict:
        ret = headersDict[name]

    return ret


  # # Add value to headers
  #
  # @param responseHeader - response header as string
  # @param name - option name string
  # @param value - option value string
  # @return changed response header string
  @staticmethod
  def addValue(responseHeader, name, value):
    # variable for result
    ret = responseHeader
    if isinstance(responseHeader, basestring) and isinstance(name, basestring) and isinstance(value, basestring):
      if responseHeader != '':
        ret = responseHeader + HTTPHeaders.HEADERS_LINE_DELIMITER

      ret = ret + name + HTTPHeaders.HEADERS_OPTIONS_DELIMITER + value

    return ret


  # # Add pubdate to rss feed crawler resource header
  #
  # @param responseHeader - response header as string
  # @param pubdateRssFeed - pubdate to rss feed
  # @return changed response header string
  @staticmethod
  def addPubdateRssFeedToHeader(responseHeader, pubdateRssFeed):
    return HTTPHeaders.addValue(responseHeader, HTTPHeaders.PUBDATE_RSS_FEED_NAME, str(pubdateRssFeed) + HTTPHeaders.DEFAULT_TIMEZONE_STRING_VALUE)


  # # Add raw pubdate to rss crawler resource header
  #
  # @param responseHeader - response header as string
  # @param pubdateRssRaw - raw pubdate string
  # @return changed response header string
  @staticmethod
  def addPubdateRssRawToHeader(responseHeader, pubdateRssRaw):
    return HTTPHeaders.addValue(responseHeader, HTTPHeaders.PUBDATE_RSS_RAW_NAME, pubdateRssRaw)


  # # Add base url to crawler resource header
  #
  # @param crawledResource - crawled resource
  # @param baseUrl - base url
  # @return - None
  @staticmethod
  def addBaseUrlToHeader(inObj, baseUrl):
    if isinstance(inObj, CrawledResource):
      inObj.response_header = HTTPHeaders.addValue(inObj.response_header, HTTPHeaders.BASE_URL_NAME, baseUrl)
    if isinstance(inObj, basestring):
      inObj = HTTPHeaders.addValue(inObj, HTTPHeaders.BASE_URL_NAME, baseUrl)

    return inObj


  # # Add feed url to rss feed crawler resource header
  #
  # @param crawledResource - crawled resource
  # @param feedUrl - feed url
  # @return - None
  @staticmethod
  def addFeedUrlToHeader(crawledResource, feedUrl):
    if isinstance(crawledResource, CrawledResource):
      crawledResource.response_header = HTTPHeaders.addValue(crawledResource.response_header, HTTPHeaders.RSS_FEED_URL_NAME, feedUrl)


  # # Add canonical url to crawler resource header
  #
  # @param crawledResource - crawled resource
  # @param canonicalUrl - canonical url
  # @return - None
  @staticmethod
  def addCanonicalUrlToHeader(crawledResource, canonicalUrl):
    if isinstance(crawledResource, CrawledResource):
      crawledResource.response_header = HTTPHeaders.addValue(crawledResource.response_header, HTTPHeaders.CANONICAL_URL_NAME, canonicalUrl)


  # # Add request date to crawler resource header
  #
  # @param crawledResource - crawled resource
  # @param requestDate - request date string in ISO format
  # @param log - logger instance
  # @return - None
  @staticmethod
  def addRequestDateToHeader(crawledResource, requestDate=None, log=None):
    try:
      if isinstance(crawledResource, CrawledResource):
        if requestDate is None:
          requestDate = datetime.datetime.now().strftime(HTTPHeaders.DEFAULT_DATE_ISO_FORMAT)
        elif isinstance(requestDate, datetime.datetime):
          requestDate = requestDate.strftime(HTTPHeaders.DEFAULT_DATE_ISO_FORMAT)

        if isinstance(requestDate, basestring):
          crawledResource.response_header = HTTPHeaders.addValue(crawledResource.response_header, HTTPHeaders.REQUEST_DATE_NAME, requestDate)

    except Exception, err:
      if log is not None:
        log.error(str(err))


  # # Add header value to crawled resource header
  #
  # @param crawledResource - crawled resource
  # @param responseHeader - response header as string
  # @return - None
  @staticmethod
  def addHeaderToHeader(crawledResource, responseHeader):
    if isinstance(crawledResource, CrawledResource) and isinstance(responseHeader, basestring) and responseHeader != "":
      responseDict = HTTPHeaders.headerToDict(responseHeader)
      responseDict.update(HTTPHeaders.headerToDict(crawledResource.response_header))
      crawledResource.response_header = HTTPHeaders.dictToHeader(responseDict)
