# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
HTTPRedirectExtractor Class content main functional for 'HTTP_REDIRECT_RESOLVER' property.

@package: dc_crawler
@file HTTPRedirectExtractor.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2018 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

from dc_crawler.HTTPRedirectResolver import HTTPRedirectResolver
from dc_crawler.HTTPHeadersExtractor import HTTPHeadersExtractor
from dc_crawler.HTTPCookieExtractor import HTTPCookieExtractor
from dc_crawler.Fetcher import BaseFetcher
from dc_crawler.RefererHeaderResolver import RefererHeaderResolver
import dc_crawler.Constants as CRAWLER_CONSTS
import app.Consts as APP_CONSTS
import app.Utils as Utils


class HTTPRedirectExtractor(object):
  # # Constants used in class
  HTTP_METHOD_POST = 'post'
  PROXY_PROTOCOL_HTTP = 'http'
  PROXY_PROTOCOL_HTTPS = 'https'
  HEADERS_OPTION_REFFERER = 'Referer'

  DEFAULT_VALUE_FETCH_TYPE = BaseFetcher.TYP_NORMAL


  # # Initialization
  def __init__(self):
    pass


  # # fetch URL
  #
  # @param siteProperties - site properties dictionary
  # @param siteId - site ID
  # @param url - the URL to fetch
  # @param headers - request headers dictionary
  # @param httpTimeout - HTTP timeout
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxies - proxies settings dictionary
  # @param allowedContentTypes - allowed_content_types list
  # @param maxResourceSize - max resource size
  # @param maxRedirects - max allowed redirects count
  # @param httpMethod - HTTP method
  # @param fetchType - fetch type
  # @param dbWrapper - DBWrapper instance
  # @param filters - filters dictionary
  # @param depth - depth value
  # @param macro - macro code
  # @param stage - crawler stage for apply
  # @param cookiesStage - cookies stage for apply
  # @param log - logger instance
  # @param fetchAlways - boolean flag fetch always
  # @param useDetectModified - boolean flag use detect modified
  # @return URL fetch response
  @staticmethod
  def fetchUrl(siteProperties, siteId, url, headers, httpTimeout, allowRedirects=True, proxies=None, allowedContentTypes=None, maxResourceSize=None,
               maxRedirects=HTTPRedirectResolver.RedirectProperty.DEFAULT_VALUE_MAX_REDIRECTS, httpMethod=None, fetchType=None, dbWrapper=None, filters=None,
               depth=None, macro=None, stage=None, cookiesStage=None, log=None, fetchAlways=True, useDetectModified=False, **kwarg):

    # variable for result
    ret = None
    try:
      if log is not None:
        log.debug("siteProperties type = %s", str(type(siteProperties)))

      connectionTimeout = float(siteProperties[APP_CONSTS.CONNECTION_TIMEOUT]) if APP_CONSTS.CONNECTION_TIMEOUT in siteProperties \
        else CRAWLER_CONSTS.CONNECTION_TIMEOUT

      tm = int(httpTimeout) / 1000.0
      if isinstance(httpTimeout, float):
        tm += float('0' + str(httpTimeout).strip()[str(httpTimeout).strip().find('.'):])

      auth = None
      if APP_CONSTS.HTTP_AUTH_NAME in siteProperties and APP_CONSTS.HTTP_AUTH_PWD in siteProperties:
        authName = siteProperties[APP_CONSTS.HTTP_AUTH_NAME]
        authPwd = siteProperties[APP_CONSTS.HTTP_AUTH_PWD]
        if authName is not None and authPwd is not None:
          auth = (authName, authPwd)

      postForms = {}
      for key, value in siteProperties.items():
        if key.startswith(APP_CONSTS.HTTP_POST_FORM_PREFIX):
          postForms[key[len(APP_CONSTS.HTTP_POST_FORM_PREFIX):]] = value

      postData = postForms if isinstance(httpMethod, basestring) and httpMethod.lower() == HTTPRedirectExtractor.HTTP_METHOD_POST else None

      if APP_CONSTS.HTTP_REDIRECTS_MAX in siteProperties:
        maxRedirects = int(siteProperties[APP_CONSTS.HTTP_REDIRECTS_MAX])

      if fetchType is None:
        fetchType = HTTPRedirectExtractor.DEFAULT_VALUE_FETCH_TYPE

      # extract actual headers values
      if HTTPHeadersExtractor.ARGUMENT_DOMAIN not in kwarg:
        kwarg[HTTPHeadersExtractor.ARGUMENT_DOMAIN] = Utils.parseHost(url, True)
      if HTTPHeadersExtractor.ARGUMENT_SITEID not in kwarg:
        kwarg[HTTPHeadersExtractor.ARGUMENT_SITEID] = siteId

      headers = HTTPHeadersExtractor.extract(siteProperties=siteProperties, url=url, headers=headers, log=log, **kwarg)

      # update headers by referrer value
      RefererHeaderResolver.updateHeaders(siteProperties=siteProperties, headers=headers, url=url, siteId=siteId, dbWrapper=dbWrapper, log=log)

      # update headers by cookies
      headers = HTTPCookieExtractor.updateHeadersByCookies(siteProperties=siteProperties, siteId=siteId, url=url, stage=cookiesStage, headers=headers)

      redirectResolver = HTTPRedirectResolver(siteProperties=siteProperties, fetchType=fetchType, dbWrapper=dbWrapper, siteId=siteId,
                                              connectionTimeout=connectionTimeout, stage=stage)

      # make fetch URL
      ret = redirectResolver.fetchUrl(url=url, headers=headers, timeout=tm, allowRedirects=allowRedirects, proxies=proxies, auth=auth, postData=postData,
                                      allowedContentTypes=allowedContentTypes, maxResourceSize=maxResourceSize, maxRedirects=maxRedirects, filters=filters,
                                      depth=depth, macro=macro, fetchAlways=fetchAlways, useDetectModified=useDetectModified, **kwarg)

      if ret is not None and isinstance(ret.cookies, basestring):
        # save returned cookies as string (for sample it can be after SeleniumFetcher)
        HTTPCookieExtractor.addCookie(siteProperties=siteProperties, siteId=siteId, url=url, cookie=ret.cookies)

    except Exception, err:
      if log is not None:
        log.debug(str(err))
      raise err

    return ret


  # # extract http redirect resolved url
  #
  # @param siteProperties - site properties dictionary
  # @param siteId - site ID
  # @param url - the url to fetch
  # @param headers - request headers dict
  # @param httpTimeout - http timeout
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxies - proxies settings dictionary
  # @param allowedContentTypes - allowed_content_types list
  # @param maxResourceSize - max resource size
  # @param maxRedirects - max allowed redirects count
  # @param httpMethod - HTTP method
  # @param fetchType - fetch type
  # @param dbWrapper - DBWrapper instance
  # @param filters - filters dictionary
  # @param depth - depth value
  # @param macro - macro code
  # @param stage - stage for apply
  # @param cookiesStage - cookies stage for apply
  # @param log - logger instance
  # @return resolved URL string or None
  @staticmethod
  def extract(siteProperties, siteId, url, headers, httpTimeout, allowRedirects=True, proxies=None, allowedContentTypes=None, maxResourceSize=None,
              maxRedirects=HTTPRedirectResolver.RedirectProperty.DEFAULT_VALUE_MAX_REDIRECTS, httpMethod=None, fetchType=None, dbWrapper=None, filters=None,
              depth=None, macro=None, stage=None, cookiesStage=None, log=None):
    # variable for result
    ret = url
    try:
      if APP_CONSTS.HTTP_REDIRECT_RESOLVER in siteProperties and siteProperties[APP_CONSTS.HTTP_REDIRECT_RESOLVER] != "":
        # make fetch URL
        res = HTTPRedirectExtractor.fetchUrl(siteProperties=siteProperties, siteId=siteId, url=url, headers=headers, httpTimeout=httpTimeout,
                                             allowRedirects=allowRedirects, proxies=proxies, allowedContentTypes=allowedContentTypes,
                                             maxResourceSize=maxResourceSize, maxRedirects=maxRedirects, httpMethod=httpMethod, fetchType=fetchType,
                                             dbWrapper=dbWrapper, filters=filters, depth=depth, macro=macro, stage=stage, cookiesStage=cookiesStage, log=log,
                                             fetchAlways=False)
        if res is not None:
          ret = res.url

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret

  # # check flag insert as new
  #
  # @param siteProperties - site properties dictionary
  # @param url - the URL string
  # @param log - logger instance
  # @return boolean flag is insert as new value
  @staticmethod
  def isInsertAsNewUrl(siteProperties, url, log=None):
    #variable for result
    ret = False
    try:
      if APP_CONSTS.HTTP_REDIRECT_RESOLVER in siteProperties and siteProperties[APP_CONSTS.HTTP_REDIRECT_RESOLVER] != "":
        if log is not None:
          log.debug("Found property: '%s': %s", str(APP_CONSTS.HTTP_REDIRECT_RESOLVER), str(siteProperties[APP_CONSTS.HTTP_REDIRECT_RESOLVER]))

        ret = HTTPRedirectResolver(siteProperties=siteProperties).isInsertAsNewUrl(url)

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  # # update URL object if necessary
  #
  # @param siteProperties - site properties dictionary
  # @param urlObj - URLs objects instance
  # @param url - the URL string
  # @param log - logger instance
  # @return - None
  @staticmethod
  def updateUrl(siteProperties, urlObj, url, log=None):
    try:
      if APP_CONSTS.HTTP_REDIRECT_RESOLVER in siteProperties and siteProperties[APP_CONSTS.HTTP_REDIRECT_RESOLVER] != "":
        if log is not None:
          log.debug("Found property: '%s': %s", str(APP_CONSTS.HTTP_REDIRECT_RESOLVER), str(siteProperties[APP_CONSTS.HTTP_REDIRECT_RESOLVER]))

        HTTPRedirectResolver(siteProperties=siteProperties).updateUrl(urlObj, url)

    except Exception, err:
      if log is not None:
        log.error(str(err))
