# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
RequestsRedirectWrapper Class content main functional for resolve redirect.

@package: dc_crawler
@file RequestsRedirectWrapper.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2018 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import copy
import time
import requests.exceptions
import requests.structures
import requests.packages.urllib3
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from app.Utils import varDump
from app.Filters import Filters
from app.Utils import getTracebackInfo
import app.Utils as Utils
from app.Exceptions import UrlAvailableException
import dc_crawler.Constants as CONSTS
from dc_crawler.Exceptions import CrawlerFilterException
from dc_crawler.Exceptions import TooManyHTTPRedirectsException
from dc_crawler.Exceptions import TooManyHTMLRedirectsException
from dc_crawler.Exceptions import BadStatusCodeException
from dc_crawler.HTTPCookieResolver import HTTPCookieResolver
from dc_crawler.HTTPCookieExtractor import HTTPCookieExtractor
from dc_crawler.HTTPProxyResolver import HTTPProxyResolver
from dc_crawler.HTTPProxy import HTTPProxy
from dc_crawler.HostAliveExtractor import HostAliveExtractor
from dc_crawler.HTMLRedirectResolver import HTMLRedirectResolver
from dc_crawler.HTTPRedirectHeadersDebug import HTTPRedirectHeadersDebug

# disable warnings
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

logger = Utils.MPLogger().getLogger()


class RequestsRedirectWrapper(object):
  # Constants used in class
  REQUEST_COOKIE_HEADER_NAME = 'Cookie'
  RESPONSE_COOKIE_HEADER_NAME = 'set-cookie'
  REFERER_HEADER_NAME = 'Referer'
  LOCATION_HEADER_NAME = 'location'
  CONTENT_TYPE_HEADER_NAME = 'content-type'

  # header names for remove
  REDIRECT_HEADER_FIELDS_FOR_REMOVE = [CONTENT_TYPE_HEADER_NAME, LOCATION_HEADER_NAME, REQUEST_COOKIE_HEADER_NAME]

  # Usage algorithm constants
  USAGE_ALGORITHM_BASE = 0
  USAGE_ALGORITHM_CUSTOM = 1
  DEFAULT_USAGE_ALGORITHM = USAGE_ALGORITHM_CUSTOM  # USAGE_ALGORITHM_BASE

  DEFAULT_SITE_PROPERTIES = {}
  DEFAULT_VALUE_SITE_ID = '0'

  # Constants of error messages
  ERROR_BAD_STATUS_CODE_VALUE = "Not allowed status code '%s'. Allowed list: %s"
  ERROR_BAD_CONTENT_TYPE_VALUE = "Not allowed content-type = '%s'. Allowed list: %s"


  # Initialization
  #
  # @param dbWrapper  - DB wrapper instance
  # @param siteId - site ID
  # @param usageAlgorithm - usage algorithm number
  # @param redirectCodes - redirect codes numbers list
  # @param siteProperties - site properties dictionary
  def __init__(self, dbWrapper=None, siteId=None, usageAlgorithm=DEFAULT_USAGE_ALGORITHM, redirectCodes=None, siteProperties=None):
    object.__init__(self)
    self.dbWrapper = dbWrapper
    self.siteId = self.DEFAULT_VALUE_SITE_ID if siteId is None else siteId
    self.usageAlgorithm = usageAlgorithm
    self.redirectCodes = redirectCodes if isinstance(redirectCodes, list) else CONSTS.REDIRECT_HTTP_CODES
    self.siteProperties = self.DEFAULT_SITE_PROPERTIES if siteProperties is None else Utils.jsonLoadsSafe(jsonString=siteProperties, default=siteProperties,
                                                                                                          log=logger)
    # extract max allowed HTML redirects
    self.maxRedirectsHTML = HTMLRedirectResolver.getMaxAllowedRedirects(self.siteProperties, logger)
    # first redirect response
    self.firstRedirectResponse = None


  # @param url - the URL to fetch
  # @param method - fetch by HTTP method
  # @param timeout - request timeout(seconds)
  # @param headers - request headers dictionary
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxySetting - proxies setting dictionary
  # @param auth - basic authentication setting, tuple of name and password
  # @param data - post data, used only when method is post
  # @param maxRedirects - max allowed redirects count
  # @param filters - filters dictionary
  # @param maxRedirectsHTML - max allowed HTML redirects count
  # @return Response object
  def request(self, url, method, timeout, headers, allowRedirects, proxySetting, auth, data, maxRedirects, filters, maxRedirectsHTML=None):
    # variable for return
    implRes = None

    logger.info("self.siteId = '%s'", str(self.siteId))
#     logger.info("self.siteProperties:\n%s", varDump(self.siteProperties, stringifyType=0))
    logger.info("!!! is exist '%s' in self.siteProperties = %s", str(HTTPProxyResolver.USER_PROXY_PROPERTY_NAME),
                str(HTTPProxyResolver.USER_PROXY_PROPERTY_NAME in self.siteProperties))
    logger.info("Usage headers:\n%s", varDump(headers, stringifyType=0))

    logger.info("self.dbWrapper = %s", str(self.dbWrapper))

    usageAlgorithm = self.usageAlgorithm
    # create DBProxyWrapper instance if necessary and set algorithm
    if HTTPProxyResolver.USER_PROXY_PROPERTY_NAME in self.siteProperties or (HTTPProxyResolver.HTTP_PROXY_HOST_NAME in self.siteProperties and \
      HTTPProxyResolver.HTTP_PROXY_PORT_NAME in self.siteProperties):
      usageAlgorithm = self.USAGE_ALGORITHM_CUSTOM
    else:
      usageAlgorithm = self.USAGE_ALGORITHM_BASE

    logger.info("usageAlgorithm = %s, it '%s'", str(usageAlgorithm), 'USAGE_ALGORITHM_CUSTOM' if usageAlgorithm == self.USAGE_ALGORITHM_CUSTOM else\
                'USAGE_ALGORITHM_BASE')

    # make request
    if usageAlgorithm == self.USAGE_ALGORITHM_BASE:
      implRes = self.requestBase(url, method, timeout, headers, allowRedirects, proxySetting, auth, data, maxRedirects, filters)
      # accumulate cookies and save to storage if necessary
      HTTPCookieExtractor.accumulateCookies(siteProperties=self.siteProperties, siteId=self.siteId, url=url, responseCookies=implRes.cookies)

    elif usageAlgorithm == self.USAGE_ALGORITHM_CUSTOM:
      implRes = self.requestCustom(url, method, timeout, headers, allowRedirects, proxySetting, auth, data, maxRedirects, filters)
    else:
      raise Exception("Try using not support algorithm (%s) for the request!" % (str(usageAlgorithm)))

#     logger.info("!!!!!!!! maxRedirectsHTML = %s", str(maxRedirectsHTML))
#     logger.info("!!! self.maxRedirectsHTML = %s", str(self.maxRedirectsHTML))
#     logger.info("!!! implRes.headers:\n%s\ntype = %s", varDump(implRes.headers, stringifyType=0), str(type(implRes.headers)))

    # set dynamically attribute
    if self.firstRedirectResponse is not None:
      implRes.firstRedirectResponse = self.firstRedirectResponse

    isAllowedContentType = HTMLRedirectResolver.isAllowedContentType(implRes.headers, logger)
    logger.debug("Is allowed content-type for resolve HTML redirect = %s", str(isAllowedContentType))

    if isAllowedContentType:
      if maxRedirectsHTML is None:
        maxRedirectsHTML = self.maxRedirectsHTML
      # extract HTML redirects
      extractedUrl = HTMLRedirectResolver.extractURL(implRes.url, implRes.content, logger)
      if extractedUrl is not None:
        # check max allowed redirects value
        if maxRedirectsHTML is not None:
          maxRedirectsHTML -= 1
          if maxRedirectsHTML < 0:
            raise TooManyHTMLRedirectsException("Exceeded %s HTML redirects for URL: %s" % (str(self.maxRedirectsHTML), str(url)), response=implRes)

        implRes = self.request(extractedUrl, method, timeout, headers, allowRedirects, proxySetting, auth, data, maxRedirects, filters, maxRedirectsHTML)

#     logger.info("!!! self.firstRedirectResponse\n%s", varDump(self.firstRedirectResponse, stringifyType=0))

    return implRes


  # # Request base
  #
  # @param url - the url to fetch
  # @param method - fetch by HTTP method
  # @param timeout - request timeout(seconds)
  # @param headers - request headers dict
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxySetting - proxies setting dictionary
  # @param auth - basic auth setting, tuple of name and password
  # @param data - post data, used only when method is post
  # @param maxRedirects - max allowed redirects count
  # @param filters - filters dict
  # @return Response object
  def requestBase(self, url, method, timeout, headers, allowRedirects, proxySetting, auth, data, maxRedirects, filters):
    # variable for return
    res = None
    try:
      # Checking is available URL
      if not HostAliveExtractor.isAvailableUrl(siteProperties=self.siteProperties, url=url, log=logger):
        logger.debug("Host '%s' is not available!", str(url))
        raise UrlAvailableException("Host '%s' is not available!" % str(url))

      # prepare request
      req = requests.Request(method=method,
                             url=url,
                             headers=headers,
                             auth=auth,
                             data=data,
                             hooks={'response':[RequestsRedirectWrapper.checkRedirect]})

#       logger.debug("!!! headers: %s, type: %s", varDump(headers), str(type(headers)))

      reqv = req.prepare()

      rSession = requests.Session()
      rSession.max_redirects = int(maxRedirects)
      rSession.stream = True
      rSession.verify = False  # don't verify ssl
      rSession.proxies = proxySetting
      # send request
      res = rSession.send(request=reqv, allow_redirects=allowRedirects, timeout=timeout)

      self.__checkResponse(res, filters)

    except requests.exceptions.RequestException, err:
      logger.debug("!!! RequestException: %s, exception type = %s", str(err), str(type(err)))
      raise err
    except Exception, err:
      logger.debug("!!! Exception: %s, exception type = %s", str(err), str(type(err)))
      logger.info(getTracebackInfo())
      raise err

    return res


  # #Check allowed Response object
  # # if check not passed will be raise exception
  #
  # @param res - Response object
  # @param filters - filters dict
  # @return - None
  def __checkResponse(self, res, filters):

    redirectUrlsList = []
    for history in res.history:
      # save redirect URLs for extra header fields
      redirectUrlsList = HTTPRedirectHeadersDebug.addData(history.url, history.status_code, self.siteProperties, logger, redirectUrlsList)
    # update headers by extra fields
    res = HTTPRedirectHeadersDebug.updateResponseHeaders(redirectUrlsList, res)

    if self.firstRedirectResponse is None:
      if len(res.history) > 0:
        self.firstRedirectResponse = copy.deepcopy(res.history[0])
      else:
        self.firstRedirectResponse = copy.deepcopy(res)

      logger.debug("First redirect response was created: %s", varDump(self.firstRedirectResponse))
    else:
      logger.debug("First redirect response already exists. Skipped...")

    if not self.__isAllowedUrl(res.url, filters):
      raise CrawlerFilterException("Url %s not passed filter" % str(res.url), response=self.firstRedirectResponse)

    for history in res.history:
      logger.debug("!!! history.url: %s", varDump(history.url))
      logger.debug("!!! history.status_code: %s", varDump(history.status_code))
      # check filters
      if not self.__isAllowedUrl(history.url, filters):
        raise CrawlerFilterException("URL '%s' not passed filter" % str(history.url), response=self.firstRedirectResponse)
      # check status codes
      if isinstance(self.redirectCodes, list) and history.status_code not in self.redirectCodes:
        raise BadStatusCodeException(self.ERROR_BAD_STATUS_CODE_VALUE % \
                                     (str(history.status_code), str(self.redirectCodes)), response=self.firstRedirectResponse)


  # # Request custom
  #
  # @param url - the url to fetch
  # @param method - fetch by HTTP method
  # @param timeout - request timeout(seconds)
  # @param headers - request headers dict
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxySetting - proxies setting dictionary
  # @param auth - basic auth setting, tuple of name and password
  # @param data - post data, used only when method is post
  # @param maxRedirects - max allowed redirects count
  # @param filters - filters dict
  # @return Response object
  def requestCustom(self, url, method, timeout, headers, allowRedirects, proxySetting, auth, data, maxRedirects, filters):
    # variable for return
    implRes = None
    applyHeaders = copy.deepcopy(headers)
#     logger.debug("!!! applyHeaders: %s, type = %s", str(applyHeaders), str(type(applyHeaders)))
    # logger.debug("!!! request enter ...  applyHeaders: %s", varDump(applyHeaders))
#     logger.info("!!! self.siteProperties:\n%s", varDump(self.siteProperties, stringifyType=0))
    logger.info("!!! is exist '%s' in siteProperties = %s", str(HTTPProxyResolver.USER_PROXY_PROPERTY_NAME),
                str(HTTPProxyResolver.USER_PROXY_PROPERTY_NAME in self.siteProperties))
    logger.info("!!! is exist '%s' in siteProperties.keys() = %s", str(HTTPProxyResolver.USER_PROXY_PROPERTY_NAME),
                str(HTTPProxyResolver.USER_PROXY_PROPERTY_NAME in self.siteProperties.keys()))

    logger.debug('#' * 40)
    refererValue = None

    if RequestsRedirectWrapper.isExistsHeaderField(applyHeaders, RequestsRedirectWrapper.REFERER_HEADER_NAME):
      refererValue = RequestsRedirectWrapper.getHeaderFieldValue(applyHeaders, RequestsRedirectWrapper.REFERER_HEADER_NAME)
    logger.debug("refererValue: '%s'", str(refererValue))

    cookieResolver = HTTPCookieResolver(siteId=self.siteId)
    cachedCookies = HTTPCookieExtractor.getCookies(siteProperties=self.siteProperties, siteId=self.siteId, url=url, stage=HTTPCookieResolver.STAGE_REDIRECT)
    cookieResolver.addCookie(siteId=self.siteId, url=url, cookie=cachedCookies)

    redirectUrlsList = []
    redirectsCount = 0
    while redirectsCount < int(maxRedirects):
      retriesCount = HTTPProxyResolver.getTriesCount(self.siteProperties)
      logger.debug("retriesCount = %s", str(retriesCount))

      proxyTriesCount = 0
      for triesCount in range(0, retriesCount + 1):
        logger.debug("Try iteration %s from %s retries.", str(triesCount), str(retriesCount))
        logger.debug("proxyTriesCount = %s", str(proxyTriesCount))
        # check is exceeded tries count
        HTTPProxyResolver.checkTriesCount(siteProperties=self.siteProperties, currentTriesCount=proxyTriesCount)

        if isinstance(proxySetting, dict) and len(proxySetting) > 0:
          logger.info("Input proxySetting: %s", varDump(proxySetting, stringifyType=0))

        logger.info("!!! Try get new proxy name for url: %s", str(url))
        # get proxy name
        proxyName, isValidProxy = HTTPProxyResolver.getProxyName(siteProperties=self.siteProperties, siteId=self.siteId, url=url, dbWrapper=self.dbWrapper,
                                                                 log=logger)

        logger.info("!!! Extracted proxyName: '%s', isValidProxy = %s", str(proxyName), str(isValidProxy))
        # increment counter
        if proxyName is not None:
          proxyTriesCount += 1

        if not isValidProxy and proxyName is not None:
          logger.debug("Got invalid proxy '%s'. Fetching skipped. Try get next proxy...", str(proxyName))
          continue

        logger.info("!!! Try get new proxySetting...")
        resProxySetting = HTTPProxy.getProxies(proxyName=proxyName, log=logger)
        logger.debug("!!! Extracted proxySetting: %s", str(resProxySetting))
        if isinstance(resProxySetting, dict) and len(resProxySetting) > 0:
          proxySetting = resProxySetting

        logger.debug("Use proxySetting: %s", varDump(proxySetting, stringifyType=0))

        logger.info("!!! Try get cookies... for %s", str(url))
        cachedCookies = cookieResolver.getCookie(siteId=self.siteId, url=url)
        logger.debug("!!! Extracted cachedCookies: '%s'", str(cachedCookies))
        if isinstance(cachedCookies, basestring) and cachedCookies != "":
          applyHeaders[RequestsRedirectWrapper.REQUEST_COOKIE_HEADER_NAME] = cachedCookies

        if isinstance(refererValue, basestring) and refererValue != "":
          applyHeaders[RequestsRedirectWrapper.REFERER_HEADER_NAME] = refererValue

        implRes = self.requestBase(url, method, timeout, applyHeaders, False, proxySetting, auth, data, maxRedirects=1, filters=filters)

        logger.debug("!!! implRes.content len = %s", str(len(implRes.content)))
        logger.debug("!!! implRes.status_code = %s", str(implRes.status_code))
        logger.debug("!!! implRes.headers: %s", str(implRes.headers))
        logger.debug("!!! implRes.cookies: %s, type: %s", varDump(implRes.cookies), str(type(implRes.cookies)))
        logger.debug("!!! implRes:\n%s\ntype: %s", varDump(implRes, maxDepth=10), str(type(implRes)))

#         # remove extra header fields
#         applyHeaders = self.removeExtraHeaderFields(applyHeaders)
#         logger.debug("!!! applyHeaders after cleanup: %s", str(applyHeaders))

        cookieResolver.appendCookies(siteId=self.siteId, responseCookies=implRes.cookies)

        HTTPCookieExtractor.accumulateCookies(siteProperties=self.siteProperties, siteId=self.siteId, url=url, responseCookies=implRes.cookies)

        # check necessary the rotation of the proxies
        if implRes.status_code not in self.redirectCodes and \
          HTTPProxyResolver.isNeedRotate(siteProperties=self.siteProperties, siteId=self.siteId, proxyName=proxyName, dbWrapper=self.dbWrapper,
                                         rawContent=implRes.content):
          # remove extra header fields
          applyHeaders = self.removeExtraHeaderFields(applyHeaders)
          logger.debug("!!! applyHeaders after cleanup: %s", str(applyHeaders))
          logger.debug('Necessary rotate proxy. Go to the next...')
          time.sleep(1)
          continue
        else:
          break

      if implRes.status_code not in self.redirectCodes or not allowRedirects:
        logger.debug("!!! break !!!")
        break

      # set URL default value
      url = implRes.url
      # save resolved URL value for next iteration
      if RequestsRedirectWrapper.isExistsHeaderField(implRes.headers, self.LOCATION_HEADER_NAME):
        # extract location URL
        locationURL = RequestsRedirectWrapper.getHeaderFieldValue(implRes.headers, self.LOCATION_HEADER_NAME)
        # save redirect URLs for extra header fields
        redirectUrlsList = HTTPRedirectHeadersDebug.addData(locationURL, implRes.status_code, self.siteProperties, logger, redirectUrlsList)
        # normalization of the URL
        url = Utils.urlNormalization(base=implRes.url, url=locationURL, log=logger)
        logger.info("!!! Set URL from location: %s", str(url))

      if redirectsCount > 0:
        if not self.__isAllowedUrl(url, filters):
          implRes = HTTPRedirectHeadersDebug.updateResponseHeaders(redirectUrlsList, implRes)
          raise CrawlerFilterException("URL %s not passed filter" % str(url), response=implRes)

      redirectsCount += 1
      logger.debug("!!! redirectsCount = %s, maxRedirects = %s", str(redirectsCount), str(maxRedirects))

    # append extra redirect URLs fields
    implRes = HTTPRedirectHeadersDebug.updateResponseHeaders(redirectUrlsList, implRes)
    if implRes is not None and implRes.status_code in self.redirectCodes:
      raise TooManyHTTPRedirectsException("Exceeded %s HTTP redirects for URL: %s" % (str(maxRedirects), str(url)), response=implRes)

    return implRes

  # # check is exist header field name
  #
  # @param headers - request headers dictionary
  # @param name - the headers field name
  # @return boolean flag True if exists or otherwise False
  @staticmethod
  def isExistsHeaderField(headers, name):
    # variable for result
    ret = False
    if isinstance(headers, (dict, requests.structures.CaseInsensitiveDict)):
      for key in headers.keys():
        if key.lower() == name.lower():
          ret = True
          break

    return ret


  # # get value from header by name
  #
  # @param headers - request headers dictionary
  # @param name - the headers field name
  # @return value from header
  @staticmethod
  def getHeaderFieldValue(headers, name):
    # variable for result
    ret = None
    if isinstance(headers, (dict, requests.structures.CaseInsensitiveDict)):
      for key in headers.keys():
        if key.lower() == name.lower():
          ret = headers[key]
          break

    return ret


  # # remove one field from header fields
  #
  # @param headers - request headers dictionary
  # @param name - the headers field name
  # @return headers - result headers already updated
  @staticmethod
  def removeOneHeaderField(headers, name):
    if isinstance(headers, (dict, requests.structures.CaseInsensitiveDict)):
      for key in headers.keys():
        if key.lower() == name.lower():
          del headers[key]
          break

    return headers


  # # remove extra header fields
  #
  # @param headers - request headers dictionary
  # @return headers - result headers already updated
  @staticmethod
  def removeExtraHeaderFields(headers):
    for name in RequestsRedirectWrapper.REDIRECT_HEADER_FIELDS_FOR_REMOVE:
      headers = RequestsRedirectWrapper.removeOneHeaderField(headers, name)

    return headers


  # # check allowed url by filter
  #
  # @param url - url for check
  # @param inputFilters - filters dict
  # @return - True if allowed by filter or othrwise False
  def __isAllowedUrl(self, url, inputFilters=None):
    # variable for result
    ret = True

    if self.dbWrapper is not None and self.siteId is not None and inputFilters is not None:
      # Check redirect url use regular expression
      from dc_crawler.CollectURLs import CollectURLs

      filters = copy.deepcopy(inputFilters)
      if filters is not None:
        for inputFilter in filters:
          if inputFilter.stage == Filters.STAGE_ALL or inputFilter.stage == Filters.STAGE_REDIRECT_URL:
            inputFilter.stage = Filters.STAGE_COLLECT_URLS

      ret = CollectURLs.filtersApply(filters, url, 0, None, self.siteId, None, Filters.OC_RE, Filters.STAGE_COLLECT_URLS, siteProperties=self.siteProperties,
                                     url=url)

    return ret


  # # Extract content-type from requests response headers
  #
  # @param response - requests response object
  # @return content-type string
  @staticmethod
  def getContentType(response):
    # variable for result
    ret = ''
    try:
      if response is not None and response.headers is not None:
        ret = response.headers.get(RequestsRedirectWrapper.CONTENT_TYPE_HEADER_NAME, '').lower().split(";")[0]
    except Exception, err:
      logger.error(str(err))

    return ret

  # # Next two callback function awhile not use, but possible can use in future.
  #   Now, they show different samples of usage hooks.

#   # # check redirect max allowed count
#   #
#   # @param handler - RequestsRedirectWrapper instance reference
#   @staticmethod
#   def checkRedirectMax(handler, *args, **kwargs):
#     logger.debug('handler: %s', varDump(handler))
#     logger.debug('args = ' + str(args))
#     logger.debug('kwargs = ' + str(kwargs))
#
#     if handler is not None:
#       handler.redirectCount += 1
#       if handler.redirectCount > handler.maxRedirects:
#         raise requests.exceptions.TooManyRedirects('Exceeded %s redirects.' % str(handler.maxRedirects))


  # # check redirect
  #
  # @param r - result object instance
  @staticmethod
  def checkRedirect(r, *args, **kwargs):
    # logger.debug('r: %s', varDump(r))
    logger.debug('args = %s', str(args))
    logger.debug('kwargs = %s', str(kwargs))
    logger.debug('r.url: %s', str(r.url))
    logger.debug('r.status_code = %s', str(r.status_code))
