"""
HCE project,  Python bindings, Distributed Tasks Manager application.
TagLimitsHandler Class content main functional for check tags limits

@package: dc_processor
@file TagLimitsHandler.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import re

import app.Utils as Utils
from dc_processor.scraper_result import Result


class TagLimitsHandler(object):
  # # Constants used in class

  # # Constants of error messages used in class
  ERROR_MSG_WRONG_DATA_TYPE = "Initialization failed. Loaded % from input property %s"
  ERROR_MSG_WRONG_INPUT_DATA = "Wrong input data %s for '%s'"

  # # Internal class
  class TagLimits(object):
    # # Constants used in class
    OPTION_MIN_BYTES = "min_bytes"
    OPTION_MAX_BYTES = "max_bytes"
    OPTION_DEFAULT = "default"
    OPTION_TRUNCATE = "truncate"

    TRUNCATE_VALUE_OFF = 0
    TRUNCATE_VALUE_ON = 1

    DEFAULT_VALUE_TRUNCATE = TRUNCATE_VALUE_OFF


    def __init__(self):
      self.minBytes = None
      self.maxBytes = None
      self.default = None
      self.truncate = self.DEFAULT_VALUE_TRUNCATE


  # Initialization
  def __init__(self):
    pass


  # # Get limits
  #
  # @param propertiesDict - properties dictionary
  # @param tagName - tag name
  # @param url - URL string
  # @param log - logger instance
  # @return TagLimits instance or None
  @staticmethod
  def getLimits(propertiesDict, tagName, url, log=None):
    # variable for result
    ret = None
    try:
      if not isinstance(tagName, basestring):
        raise Exception(TagLimitsHandler.ERROR_MSG_WRONG_INPUT_DATA % (str(tagName), 'tagName'))

      if not isinstance(url, basestring):
        raise Exception(TagLimitsHandler.ERROR_MSG_WRONG_INPUT_DATA % (str(url), 'url'))

      if not isinstance(propertiesDict, dict):
        raise Exception(TagLimitsHandler.ERROR_MSG_WRONG_DATA_TYPE % (str(propertiesDict), str(propertiesDict)))

      for elemTagName, elemTagValue in propertiesDict.items():
        if elemTagName.lower() == tagName.lower() and isinstance(elemTagValue, dict):
          for pattern, limitsDict in  elemTagValue.items():
            if re.search(pattern=pattern, string=url, flags=re.I + re.U) is not None and isinstance(limitsDict, dict):
              # make result object
              ret = TagLimitsHandler.TagLimits()
              if TagLimitsHandler.TagLimits.OPTION_MIN_BYTES in limitsDict:
                ret.minBytes = int(limitsDict[TagLimitsHandler.TagLimits.OPTION_MIN_BYTES])

              if TagLimitsHandler.TagLimits.OPTION_MAX_BYTES in limitsDict:
                ret.maxBytes = int(limitsDict[TagLimitsHandler.TagLimits.OPTION_MAX_BYTES])

              if TagLimitsHandler.TagLimits.OPTION_DEFAULT in limitsDict:
                ret.default = str(limitsDict[TagLimitsHandler.TagLimits.OPTION_DEFAULT])

              if TagLimitsHandler.TagLimits.OPTION_TRUNCATE in limitsDict:
                ret.truncate = int(limitsDict[TagLimitsHandler.TagLimits.OPTION_TRUNCATE])

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  ## Apply limits
  #
  # @param tagLimits - TagLimits instance
  # @param tagName - tag name
  # @param response - scraper Result instance for apply limits
  # @param log - logger instance
  # @return modified scraper Result instance
  @staticmethod
  def applyLimits(tagLimits, tagName, response, log=None):
    try:
      if not isinstance(tagLimits, TagLimitsHandler.TagLimits):
        raise Exception(TagLimitsHandler.ERROR_MSG_WRONG_INPUT_DATA % (str(tagLimits), 'tagLimits'))

      if isinstance(response, Result) and isinstance(response.tags, dict):
        for elemTagName, elemTagValue in response.tags.items():
          if elemTagName.lower() == tagName.lower() and isinstance(elemTagValue, dict) and Result.TAGS_DATA_NAME in elemTagValue:
            if log is not None:
#               log.debug("data: %s", str(elemTagValue[Result.TAGS_DATA_NAME]))
              log.debug("data len = %s", str(len(elemTagValue[Result.TAGS_DATA_NAME])))

            if isinstance(elemTagValue[Result.TAGS_DATA_NAME], basestring):
              dataLength = len(elemTagValue[Result.TAGS_DATA_NAME])

              if (isinstance(tagLimits.minBytes, int) and tagLimits.minBytes > 0 and tagLimits.minBytes > dataLength) or \
                (isinstance(tagLimits.maxBytes, int) and tagLimits.maxBytes > 0 and tagLimits.maxBytes < dataLength):
                if isinstance(tagLimits.default, basestring):
                  elemTagValue[Result.TAGS_DATA_NAME] = tagLimits.default
                  if log is not None:
                    log.debug("Set default value, because not passed limits.")
                elif isinstance(tagLimits.truncate, int) and tagLimits.truncate == TagLimitsHandler.TagLimits.TRUNCATE_VALUE_ON and \
                  isinstance(tagLimits.maxBytes, int) and tagLimits.maxBytes > 0:
                  elemTagValue[Result.TAGS_DATA_NAME] = elemTagValue[Result.TAGS_DATA_NAME][:tagLimits.maxBytes]
                  if log is not None:
                    log.debug("Truncate value by max bytes limits == %s", str(tagLimits.maxBytes))
                else:
                  del response.tags[elemTagName]
                  if log is not None:
                    log.debug("Remove tag '%s' as not passed limits.")

            break

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return response


  # # tag limits execution
  #
  # @param siteProperties - site properties dictionary
  # @param url - URL string
  # @param response - scraper Result instance
  # @param log - logger instance
  # @return modified scraper Result instance
  @staticmethod
  def execute(siteProperties, url, response, log=None):
    try:
      if not isinstance(response, Result):
        raise Exception(TagLimitsHandler.ERROR_MSG_WRONG_INPUT_DATA % (str(response), 'response'))

      propertiesDict = Utils.jsonLoadsSafe(jsonString=siteProperties, default=siteProperties, log=log)
      if isinstance(propertiesDict, dict):
        for tagName in propertiesDict.keys():
          limitsObj = TagLimitsHandler.getLimits(propertiesDict=propertiesDict, tagName=tagName, url=url, log=log)
    #       if log is not None:
    #         log.debug("limitsObj: %s", Utils.varDump(limitsObj, stringifyType=0))

          if limitsObj is not None:
            response = TagLimitsHandler.applyLimits(tagLimits=limitsObj, tagName=tagName, response=response, log=log)

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return response
