# coding: utf-8
"""@package docstring
 @file scrapy_extractor.py
 @author Alexey <developers.hce@gmail.com>
 @link http://hierarchical-cluster-engine.com/
 @copyright Copyright &copy; 2013 IOIX Ukraine
 @license http://hierarchical-cluster-engine.com/license/
 @package HCE project node API
 @since 0.1
"""

import re
import json
import copy
import ConfigParser
import dc_processor.Constants as CONSTS
from dc_processor.base_extractor import BaseExtractor
from app.SelectorWrapper import SelectorWrapper
from app.Utils import varDump
from app.Utils import ExceptionLog
import app.Utils as Utils  # pylint: disable=F0401
from app.Exceptions import ScraperTemplateException
from app.Url import Url
from app.DateTimeType import DateTimeType

# Logger initialization
logger = Utils.MPLogger().getLogger()


# #ScrapyExtractor class implements data extracting using Scrapy module with prepared XPathes/Css
#
class ScrapyExtractor(BaseExtractor):

  # # Constants used in class
  TEMPLATE_FILE_RULE_XPATH = 'xpath'
  TEMPLATE_FILE_RULE_REPLACE = 'replace'
  TEMPLATE_FILE_RULE_EXCLUDE = 'exclude'
  TEMPLATE_FILE_RULE_PUBDATE_FORMAT = 'format'

  DISABLE_XPATH_CHARS_LIST = [';', '#']

  # #class constructor
  #
  # @param config - incoming app config
  # @param templ - optionality dict with base set of extractor rules/xpathes
  # @param domain - optionality param, processing url domain
  # @param processorProperties - optionality param, incoming app processor properties
  # @param template - optionality param, incoming template set
  def __init__(self, config, templ=None, domain=None, processorProperties=None):
    BaseExtractor.__init__(self, config, templ, domain, processorProperties)
    logger.debug("Properties: %s", varDump(self.properties))

    self.setName(CONSTS.EXTRACTOR_SCRAPY_NAME)
    # set module rank from module's properties
    self.rankReading(self.__class__.__name__)

    self.closeVoid = None
    if processorProperties is not None and CONSTS.TAG_CLOSE_VOID_PROP_NAME in processorProperties and \
      processorProperties[CONSTS.TAG_CLOSE_VOID_PROP_NAME] is not None:
      self.closeVoid = int(processorProperties[CONSTS.TAG_CLOSE_VOID_PROP_NAME])

    self.keepAttributes = None
    if processorProperties is not None and CONSTS.TAG_KEEP_ATTRIBUTES_PROP_NAME in processorProperties and \
      processorProperties[CONSTS.TAG_KEEP_ATTRIBUTES_PROP_NAME] is not None:
      self.keepAttributes = {}
      for key in processorProperties[CONSTS.TAG_KEEP_ATTRIBUTES_PROP_NAME]:
        self.keepAttributes[key.lower()] = processorProperties[CONSTS.TAG_KEEP_ATTRIBUTES_PROP_NAME][key]

    self.innerTextTagReplacers = None
    if processorProperties is not None and CONSTS.TAG_MARKUP_PROP_NAME in processorProperties and \
      processorProperties[CONSTS.TAG_MARKUP_PROP_NAME] is not None:
      self.innerTextTagReplacers = {}
      for key in processorProperties[CONSTS.TAG_MARKUP_PROP_NAME]:
        self.innerTextTagReplacers[key.lower()] = processorProperties[CONSTS.TAG_MARKUP_PROP_NAME][key]

    self.sel = None
    self.resource = None
    # for post processing
    self.postReplace = {}
    self.postExclude = {}
    # for pubdate format
    self.pubdateFormat = {}
    # for dc_date format
    self.dcdateFormat = {}

    if processorProperties is not None and "SCRAPER_SCRAPY_PRECONFIGURED" in processorProperties:
      self.templates = self.generateTemplatesFromRowTemplates(json.loads(processorProperties\
                                                                         ["SCRAPER_SCRAPY_PRECONFIGURED"]), domain)
    else:
      self.templates = [{self.name + "_default": self.templateLoad(config, templ, domain)}]

    try:
      defaultConfigTemplate = config.get("Application", "default_template", None)
    except ConfigParser.NoOptionError:
      defaultConfigTemplate = None
    if defaultConfigTemplate is not None:
      logger.debug(">>> Extend Templates with config default template")
      tempTemplates = self.generateTemplatesFromRowTemplates(json.loads(defaultConfigTemplate), domain)
      if len(tempTemplates) > 0:
        newTemplates = []
        for templeteElemConfig in tempTemplates:
          for templeteElemProperty in self.templates:
            for templeteKeyProperty in templeteElemProperty:
              if templeteKeyProperty in templeteElemConfig:
                templeteElemConfig = None
                break
            if templeteElemConfig is None:
              break
          if templeteElemConfig is not None:
            newTemplates.append(templeteElemConfig)
        self.templates = self.templates + newTemplates
    self.blockedByXpathTags = []
    logger.debug("!!! INIT Template Domain: '%s'", str(domain))
#     logger.debug("!!! INIT Template: %s", str(self.templates))


  # #generateTemplatesFromRowTemplates method extract templates from incoming rowTemplates data
  #
  # @param rowTemplates - incoming rowTemplates
  # @return result - list of template dicts
  def generateTemplatesFromRowTemplates(self, rowTemplates, domain=None):
    ret = []
    try:
      if "sets" in rowTemplates:
        ret = rowTemplates["sets"]
        for elem in ret:
          for setName in elem:
            if isinstance(elem[setName], basestring):
              try:
                with open(elem[setName], "rb") as fd:
                  elem[setName] = json.loads(fd.read())
              except Exception as excp:
                logger.debug(">>> generateTemplatesFromRowTemplates element[%s] file/json operations error, %s",
                             setName, str(type(elem[setName])))
                elem[setName] = {}
            elif not isinstance(elem[setName], dict):
              logger.debug(">>> generateTemplatesFromRowTemplates element[%s] wrong type is %s", setName,
                           str(type(elem[setName])))
              elem[setName] = {}

            elem[setName] = self.templatePreparer(None, domain, elem[setName])
            break
    except ScraperTemplateException, err:
      raise err
    except Exception as excp:
      logger.debug(">>> Some error during generateTemplatesFromRowTemplates = " + str(excp))
    return ret


  # #templateLoad method which fills internal template dict with preparatory extractor rules/xpathes
  #
  # @param config - incoming app config
  # @param templ - optionality dict with base set of extractor rules/xpathes
  # @param domain - optionality param, processing url domain
  # @return result template dict
  def templateLoad(self, config, templ=None, domain=None):
    logger.debug("templ: %s, domain: '%s'", str(templ), str(domain))

    ret = {}
    defaultTemplate = None
    try:
      templateFile = config.get("Application", "template", None)
    except ConfigParser.NoOptionError:
      templateFile = None
    if templateFile:
      try:
        logger.debug("Read template from file. %s", templateFile)
        with open(templateFile, "rb") as fd:
          defaultTemplate = self.templatePreparer(fd.read(), domain, {})
      except ScraperTemplateException, err:
        raise err
      except Exception, err:
        logger.error("Error Read template from file. %s", str(err))

    if self.properties is not None and CONSTS.TEMPLATE_KEY not in self.properties:
      ret = self.templatePreparer(self.properties[CONSTS.TEMPLATE_KEY], domain, {})
    elif templ is not None:
      if isinstance(templ, dict):
        ret = self.templatePreparer(None, domain, templ)
      else:
        ret = self.templatePreparer(templ, domain, {})

      # merge default template and custom one
      if defaultTemplate is not None:
        logger.debug("merge default template and custom one")
        defaultTags = defaultTemplate.keys()
        customTags = ret.keys()
        logger.debug("tags in default template:\n%s\nin custom template:\n%s", str(defaultTags), str(customTags))
        for tag in defaultTags:
          if tag not in customTags:
            ret[tag] = defaultTemplate[tag]
            logger.debug("%s was replaced from custom template", str(tag))
    elif defaultTemplate is not None:
      ret = defaultTemplate
    else:
      logger.error("Error Read template.")
    return ret


  # #pasteLists pastes same elements in 2 incoming dicts
  #
  # @param lhs - incoming destination dict
  # @param rhs - incoming source dict
  def pasteLists(self, lhs, rhs):
#     logger.debug("lhs: %s, type: %s", str(lhs), str(type(lhs)))
#     logger.debug("rhs: %s, type: %s", str(rhs), str(type(rhs)))

    if isinstance(lhs, dict) and isinstance(rhs, dict):
      for elem in rhs:

        self.postReplace[elem] = []
        if elem in lhs and self.TEMPLATE_FILE_RULE_REPLACE in lhs[elem] and isinstance(lhs[elem][self.TEMPLATE_FILE_RULE_REPLACE], dict):
          self.postReplace[elem].append(lhs[elem][self.TEMPLATE_FILE_RULE_REPLACE])
#           logger.debug("!!! lhs self.postReplace: %s", str(lhs[elem][self.TEMPLATE_FILE_RULE_REPLACE]))
#           logger.debug("!!! self.postReplace: %s", str(self.postReplace))

        if elem in rhs and self.TEMPLATE_FILE_RULE_REPLACE in rhs[elem] and isinstance(rhs[elem][self.TEMPLATE_FILE_RULE_REPLACE], dict):
          self.postReplace[elem].append(rhs[elem][self.TEMPLATE_FILE_RULE_REPLACE])
#           logger.debug("!!! rhs self.postReplace: %s", str(rhs[elem][self.TEMPLATE_FILE_RULE_REPLACE]))
#           logger.debug("!!! self.postReplace: %s", str(self.postReplace))

        self.postExclude[elem] = []
        if elem in lhs and self.TEMPLATE_FILE_RULE_EXCLUDE in lhs[elem] and isinstance(lhs[elem][self.TEMPLATE_FILE_RULE_EXCLUDE], list):
          self.postExclude[elem].extend(lhs[elem][self.TEMPLATE_FILE_RULE_EXCLUDE])

        if elem in rhs and self.TEMPLATE_FILE_RULE_EXCLUDE in rhs[elem] and isinstance(rhs[elem][self.TEMPLATE_FILE_RULE_EXCLUDE], list):
          self.postExclude[elem].extend(rhs[elem][self.TEMPLATE_FILE_RULE_EXCLUDE])

#         logger.debug("!!! self.postExclude['%s']: %s", str(elem), str(self.postExclude[elem]))

        self.pubdateFormat[elem] = []
        if elem in lhs and self.TEMPLATE_FILE_RULE_PUBDATE_FORMAT in lhs[elem] and isinstance(lhs[elem][self.TEMPLATE_FILE_RULE_PUBDATE_FORMAT], list):
          self.pubdateFormat[elem].extend(lhs[elem][self.TEMPLATE_FILE_RULE_PUBDATE_FORMAT])

        if elem in rhs and self.TEMPLATE_FILE_RULE_PUBDATE_FORMAT in rhs[elem] and isinstance(rhs[elem][self.TEMPLATE_FILE_RULE_PUBDATE_FORMAT], list):
          self.pubdateFormat[elem].extend(rhs[elem][self.TEMPLATE_FILE_RULE_PUBDATE_FORMAT])

        lXpathList = []
        rXpathList = []

        if elem in lhs and isinstance(lhs[elem], dict) and self.TEMPLATE_FILE_RULE_XPATH in lhs[elem] and \
          isinstance(lhs[elem][self.TEMPLATE_FILE_RULE_XPATH], list):
          lXpathList = lhs[elem][self.TEMPLATE_FILE_RULE_XPATH]

        if elem in lhs and isinstance(lhs[elem], list):
          lXpathList = lhs[elem]

        if elem in rhs and isinstance(rhs[elem], dict) and self.TEMPLATE_FILE_RULE_XPATH in rhs[elem] and \
          isinstance(rhs[elem][self.TEMPLATE_FILE_RULE_XPATH], list):
          rXpathList = rhs[elem][self.TEMPLATE_FILE_RULE_XPATH]

        if elem in rhs and isinstance(rhs[elem], list):
          rXpathList = rhs[elem]

#         logger.debug("!!! lXpathList: %s", varDump(lXpathList))
#         logger.debug("!!! rXpathList: %s", varDump(rXpathList))

        lhs[elem] = lXpathList + rXpathList
#         logger.debug("!!! lhs[elem]: %s", varDump(lhs[elem]))


  # # Common method of prepared templates extract
  #
  # @param jsonBuf - incoming json with templates
  # @param domains - domains list
  # @param globalTemplate - incoming templates
  # @return template dict, that corresponds incoming domainCrc or "*"
  def templatePreparer(self, jsonBuf, domains, globalTemplate):
    ret = {}
    if len(globalTemplate) == 0:
      try:
        globalTemplate = Utils.duplicateKeyCheck(jsonBuf)
      except Exception, err:
        erroMsg = "Wrong json format. %s" % str(err)
        logger.error(erroMsg)
        raise ScraperTemplateException(erroMsg)

    if len(globalTemplate) > 0:
      try:
        if domains is not None:
#           logger.debug("!!! domains: '%s', type: %s", str(domains), str(type(domains)))
#           logger.debug("!!! globalTemplate: '%s'", str(globalTemplate))
#           logger.debug("!!! type(globalTemplate): '%s'", str(type(globalTemplate)))
          if isinstance(domains, basestring):
            domains = [domains]

          for domain in domains:
            found = False
#             logger.info("!!! globalTemplate after sort: %s", varDump(sorted(globalTemplate.keys(), key=len, reverse=True), stringifyType=0))
            for pattern in sorted(globalTemplate.keys(), key=len, reverse=True):
              try:
                searchPatterns = pattern.split()
#                 logger.debug("!!! searchPatterns: '%s'", str(searchPatterns))
                for searchPattern in searchPatterns:
                  if searchPattern != '*':
                    if re.search(searchPattern, domain, re.UNICODE + re.I) is not None:
#                       logger.debug("!!! Found searchPattern: '%s'", str(searchPattern))
#                       logger.debug("!!! Found pattern: '%s'", str(pattern))
#                       logger.debug("!!! globalTemplate[pattern]: %s, type = %s", str(globalTemplate[pattern]), str(type(globalTemplate[pattern])))

                      if isinstance(globalTemplate[pattern], dict):
                        ret = globalTemplate[pattern]
                        found = True
#                         logger.debug("!!! Set found = True, break...")
                        break

                if found:
#                   logger.debug("!!! found is True (for pattern in globalTemplate:), break...")
                  break
              except Exception, err:
                logger.debug("Regular expression error: %s, pattern: '%s', domain: '%s'", str(err), str(pattern), str(domain))

            if found:
#               logger.debug("!!! found is True (for domain in domains:), break...")
              break

            # If was fail use old algorithm
            if len(ret) == 0 and domain in globalTemplate and isinstance(globalTemplate[domain], dict):
              ret = globalTemplate[domain]

        if domains is not None:
          for domain in domains:
            if len(ret) == 0:
              while domain.find(".") != -1:
                domain = domain[domain.find(".") + 1: len(domain)]
                if domain is not None and domain in globalTemplate:
                  self.pasteLists(ret, globalTemplate[domain])
              if domain is not None and domain in globalTemplate:
                self.pasteLists(ret, globalTemplate[domain])

        domain = "*"
        if domains is not None and domain in globalTemplate:
          self.pasteLists(ret, globalTemplate[domain])

      except Exception, err:
        ExceptionLog.handler(logger, err, 'Exception: ', (ret))

    for key, value in ret.items():
      if isinstance(value, list):
        removeList = []
        for elemXPath in value:
          if elemXPath != "" and elemXPath[0] in self.DISABLE_XPATH_CHARS_LIST:
            removeList.append(elemXPath)

        for removeElem in removeList:
          value.remove(removeElem)
          logger.debug("For '%s' found disabled xpath: %s", str(key), str(removeElem))

#     logger.info("!!! ret: %s", varDump(ret, stringifyType=0))
    return ret


  # #extractTag method extracts for concrete tag
  #
  # @param tagName - incoming tag name
  # @param result - incoming result object
  # @param textHandler - optionality param with text processing callback function
  # @param delimiter - optionality delimiter between extracted elements
  def extractTag(self, tagName, result, template, textHandler=None, delimiter=' '):
    try:
      if tagName in template:
#         logger.debug("!!! tagName: %s, xpath list: %s", str(tagName), str(template[tagName]))
        for path in template[tagName]:
#           logger.debug("!!! Try usage xpath: '%s'", str(path))

          if tagName in self.blockedByXpathTags:
            break

          if path == "":
            if tagName not in result.blockedByXpathTags:
              result.blockedByXpathTags.append(tagName)
            break
          elif path == "none":
            if tagName not in self.blockedByXpathTags:
              self.blockedByXpathTags.append(tagName)
            break

#           logger.debug("!!! self.tagsValidator = %s, self.name = %s", str(self.tagsValidator), str(self.name))
#           logger.debug("!!! self.sel = %s", str(self.sel))
#           logger.debug("!!! self.sel.xpath(path): '%s'", str(self.sel.xpath(path)))

          try:
            if textHandler is not None:
              conditions = None
              if self.tagsValidator is not None and self.name in self.tagsValidator and \
              tagName in self.tagsValidator[self.name]:
                conditions = self.tagsValidator[self.name][tagName]
              localValue = textHandler(self.sel.xpath(path), delimiter, delimiter, self.innerTextTagReplacers, conditions,
                                       keepAttributes=self.keepAttributes, baseUrl=self.resource.url,
                                       closeVoid=self.closeVoid, excludeNodes=self.postExclude[tagName] if tagName in self.postExclude else None)
            else:
              localValue = self.sel.xpath(path).extract()
          except Exception, err:
            logger.error("Apply xpath for tag '%s' failed. Error: %s", str(tagName), str(err))
            continue

#           if tagName == 'content_encoded' and localValue != "":
#             logger.debug("!!! tagName: %s", str(tagName))
#             logger.debug("!!! xpath: %s", str(path))
#             logger.debug("!!! value: '%s'", varDump(localValue))

#           if tagName == 'title' and localValue != '':  # or tagName == 'html_lang':
#             logger.debug("!!! tagName: %s", str(tagName))
#             logger.debug("!!! xpath: %s", str(path))
#             logger.debug("!!! value: '%s'", json.dumps(localValue, ensure_ascii=False, encoding='utf-8'))
#             logger.debug("!!! self.sel: %s", str(self.sel))

          # apply post-processing
          if isinstance(self.postReplace, dict) and tagName in self.postReplace and \
            isinstance(self.postReplace[tagName], list) and localValue != "":
#             if len(localValue) > 0:
#               logger.info("POST PROCESSING FOR TAG '%s', len = %s", str(tagName), len(localValue))
#               logger.debug("!!! localValue before: '%s'", json.dumps(localValue, ensure_ascii=False, encoding='utf-8'))

            for postReplace in self.postReplace[tagName]:
#               logger.info("postReplace: %s", str(postReplace))

              if isinstance(postReplace, dict):
                isReplaced = False
                for pattern, repl in postReplace.items():
                  if isinstance(pattern, basestring) and isinstance(repl, basestring) and localValue != "":
                    try:
                      logger.debug("!!! pattern: '%s', repl: '%s'", str(pattern), str(repl))
                      localValue = re.sub(pattern=pattern.decode('utf-8'), repl=repl, string=localValue.decode('utf-8'), count=1, flags=re.U + re.M + re.I)
                      logger.debug("Value after replace: %s", str(localValue))
                      logger.debug("Success!!!")
                      isReplaced = True
                    except Exception, err:
                      logger.error("RE error: %s", str(err))
                      logger.info("pattern: '%s', repl: '%s'", str(pattern), str(repl))

                if isReplaced:
                  break

          if isinstance(localValue, basestring):
            localValue = localValue.decode('utf-8')

          if tagName == 'title' and localValue != '':
            logger.debug("!!! localValue: '%s'", json.dumps(localValue, ensure_ascii=False, encoding='utf-8'))

          if tagName == CONSTS.TAG_LINK:
            urlObj = Url(localValue)
            if urlObj.isValid():
              self.addTag(result=result, tag_name=tagName, tag_value=localValue, xpath=path)

          elif tagName == CONSTS.TAG_PUB_DATE or tagName == CONSTS.TAG_DC_DATE:
#             logger.info("!"*50 + ' %s ' % str(tagName).upper() + "!"*50)
#             logger.debug("!!! localValue: %s", str(localValue))
#             logger.debug("!!! self.pubdateFormat:\n%s", varDump(self.pubdateFormat, stringifyType=0))

            formatList = self.pubdateFormat[tagName]
            if isinstance(formatList, list) and len(formatList) > 0 and isinstance(localValue, list) and len(localValue) > 0:
              for dataString in localValue:
                if isinstance(dataString, basestring) and dataString != "":
                  dateTimeType = DateTimeType(dataString=dataString, formatObj=formatList, continueIfNotDetected=False, log=None)
                  if dateTimeType.isError:
                    logger.error("Extract tag '%s' by format %s failed. Error: %s", str(tagName), str(formatList), str(dateTimeType.errorMsg))
                  else:
                    localValue = dateTimeType.getString()
                    logger.debug("Extracted tag '%s' by format %s has value: %s", str(tagName), str(formatList), str(localValue))
                    if localValue is None:
                      localValue = ''
                    self.addTag(result=result, tag_name=tagName, tag_value=localValue, xpath=path)
            else:
              self.addTag(result=result, tag_name=tagName, tag_value=localValue, xpath=path)

#             logger.info("!"*50 + ' FINISHED ' + "!"*50)
          else:
            self.addTag(result=result, tag_name=tagName, tag_value=localValue, xpath=path)

#           if isinstance(localValue, basestring) and localValue != "":
#             logger.debug("Success for xpath '%s', value: %s", str(path), varDump(localValue))

    except Exception, err:
      ExceptionLog.handler(logger, err, "Extract tag '%s' failed:" % str(tagName))


  # #extractTagsForOneTemplate method extract data by tags for one template and fills incoming result object
  #
  # @param resource - incoming raw data
  # @param result - incoming result object
  # @param template - current template
  # @return incoming result with additionally filled fields/tags
  def extractTagsForOneTemplate(self, resource, result, template):
    try:
      self.resource = resource
#       logger.debug("URL: %s \nresource.raw_html:\n%s ", str(self.resource.url), varDump(resource.raw_html, strTypeMaxLen=500))
      self.sel = SelectorWrapper(text=resource.raw_html)

#       logger.debug("Selector varDump(self.sel):\n%s ", varDump(self.sel, stringifyType=0))
#       logger.debug("Selector str(self.sel): %s ", str(self.sel))

      # search engine parsing
#       logger.debug("Regular parsing")
      self.extractTag(CONSTS.TAG_TITLE, result, template, Utils.innerText)
      self.extractTag(CONSTS.TAG_AUTHOR, result, template, Utils.innerText)
      self.extractTag(CONSTS.TAG_PUB_DATE, result, template)
      self.extractTag(CONSTS.TAG_DESCRIPTION, result, template, Utils.innerText)
      self.extractTag(CONSTS.TAG_DC_DATE, result, template)
      self.extractTag(CONSTS.TAG_MEDIA, result, template, Utils.innerText, self.imgDelimiter)
      self.extractTag(CONSTS.TAG_LINK, result, template, Utils.innerText)
      self.extractTag(CONSTS.TAG_CONTENT_UTF8_ENCODED, result, template, Utils.innerText, delimiter='')

      # for path in template["enclosure"]]
      self.extractTag(CONSTS.TAG_KEYWORDS, result, template, Utils.innerText)
      # Add support of html_lang tag
      self.extractTag(CONSTS.HTML_LANG, result, template, Utils.innerText)
      # Add support of html_charset tag
      self.extractTag(CONSTS.TAG_HTML_CHARSET, result, template, Utils.innerText)

    except Exception as err:
      ExceptionLog.handler(logger, err, "Parse error:", (err))

    return result


  # #extractTags public method extract data by tags and fills incoming result object
  #
  # @param resource - incoming raw data
  # @param result - incoming result object
  # @return incoming result with additionally filled fields/tags
  def extractTags(self, resource, result):
    self.blockedByXpathTags = []
    localResults = []
#     logger.info("!!! self.templates: %s", str(self.templates))

    for templateDict in self.templates:
      for templateName in templateDict:
        localResult = copy.deepcopy(result)
        self.extractTagsForOneTemplate(resource, localResult, templateDict[templateName])
        localResults.append(localResult)
        break

    for localResult in localResults:
      result.mergeResults(localResult)

    return result
