'''
HCE project, Python bindings, Distributed Tasks Manager application.
ResourcesStateMonitor object and related classes definitions.
This object acts as monitor of resources of HCE Cluster.
Periodically reads cluster schema from json schema file and make DRCE synch task request for all "replica" or
"shard" nodes. The DRCE task request is prepared and stored in the json external file that read each time when
monitoring cycle happened. The DRCE synch task acts as set of Linux commands and collects system information like
AVG CPU, disk, memory usage and so on. This information parsed and some system resource usage indicators detected.
This indicators updated the resource database that used for tasks planning.


@package: dtm
@author bgv bgv.hce@gmail.com
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
'''


import json
import logging
import re

import admin.Command
import admin.Node
import admin.NodeManagerRequest
from app.BaseServerManager import BaseServerManager
from app.LogFormatter import LogFormatterEvent
from drce.CommandConvertor import CommandConvertor
import dtm.EventObjects
from transport.ConnectionBuilderLight import ConnectionBuilderLight
from transport.UIDGenerator import UIDGenerator
import dtm.Constants as DTM_CONSTS
import transport.Consts as TRANSPORT_CONSTS
import app.Utils
from app.Utils import varDump

# Logger initialization
# logger = logging.getLogger(__name__)
logger = logging.getLogger(DTM_CONSTS.LOGGER_NAME)



# #The ResourcesStateMonitor class, is a monitor of resources of HCE Cluster.
#
# This object acts as monitor of resources of HCE Cluster.
# Periodically reads cluster schema from json schema file and make DRCE synch task request for all "replica" or
# "shard" nodes. The DRCE task request is prepared and stored in the json external file that read each time when
# monitoring cycle happened. The DRCE synch task acts as set of Linux commands and collects system information like
# AVG CPU, disk, memory usage and so on. This information parsed and some system resource usage indicators detected.
# This indicators updated the resource database that used for tasks planning.
#
class ResourcesStateMonitor(BaseServerManager):

  # Configuration settings options names
  CONFIG_RESOURCES_MANAGER_CLIENT = "clientResourcesManager"
  CONFIG_HCE_NODE_ADMIN_TIMEOUT = "HCENodeAdminTimeout"
  CONFIG_UPDATE_RESOURCES_DRCE_JSON = "FetchResourcesStateDRCEJsonFile"
  CONFIG_HCE_CLUSTER_SCHEMA_FILE = "HCEClusterSchemaFile"
  CONFIG_POLLING_TIMEOUT = "PollingTimeout"

  ERROR_UPDATE_RESOURCES = "Not all resources updated!"
  ERROR_JSON_DECODE = "Json decode error!"
  ERROR_RESOURCES_LISTS_NOT_EQUAL = "Lists of sent and received messages not equal!"
  ERROR_READ_FILE = "Error read file"
  ERROR_READ_DRCE_JSON_FILE = "Error read DRCE update resources state request json file"
  ERROR_HCE_RESPONSE_PROCESSING_EXCEPTION = "HCE node Admin API response processing exception"
  ERROR_HCE_CLUSTER_SCHEMA_STRUCTURE_WRONG = "Wrong structure of HCE cluster schema object"
  ERROR_HCE_NODE_REQUEST_ERROR = "HCE node request error"
  ERROR_HCE_RESPONSE_PROCESSING_FORMAT_ADMIN = "HCE node response format error, cant to split"
  ERROR_HCE_RESPONSE_PROCESSING_NO_RESOURCE_IN_RESPONSE = "No resource data found in response or response with error"
  ERROR_NO_ITEMS_IN_DRCE_RESPONSE = "No items in DRCE response"
  ERROR_DRCE_RESPONSE_ERROR_CODE = "DRCE response with error"
  ERROR_RE_PARSE_NO_MATCHES = "RE parse results from DRCE node, no matches!"


  # #constructor
  # initialize fields
  #
  # @param configParser config parser object
  # @param connectBuilderLight connection builder light
  #
  def __init__(self, configParser, connectionBuilderLight=None):
    super(ResourcesStateMonitor, self).__init__()

    # Instantiate the connection builder light if not set
    if connectionBuilderLight == None:
      connectionBuilderLight = ConnectionBuilderLight()

    className = self.__class__.__name__

    # Get configuration settings
    self.clientResourcesManagerName = configParser.get(className, self.CONFIG_RESOURCES_MANAGER_CLIENT)
    self.resourcesUpdateDRCERequestJsonFile = configParser.get(className, self.CONFIG_UPDATE_RESOURCES_DRCE_JSON)
    self.hceClusterSchemaFile = configParser.get(className, self.CONFIG_HCE_CLUSTER_SCHEMA_FILE)

    # Set connections poll timeout, defines period of HCE cluster monitoring cycle
    self.configVars[self.POLL_TIMEOUT_CONFIG_VAR_NAME] = configParser.getint(className, self.CONFIG_POLLING_TIMEOUT)

    # Create connections and raise bind or connect actions for correspondent connection type
    resourcesManagerConnection = connectionBuilderLight.build(TRANSPORT_CONSTS.CLIENT_CONNECT,
                                                              self.clientResourcesManagerName)

    # Add connections to the polling set
    self.addConnection(self.clientResourcesManagerName, resourcesManagerConnection)

    # Set event handler for EXECUTE_TASK event
    self.setEventHandler(DTM_CONSTS.EVENT_TYPES.UPDATE_RESOURCES_DATA_RESPONSE, self.onUpdateResourcesDataResponse)

    # Initialize HCE node Admin API
    self.hceNodeAdminTimeout = configParser.getint(className, self.CONFIG_HCE_NODE_ADMIN_TIMEOUT)
    self.hceNodeManagerRequest = admin.NodeManagerRequest.NodeManagerRequest()

    # Initialize unique Id generator
    self.drceIdGenerator = UIDGenerator()
    # Initialize DRCE commands convertor
    self.drceCommandConvertor = CommandConvertor()
    logger.debug("Constructor passed")


  # #onUpdateResourcesDataResponse event handler
  #
  # @param event instance of Event object
  def onUpdateResourcesDataResponse(self, event):
    # Get task Id from event
    generalResponse = event.eventObj
    # Get list of sent resources to update
    resourcesToUpdate = event.cookie
    if len(resourcesToUpdate) == len(generalResponse.statuses):
      errorObjects = []
      for i in range(len(generalResponse.statuses)):
        if generalResponse.statuses[i] == False:
          errorObjects.append(resourcesToUpdate[i])
          logger.error(LogFormatterEvent(event, errorObjects, self.ERROR_UPDATE_RESOURCES))
    else:
      logger.error(LogFormatterEvent(event, [], self.ERROR_RESOURCES_LISTS_NOT_EQUAL + "\n" +
                                     "send " + str(len(resourcesToUpdate)) + " returned " +
                                     str(len(generalResponse.statuses))
                                     ))



  # #Events wait timeout handler, for timeout state of the connections polling. Executes periodical processing of RSM
  # resources state monitoring
  #
  def on_poll_timeout(self):
    # Initialize list of resources objects to update
    resources = []

    # Get HCE Cluster schema
    clusterSchemaObj = self.getObjectFromJsonFile(self.hceClusterSchemaFile)
    # Get DRCE FO request to get resources state from file
    drceRequestJson = self.loadFromFile(self.resourcesUpdateDRCERequestJsonFile)
    if drceRequestJson is not None:
      # Get connected data nodes from schema
      nodes = self.getConnectedNodesFromSchema(clusterSchemaObj)
      # For each DRCE node execute HCE node admin request
      for node in nodes:
        # Send request to HCE node Admin API
        rawResponse = self.sendToHCENodeAdmin(node.host, node.port, drceRequestJson)
        if rawResponse is not None:
          try:
            # Split admin response parts
            logger.debug("Raw node response: " + str(rawResponse) + ", type = " + str(type(rawResponse)))
            parts = rawResponse.split(admin.Constants.COMMAND_DELIM)
            if len(parts) > 1:
              # Convert DRCE jason protocol response to TaskResponse object
              taskResponse = self.drceCommandConvertor.from_json(parts[1])
              logger.debug("Received taskResponse object: " + str(vars(taskResponse)))
              # Get resource info data from the TaskResponse object and create Resource object
              resource = self.getResourceFromTaskResponse(taskResponse)
              logger.debug("Received Resource object: " + str(varDump(resource)))
              # Collect resource if valid data detected
              if resource is not None:
                resources.append(resource)
              else:
                logger.error(self.ERROR_HCE_RESPONSE_PROCESSING_NO_RESOURCE_IN_RESPONSE)
            else:
              logger.error(self.ERROR_HCE_RESPONSE_PROCESSING_FORMAT_ADMIN)
          except Exception, e:
            logger.error(self.ERROR_HCE_RESPONSE_PROCESSING_EXCEPTION + " : " + str(e.message) + "\n" + \
                         app.Utils.getTracebackInfo())

      # Send update for collected resources objects
      if len(resources) > 0:
        self.sendUpdateResourceDataRequest(resources)
    else:
      logger.error(self.ERROR_READ_DRCE_JSON_FILE + " " + self.resourcesUpdateDRCERequestJsonFile)



  # #Get the Resource object created on the basis of information from the TaskResponse object of DRCE get resources data
  # request
  #
  # @param taskResponseObjects The TaskResponse objects container
  # @return the first Resource object if success or None if not
  def getResourceFromTaskResponse(self, taskResponseObjects):
    resource = None

    if len(taskResponseObjects.items) > 0:
      # logger.debug("Received taskResponseObjects.items[0]:" + str(vars(taskResponseObjects.items[0])))

      if taskResponseObjects.items[0].error_code == 0 or taskResponseObjects.items[0].error_code > 0:
        resource = dtm.EventObjects.Resource(taskResponseObjects.items[0].host + ":" +
                                             str(taskResponseObjects.items[0].port))
        resource.nodeName = taskResponseObjects.items[0].node
        resource.state = dtm.EventObjects.Resource.STATE_ACTIVE
        logger.debug("Received stdout:\n" + str(taskResponseObjects.items[0].stdout))
        reTemplate1 = r"---CPU LA---(.*)"\
                       "---cpu cores---(.*)"\
                       "---vmstat---(.*)"\
                       "---processes---(.*)"\
                       "---threads max---(.*)"\
                       "---threads actual---(.*)"\
                       "---RAM---(.*)"\
                       "---Disk---(.*)"\
                       "---uptime---(.*)"\
                       "---END---(.*)"
        groupsItems = re.match(reTemplate1, taskResponseObjects.items[0].stdout, re.M | re.I | re.S)
        if groupsItems:
          # CPU LA
          # cpuLA = int(float(groupsItems.group(1).split(" ")[0].lstrip()) * 100)
          # print "cpuLA*100%=" + str(cpuLA)

          cpuCores = 0
          try:
            # CPU cores
            logger.debug("CPU cores srs:\n" + str(groupsItems.group(2)))
            cpuCores = int(groupsItems.group(2))
            logger.debug("cpuCores:" + str(cpuCores))
          except Exception, e:
            logger.error("Error CPU cores: " + str(e))
          resource.cpuCores = cpuCores

          cpuLoad = 0
          try:
            # CPU Load
            logger.debug("CPU load srs:\n" + str(groupsItems.group(3)))
            lines = (' '.join(groupsItems.group(3).split("\n")[4].split())).split(' ')
            logger.debug("CPU load lines:\n" + str(lines))
            cpuIdle = int(lines[14])
            logger.debug("cpuIdle:" + str(cpuIdle))
            cpuLoad = 100 - cpuIdle
            if cpuLoad > 100:
              cpuLoad = 100
            else:
              if cpuLoad < 0:
                cpuLoad = 0
            logger.debug("cpuLoad:" + str(cpuLoad))
          except Exception, e:
            logger.error("Error CPU load: " + str(e))
          resource.cpu = cpuLoad

          cpuIO = 0
          try:
            # IO wait
            logger.debug("CPU IO srs:\n" + str(groupsItems.group(3)))
            lines = (' '.join(groupsItems.group(3).split("\n")[4].split())).split(' ')
            logger.debug("CPU IO lines:\n" + str(lines))
            cpuIO = int(lines[15])
            logger.debug("cpuIO:" + str(cpuIO))
          except Exception, e:
            logger.error("Error CPU IO: " + str(e))
          resource.io = cpuIO

          processes = 0
          try:
            # Processes
            logger.debug("CPU Processes srs:\n" + str(groupsItems.group(4)))
            processes = int(groupsItems.group(4).strip())
            logger.debug("processes:" + str(processes))
          except Exception, e:
            logger.error("Error CPU processes: " + str(e))
          resource.processes = processes

          # Threads max
          # threadsMax = int(groupsItems.group(5).lstrip().rstrip())
          # print "threadsMax=" + str(threadsMax)

          threadsActual = 0
          try:
            # Threads actual
            logger.debug("CPU Threads actual srs:\n" + str(groupsItems.group(6)))
            threadsActual = int(groupsItems.group(6).strip())
            logger.debug("threadsActual:" + str(threadsActual))
          except Exception, e:
            logger.error("Error threads actual: " + str(e))
          resource.threads = threadsActual

          lineSWAP = u''
          oldFormat = True
          lineSWAP = 3
          try:
            # RAM
            logger.debug("RAM srs:\n" + str(groupsItems.group(7)))
            lines = groupsItems.group(7).split("\n")
            if u'buffers/cache' in lines[3]:
              oldFormat = False
            if not oldFormat:
              lineSWAP = 4
          except Exception, e:
            logger.error("Error of RAM lines preparation: " + str(e) + ", source buffer:\n" + str(groupsItems.group(7)))
          logger.debug("RAM oldFormat:" + str(oldFormat) + ", lineSWAP:" + str(lineSWAP))
          ramTotal = 0
          ramUsed = 0
          ramCached = 0
          ramVirt = 0
          ramVirtUsed = 0
          try:
            linesRAM = [item for item in lines[2].split(u' ') if item]
            if len(linesRAM) > 6:
              ramTotal = int(linesRAM[1])
              ramUsed = int(linesRAM[2])
              ramCached = int(linesRAM[5])
            else:
              if oldFormat:
                o1 = 4
                o2 = 18
                o3 = 19
                o4 = 31
                o5 = 55
                o6 = 67
              else:
                o1 = 4
                o2 = 18
                o3 = 19
                o4 = 29
                o5 = 63
                o6 = len(lines[2])
              # RAM total
              ramTotal = int(lines[2][o1: o2].strip()) * 1000
              # print "[" + str(ramTotal) + "]"
              # RAM used
              ramUsed = int(lines[2][o3: o4].strip()) * 1000
              # print "[" + str(ramUsed) + "]"
              # RAM cached
              ramCached = int(lines[2][o5:o6].strip()) * 1000
              # print "[" + str(ramCached) + "]"
          except Exception, e:
            logger.error("Error of some RAM value: " + str(e) + "\n" + str(groupsItems.group(7)))
          logger.debug("ramTotal:" + str(ramTotal))
          logger.debug("ramUsed:" + str(ramUsed))
          logger.debug("ramCached:" + str(ramCached))
          resource.ramR = ramTotal
          resource.ramRU = ramUsed
          resource.ramC = ramCached
          #resource.ramRU = resource.ramRU - ramCached
          resource.ramV = ramVirt
          resource.ramVU = ramVirtUsed
          #RAM swap
          swapTotal = 0
          swapUsed = 0
          try:
            linesSWAP = [item for item in lines[lineSWAP].split(u' ') if item]
            if len(linesSWAP) > 2:
              swapTotal = int(linesSWAP[1])
              swapUsed = int(linesSWAP[2])
            else:
              if oldFormat:
                o7 = 5
                o8 = 19
                o9 = 19
                o10 = 31
              else:
                o7 = 5
                o8 = 18
                o9 = 19
                o10 = 29
              # Swap total
              swapTotal = int(lines[lineSWAP][o7: o8].strip()) * 1000
              # print "swapTotal=" + str(swapTotal)
              # Swap used
              swapUsed = int(lines[lineSWAP][o9: o10].strip()) * 1000
              # print "swapUsed=" + str(swapUsed)
          except Exception, e:
            logger.error("Error of some swap value: " + str(e))
          resource.swap = swapTotal
          resource.swapU = swapUsed
          logger.debug("swapTotal:" + str(swapTotal))
          logger.debug("swapUsed:" + str(swapUsed))

          diskTotal = 0
          diskUsed = 0
          try:
            # Disk
            logger.debug("Disk srs:\n" + str(groupsItems.group(8)))
            lines = groupsItems.group(8).split("\n")
            for l in lines:
              if u'/dev/sd' in l or u'/dev/md' in l:
                logger.debug("disk line src:\n" + str(l))
                linesItems = (' '.join(l.split())).split(" ")
                # Disk total
                diskTotal += int(linesItems[1]) * 1024
                # Disk used
                diskUsed += int(linesItems[2]) * 1024
          except Exception, e:
            logger.error("Error of some disk value: " + str(e))
          resource.disk = diskTotal
          resource.diskU = diskUsed
          logger.debug("diskTotal:" + str(diskTotal))
          logger.debug("diskUsed:" + str(diskUsed))
        else:
          logger.error(self.ERROR_RE_PARSE_NO_MATCHES + ", node:" + resource.nodeName + "\nstdout:\n" +
                       taskResponseObjects.items[0].stdout + "\nstderr:\n" + taskResponseObjects.items[0].stderror)
          resource = None
      else:
        logger.error(self.ERROR_DRCE_RESPONSE_ERROR_CODE + " : " + str(taskResponseObjects.items[0].error_code) + \
                     " : " + taskResponseObjects.items[0].error_message)
        resource = None
    else:
      logger.error(self.ERROR_NO_ITEMS_IN_DRCE_RESPONSE)

    return resource



  # #Send UpdateResourceData request to the ResourcesManager object
  #
  # @param resourcesList The list of Resource objects to send
  def sendUpdateResourceDataRequest(self, resourcesList):
    # Get TaskManager fields
    # Prepare synch GetTaskFields request to the TasksManager
    updateResourceDataEvent = self.eventBuilder.build(DTM_CONSTS.EVENT_TYPES.UPDATE_RESOURCES_DATA, resourcesList)
    updateResourceDataEvent.cookie = resourcesList
    self.send(self.clientResourcesManagerName, updateResourceDataEvent)



  # #Get the list of admin.Node objects from the cluster schema representation
  #
  # @param clusterSchemaObj The schema dic object
  # @return the list of admin.Node objects or empty list if error of schema object structure
  def getConnectedNodesFromSchema(self, clusterSchemaObj):
    nodes = []
    # Not sure in structure integrity
    try:
      # Seek on connected data node
      for item in clusterSchemaObj["cluster"]["nodes"]:
        if (item["role"] == "replica" or item["role"] == "shard") and ('connection' in item):
          port = item["admin"].split(":")
          if len(port) > 2:
            nodes.append(admin.Node.Node(item["host"], port[2]))
    except Exception, e:
      logger.error(self.ERROR_HCE_CLUSTER_SCHEMA_STRUCTURE_WRONG + " : " + str(e.message))

    return nodes



  # #Send to EE transport node admin connection
  #
  # @param host HCE node host
  # @param port HCE node port
  # @param messageParameters HCE node Admin request message parameters string
  # @return the raw body of HCE Admin API response
  def sendToHCENodeAdmin(self, host, port, messageParameters):
    response = None

    logger.debug("sendToHCENodeAdmin() use Host: '%s' and Port '%s'", str(host), str(port))

    # Execute EE node admin request
    node = admin.Node.Node(host, port)
    params = [messageParameters]
    command = admin.Command.Command(admin.Constants.COMMAND_NAMES.DRCE,
                                    params,
                                    admin.Constants.ADMIN_HANDLER_TYPES.DATA_PROCESSOR_DATA
                                    )
    requestBody = command.generateBody()
    message = {admin.Constants.STRING_MSGID_NAME : self.drceIdGenerator.get_uid(),
               admin.Constants.STRING_BODY_NAME : requestBody}
    try:
      logger.debug("!!! Before makeRequest() msg_id = %s", str(message[admin.Constants.STRING_MSGID_NAME]))
      response = self.hceNodeManagerRequest.makeRequest(node, message, self.hceNodeAdminTimeout)
    except Exception, e:
      logger.error(self.ERROR_HCE_NODE_REQUEST_ERROR + " : " + str(e.message))

    if response is None:
      return response
    else:
      return response.getBody()



  # #Get dict object from json file
  #
  # @param filePath The file path name
  # @return the dict object or None if error
  def getObjectFromJsonFile(self, filePath):
    # Initialize returned dic
    schemaDic = None

    # Load schema from file
    schemaJsonString = self.loadFromFile(filePath)
    if schemaJsonString is not None:
      try:
        # Decode json
        # schemaDic = json.loads(schemaJsonString).decode('utf-8')
        schemaDic = json.loads(str(schemaJsonString))
      except ValueError:
        logger.error(self.ERROR_JSON_DECODE)
        logger.debug(schemaJsonString)

    return schemaDic



  # #Get DRCE request json formated to fetch resources state of HCE node
  # Supposes execution of regular Linux commands and accumulation the textual printout to parse
  #
  # @param filePath The file pathname
  # @return the file content or None if error
  def loadFromFile(self, filePath):
    # Initialize file content
    fileContent = None

    try:
      fileContent = open(filePath, 'r').read()
    except IOError:
      logger.error(self.ERROR_READ_FILE + " " + filePath)

    return fileContent



